//--------------------------------------------------------------------
// $Id: CJM_algorithm.hpp 4721 2007-09-15 04:03:10Z cjm $
//--------------------------------------------------------------------
//
//   STL Compatible Algorithms
//   Copyright 1998 by Christopher J. Madsen
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License along
//   with this program; if not, write to the Free Software Foundation, Inc.,
//   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//--------------------------------------------------------------------

#ifndef INCLUDED_CJM_ALGORITHM_HPP
#define INCLUDED_CJM_ALGORITHM_HPP
#pragma once

//--------------------------------------------------------------------
// Remember the original order of a container:
//
// Input:
//   before:  A copy of the container before it was sorted
//   after:   An iterator at the beginning of the sorted container
//
// Output:
//   order:
//     If before[B] == after[A], then order[B] == A

template<class Cont, class IdxCont>
void original_order(IdxCont& order, const Cont& before,
                    Cont::const_iterator after)
{
  order.resize(before.size());

  Cont::const_iterator  e = before.begin();
  const Cont::const_iterator  aEnd = after + before.size();
  for (IdxCont::iterator i = order.begin(); i != order.end(); ++i, ++e)
    *i = find(after, aEnd, *e) - after;
} // end original_order

//--------------------------------------------------------------------
// Reorder a container based on a position vector:
//
// Input:
//   order:  A list such as that generated by original_order
//   itr:    An iterator at the beginning of the container

template<class Cont, class IdxCont> inline
void reorder(const IdxCont& order, Cont::iterator itr)
{
  const Cont  temp(itr, itr + order.size());
  for (IdxCont::const_iterator i = order.begin(); i != order.end(); ++i, ++itr)
    *itr = temp[*i];
} // end reorder

//--------------------------------------------------------------------
// Reorder a container based on a position vector:
//
// Input:
//   order:  A list such as that generated by original_order
//   itr:    An iterator at the beginning of the container
//
// Output:
//   order:  A list which can be passed to reorder to reverse the operation

template<class Cont, class IdxCont>
void reorder_with_undo(IdxCont& order, Cont::iterator itr)
{
  const Cont  temp(itr, itr + order.size());
  IdxCont     undo(order.size());
  int j = -1;
  for (IdxCont::const_iterator i = order.begin(); i != order.end(); ++i,++itr) {
    *itr = temp[*i];
    undo[*i] = ++j;
  }

  order.swap(undo);
} // end reorder_with_undo

//--------------------------------------------------------------------
// Fill a container with values in ascending order:
//
// Input:
//   cont:   The container to resize
//   size:   The number of elements the container should have
//   first:  The value of the first element
//
// Output:
//   cont:   The container, now with SIZE elements in ascending order

template<class Cont> inline
void resize_ascending(Cont& cont, Cont::size_type size, Cont::value_type first)
{
  cont.resize(size);
  Cont::iterator i = cont.begin();
  while (i != cont.end())
    *(i++) = first++;
} // end resize_ascending

//--------------------------------------------------------------------
// Compute the new order from the original order list:
//
// Input:
//   original:  A list like that created by original_order
//
// Output:
//   order:     A list that is the reverse of the original order

template<class IdxCont> inline
void reverse_order_list(IdxCont& order, const IdxCont& original)
{
  order.resize(original.size());
  int j = -1;
  for (IdxCont::const_iterator i = original.begin();
       i != original.end(); ++i)
    order[*i] = ++j;
} // end reverse_order_list

//--------------------------------------------------------------------
// Insert an element into a container if it's not already there:
//
// Input:
//   cont:     The container (must be sorted in ascending order)
//   element:  The element to add
//
// Output:
//   cont:  The updated container
//
// Returns:
//   true:   The element was added
//   false:  The element was already in the container

template<class Cont>
bool insert_unique(Cont& cont, Cont::const_reference element)
{
  Cont::iterator  i = lower_bound(cont.begin(), cont.end(), element);

  if ((i != cont.end()) && (*i == element))
    return false;

  cont.insert(i, element);
  return true;
} // end insert_unique

#endif // INCLUDED_CJM_ALGORITHM_HPP
