//--------------------------------------------------------------------
// $Id: Clip.cpp 4718 2007-09-15 02:34:13Z cjm $
//--------------------------------------------------------------------
//
//   GUEmap
//   Copyright 1997-2007 by Christopher J. Madsen
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License along
//   with this program; if not, write to the Free Software Foundation, Inc.,
//   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//   Clipboard support
//
//--------------------------------------------------------------------

#include "StdAfx.hpp"
#include "GUEmap.hpp"

#include "MapDoc.hpp"
#include "Undo.hpp"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//====================================================================
// Class CMapApp:
//--------------------------------------------------------------------
// Store rooms on the clipboard:
//
// Input:
//   newScrap:
//     A pointer to the new clipboard contents
//     May be NULL, in which case the clipboard is emptied.
//     The application owns the RoomScrap, and will delete it when done.

void CMapApp::setClipboard(RoomScrap* newScrap)
{
  delete clipboard;
  clipboard = newScrap;
} // end CMapApp::setClipboard

//====================================================================
// Class RoomScrap:
//
// Stores rooms and their connections on the clipboard.
//
// Member Variables:
//   edges:  The edges we copied
//   rooms:  The rooms we copied
//
//--------------------------------------------------------------------
// Constructor:
//
// Copy all selected rooms and the edges that connect them.  An edge
// that leads from a selected room to an unselected room is copied as
// a stub.
//
// Input:
//   doc:               The map to copy from
//   numSelected:       The number of selected rooms
//   selectedRooms:     TRUE means the room is selected
//   numSelectedPages:  The number of selected pages
//   selectedPages:     TRUE means the page is selected

RoomScrap::RoomScrap(const CMapDoc& doc, int numSelected,
                     const ByteVec& selectedRooms, int numSelectedPages,
                     const ByteVec& selectedPages)
{
  int i;

  pages.resize(numSelectedPages);
  PageItr  p = pages.end();
  for (i = doc.getPages().size()-1; numSelectedPages; --i)
    if (selectedPages[i]) {
      --numSelectedPages;
      ASSERT(p != pages.begin());
      *(--p) = doc.getPage(i);
    }

  const int   numRooms = doc.getRoomCount();
  RoomNumVec  roomNums;

  roomNums.resize(numRooms, 0);
  rooms.resize(numSelected);
  const RoomVec&  room = doc.getRooms();

  RoomItr  r = rooms.getVector().end();
  for (i = doc.getRoomCount()-1; numSelected; i--)
    if (selectedRooms[i]) {
      roomNums[i] = --numSelected;
      ASSERT(r != rooms.getVector().begin());
      *(--r) = new MapRoom(*room[i]);
    }

  EdgeConstItr edge = doc.getEdges().begin();
  MapEdge e;
  for (i = doc.getEdgeCount(); i > 0; edge++, i--) {
    if (selectedRooms[edge->room1] &&
        ((edge->type1 & etNoRoom2) || selectedRooms[edge->room2])) {
      e = *edge;
      e.room1 = roomNums[e.room1];
      if (!(e.type1 & etNoRoom2))
        e.room2 = roomNums[e.room2];
      edges.push_back(e);
    } // end if edge connects selected rooms
    else if (!(edge->type1 & etNoRoom2)) {
      if (selectedRooms[edge->room1]) {
        e.room1 = edge->room1;
        e.type1 = edge->type1;
        e.end1  = edge->end1;
      } else if (selectedRooms[edge->room2] && !(edge->type1 & etOneWay)) {
        e.room1 = edge->room2;
        e.type1 = edge->type2;
        e.end1  = edge->end2;
      } else
        continue; // This edge does not come from a selected room
      // This edge has one end at a selected room, make it a stub:
      e.room1 = roomNums[e.room1];
      e.type1 &= ~etSpecial;
      e.type1 |= etUnexplored;
      e.type2 = etNormal;
      e.room2 = 0;
      e.end2  = rcSW;
      edges.push_back(e);
    } // end else if edge connects two rooms
  } // end for each edge in map
} // end RoomScrap::RoomScrap

//--------------------------------------------------------------------
// Paste the rooms into a map:
//
// Rooms are appended to the end of the room list, and the edges are
// fixed to reflect the new room numbers.
//
// Input:
//   doc:  The map to paste into
//
// Returns:
//   A pointer to the undo record

UndoRec* RoomScrap::paste(CMapDoc& doc) const
{
  const bool  wasModified = doc.IsModified();

  PageNumVec  pageNums;
  if (!pages.empty()) {
    const int n = doc.getPages().size();
    doc.addPages(pages);
    resize_ascending(pageNums, doc.getPages().size() - n, n);
  }

  RoomNumVec roomNums;
  roomNums.resize(rooms.size());

  const int n = doc.getRoomCount();
  int i;
  doc.addRooms(rooms.size());
  for (i = 0; i < rooms.size(); i++) {
    doc.addRoom(n+i, new MapRoom(rooms[i]));
    roomNums[i] = n+i;
  }

  doc.addEdges(edges.size());
  for (i = 0; i < edges.size(); i++) {
    MapEdge e = edges[i];
    e.room1 += n;
    if (!(e.type1 & etNoRoom2)) e.room2 += n;
    doc.addEdge(e);
  }

  return new UndoAdd(wasModified, edges.size(), roomNums, pageNums);
} // end RoomScrap::paste

/////////////////////////////////////////////////////////////////////////////
// Local Variables:
// cjm-related-file: "MapDoc.hpp"
// End:
