//--------------------------------------------------------------------
// $Id: Export.cpp 4718 2007-09-15 02:34:13Z cjm $
//--------------------------------------------------------------------
//
//   GUEmap
//   Copyright 1997-2007 by Christopher J. Madsen
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License along
//   with this program; if not, write to the Free Software Foundation, Inc.,
//   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//   Export room data to a text file
//
//--------------------------------------------------------------------

#include "StdAfx.hpp"
#include "GUEmap.hpp"

#include "MapDoc.hpp"

#include <fstream>
using namespace std;

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

class RoomName
{
 public:
  String   name;
  RoomNum  num;

  RoomName() {};
  void assign(RoomNum n, const MapRoom& r);
  bool operator<(const RoomName& r)  const { return name < r.name;  };
  bool operator==(const RoomName& r) const { return name == r.name; };
}; // end RoomName

typedef vector<RoomName> RVec;

//====================================================================
// Miscellaneous Functions:
//--------------------------------------------------------------------
// Output a string with word wrap:
//
// String must use CRLF, not LF.
//
// FIXME doesn't handle Tab characters properly.
//
// Input:
//   out:  The stream to write to
//   str:  The string to write

const StrIdx  maxlen = 70;      // Maximum line length

void wordWrap(ostream& out, const String& str)
{
  StrConstItr start,end,next;
  const StrConstItr  strend = str.end();

  next = str.begin();
  while (next != strend) {
    start = end = next;
    next += min(maxlen, strend - start); // Maximum value for end

    // Fill up the line until it's full or we find a CR:
    while ((end < next) && (*end != _T('\r')))
      ++end;

    // Compute next (the first char of the next line)
    // and end (just after the last char of this line):
    if (end == strend) // This is the last line
      next = end;
    else if (*end == _T('\r')) // We found a hard return
      next = end + 2;
    else if (isspace(*end)) { // We can break the line right here
      // Begin the next line at the the next non-space character:
      next = end+1;
      while ((next < strend) && isspace(*next) && (*next != _T('\r')))
        ++next;
      if (*next == _T('\r')) next += 2; // Hard return preceeded by space
    } else { // Must back up and find a place to break the line
      next = end;               // Remember where we are
      // FIXME allow break at hyphen
      while ((end > start) && !isspace(*end)) // Find the last space
        --end;
      if (start == end)
        end = next;             // No spaces on entire line
      else
        next = end+1;           // We found a space, break here
    } // end else back up and find a breakpoint

    // Don't print trailing spaces:
    while ((end > start) && isspace(end[-1]))
      --end;

    out.write(start, end - start);
    out << endl;
  } // end while more characters in string
} // end wordWrap

//====================================================================
// Class RoomName:
//
// RoomName stores a 1-line room name and its room number, so a vector
// of them can be sorted.
//
// Member Variables:
//   name:  The room name (with CRLF changed to space)
//   num:   The room number
//--------------------------------------------------------------------
// Initialize the record with room name and number:
//
// Converts CRLF in the room name to a space.
//
// Input:
//   n:  The room number
//   r:  The room

void RoomName::assign(RoomNum n, const MapRoom& r)
{
  copyToOneLine(name, r.name);
  num = n;
} // end RoomName::operator=

//====================================================================
// Class CMapDoc:
//--------------------------------------------------------------------
// Export room comments:

void CMapDoc::OnFileExport()
{
  String path;

  // Copy mapfile.txt into path (default name for dialog):
  const CString& docPath = GetPathName();
  if (!docPath.IsEmpty()) {
    int start = docPath.ReverseFind(_T('\\')) + 1;
    if (start > 0) {
      int end = docPath.ReverseFind(_T('.'));
      if (end == -1) end = docPath.GetLength();
      path.assign(LPCTSTR(docPath) + start, end - start);
      path += _T(".txt");
    }
  } // end if document path is not empty

  CFileDialog fd(FALSE, _T("txt"), path.c_str(),
                 OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT,
                 _T("Text files (*.txt)|*.txt|All Files (*.*)|*.*||"));

  fd.m_ofn.lpstrTitle = _T("Export Comments");

  if (fd.DoModal() != IDOK) return;

  Strcpy(path, fd.GetPathName());

  // Build a list of rooms which have comments:
  RVec rooms;
  rooms.reserve(room.size());

  RoomConstItr  r(room.getVector().begin());
  for (RoomNum n=0; r != room.getVector().end(); ++r, ++n)
    if (!(*r)->note.empty()) {
      rooms.resize(rooms.size()+1);
      rooms.back().assign(n, **r);
    }

  sort(rooms.begin(), rooms.end());

  // Write room comments to a file:
  ofstream out(path.c_str());

  // Print the map title & comments:
  if (!name.empty()) out << name << endl;
  if (!note.empty()) wordWrap(out, note);
  if (!name.empty() || !note.empty()) out << endl;

  // Print the room comments:
  for (RVec::const_iterator  rp(rooms.begin()); rp != rooms.end(); rp++) {
    out << rp->name << endl;
    wordWrap(out, room[rp->num].note);
    out << endl;
  }

  out.close();
} // end CMapDoc::OnFileExport

//////////////////////////////////////////////////////////////////////
// Local Variables:
// compile-command: "nmake -f GUEmap.mak"
// End:
