//--------------------------------------------------------------------
// $Id: File.cpp 4718 2007-09-15 02:34:13Z cjm $
//--------------------------------------------------------------------
//
//   GUEmap
//   Copyright 1997-2007 by Christopher J. Madsen
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License along
//   with this program; if not, write to the Free Software Foundation, Inc.,
//   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//   Implementation of the FileReader class
//
//--------------------------------------------------------------------

#include "StdAfx.hpp"

#include "File.hpp"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

//====================================================================
// Miscellaneous functions:
//--------------------------------------------------------------------
// Write a string to a CArchive:
//
// The length is stored (LSB first) followed by the data.
//   0x00   - 0xFE    byte
//   0xFF   - 0xFFFE  FF followed by word
//   0xFFFF ->        FFFFFF followed by double word
//
// Input:
//   ar:  The CArchive to write to
//   s:   The String to write
//
// Returns: ar

CArchive& operator<<(CArchive& ar, const String& s)
{
  const StrIdx  len(s.length());

  if (len < 0xFF)
    ar << BYTE(len);
  else {
    ar << BYTE(0xFF);
    if (len < 0xFFFF)
      ar << WORD(len);
    else
      ar << WORD(0xFFFF) << DWORD(len);
  }

  if (len)
    ar.Write(s.c_str(), len);

  return ar;
} // end CArchive operator<< String

//--------------------------------------------------------------------
// Read a string from a CArchive:
//
// The length is stored (LSB first) followed by the data.
//   0x00   - 0xFE    byte
//   0xFF   - 0xFFFE  FF followed by word
//   0xFFFF ->        FFFFFF followed by double word
//
// Input:
//   ar:  The CArchive to read from
//   s:   The String to store into
//
// Returns: ar

CArchive& operator>>(CArchive& ar, String& s)
{
  StrIdx  len;
  BYTE    lenB;

  ar >> lenB;

  if (lenB != 0xFF)
    len = lenB;
  else {
    WORD lenW;
    ar >> lenW;
    if (lenW != 0xFFFF)
      len = lenW;
    else {
      DWORD lenDW;
      ar >> lenDW;
      len = lenDW;
    }
  }

  if (len) {
    s.resize(len, 0);
    ar.Read(s.begin(), len);    // FIXME
    *(s.begin() + len) = _T('\0');
  } else
    s.erase();

  return ar;
} // end CArchive operator>> String

//====================================================================
// Class FileReader:
//--------------------------------------------------------------------
// Member Variables:
//   ar:  The CArchive we're reading from
//   rcCache:
//     A cached record code (from peekNextRecord)
//      -1 if cache is empty
//       0 if at end of file
//
//--------------------------------------------------------------------
// Constructor:
//
// Input:
//   archive:  The CArchive to read from

FileReader::FileReader(CArchive& archive)
: ar(archive),
  rcCache(-1)
{
} // end FileReader::FileReader

//--------------------------------------------------------------------
// Skip to the beginning of a record:
//
// Throws a badIndex (corrupt file) exception if the stream is not at
// the beginning of a record.  If the record is not of the specified
// type, it is skipped and the process repeats until the specified
// type of record is found.  If EOF is encountered while searching, it
// throws endOfFile.
//
// If no errors occur, the record header is removed from the stream.
//
// Input:
//   type:  The record type to search for

void FileReader::findRecord(BYTE type)
{
  BYTE rc,c;

  for (;;) {                    // Repeat until error or success
    // Make sure we are at the beginning of a record:
    if (rcCache > 0) {
      rc = rcCache;
      rcCache = -1;
    } else {
      if ((rcCache == 0) || (byte() != fcStartRecord)) {
       corrupt:
        AfxThrowArchiveException(CArchiveException::badIndex,
                                 ar.GetFile()->GetFilePath());
      }
      rc = byte();
    } // end else no record code in cache

    // See if we found the right kind of record:
    if (rc == type) return;

    // Wrong record type, so skip fields until we find the end of the record
    while ((c = byte()) != fcEndRecord) {
//      if ((c & fcExtended) == fcExtended)
//        byte();
      skipField(c);
    }

    if (byte() != rc)
      // The ending record type does not match the beginning type
      goto corrupt;
  } // end forever
} // end FileReader::findRecord

//--------------------------------------------------------------------
// Find out what type the next record is:
//
// The stream must be at the beginning of a record or at end of file.
//
// Returns:
//   The type of the record about to be read
//   0 if stream is at end of file (WITHOUT throwing endOfFile)

BYTE FileReader::peekNextRecord()
{
  if (rcCache >= 0)
    return rcCache;

  BYTE b;

  try {
    ar >> b;
  } catch (CArchiveException* e) {
    if (e->m_cause == CArchiveException::endOfFile) {
      e->Delete();
      return (rcCache = 0);     // Signal end-of-file
    }
    throw;                      // Re-throw any other exception
  } // end catch

  if (b != fcStartRecord)
    AfxThrowArchiveException(CArchiveException::badIndex,
                             ar.GetFile()->GetFilePath());

  ar >> b;

  rcCache = b;
  return b;
} // end FileReader::peekNextRecord

//--------------------------------------------------------------------
// Skip an unrecognized field:
//
// Skips over an unknown or unwanted field.  If this is an extended
// field code, you MUST read the extended type info before calling
// skipField().
//
// Input:
//   fieldCode:  The field code for this field.  For an extended field,
//               this is the fcExtended code (with size information),
//               not the extended ID byte.

void FileReader::skipField(BYTE fieldCode)
{
  DWORD  size = fieldCode & fcSizeMask; // Compute the field size

  switch (size) {
   case 0: // String
     size = byte();
     if (size == 0xFF) {
       size = word();
       if (size == 0xFFFF) ar >> size;
     }
     break;
   case 1:                      // 1-4 byte fields
   case 2:
   case 3:
   case 4:
    break;
   case 5: size = 8;      break; // 8 byte field
   case 6: size = byte(); break; // Variable length field with size byte
   case 7: size = word(); break; // Variable length field with size word
#ifdef _lint
   default:
    //lint -unreachable
#endif
  } // end switch

  BYTE b;

  while (size-- > 0)
    ar >> b;
} // end FileReader::skipField

//====================================================================
// Local Variables:
// compile-command: "nmake -f GUEmap.mak"
// End:
