//--------------------------------------------------------------------
// $Id: FindDlg.cpp 4718 2007-09-15 02:34:13Z cjm $
//--------------------------------------------------------------------
//
//   GUEmap
//   Copyright 1996-2007 by Christopher J. Madsen
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License along
//   with this program; if not, write to the Free Software Foundation, Inc.,
//   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//   Implementation of the CFindDlg class
//
//--------------------------------------------------------------------

#include "StdAfx.hpp"
#include "GUEmap.hpp"
#include "CommentDlg.hpp"
#include "FindDlg.hpp"
#include "MapDoc.hpp"
#include "MapView.hpp"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

const int searchInComments = 0;

/////////////////////////////////////////////////////////////////////////////
// CFindDlg dialog

CFindDlg::CFindDlg(CCommentDlg* theCommentDlg, CWnd* pParent /*=NULL*/)
: CDialog(CFindDlg::IDD, pParent),
  commentDlg(theCommentDlg),
  match(String::npos),
  whichRoom(-2)
{
	//{{AFX_DATA_INIT(CFindDlg)
	what = _T("");
	matchCase = FALSE;
	searchIn = 0;
	//}}AFX_DATA_INIT
} // end CFindDlg::CFindDlg


void CFindDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CFindDlg)
	DDX_CBString(pDX, IDC_FIND_WHAT, what);
	DDX_Radio(pDX, IDC_SEARCH_COMMENTS, searchIn);
	DDX_Check(pDX, IDC_FIND_CASE, matchCase);
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(CFindDlg, CDialog)
	//{{AFX_MSG_MAP(CFindDlg)
	ON_BN_CLICKED(IDOK, OnFindNext)
	ON_BN_CLICKED(IDC_FIND_EDIT, OnEditComments)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP();

//====================================================================
// CFindDlg searching (based on Jul 97 Snippets):
//--------------------------------------------------------------------

const int LARGE = 0x20000;

/* +++Date last modified: 05-Jul-1997 */

/*
**  Case-sensitive Boyer-Moore-Horspool pattern match
**
**  public domain by Raymond Gardner 7/92
**
**  limitation: pattern length + string length must be less than 0x20000
**
**  10/21/93 rdg  Fixed bug found by Jeff Dunlop
*/

typedef unsigned char uchar;

void CFindDlg::bmh_init(bool ignoreCase)
{
  int i, lastpatchar;
  const int patlen = what.GetLength();

  for (i = 0; i <= UCHAR_MAX; ++i)                  /* rdg 10/93 */
    skip[i] = patlen;
  for (i = 0; i < patlen; ++i)
    skip[what[i]] = patlen - i - 1;
  lastpatchar = what[patlen - 1];
  skip[lastpatchar] = LARGE;
  skip2 = patlen;                 /* Horspool's fixed second shift */
  for (i = 0; i < patlen - 1; ++i) {
    if (what[i] == lastpatchar)
      skip2 = patlen - i - 1;
  }

  if (ignoreCase)
    for (i = 0; i <= UCHAR_MAX; ++i)
      if (tolower(i) != i)
        skip[tolower(i)] = skip[i];
} // end CFindDlg::bmh_init

StrIdx CFindDlg::bmh_search(const String& str)
{
  const int patlen1 = what.GetLength() - 1;
  int i, j;
  StrConstItr s;

  i = patlen1 - str.size();
  if (i >= 0)
    return String::npos;
  const StrConstItr  strend = str.end();
  for ( ;; ) {
    while ( (i += skip[((uchar *)strend)[i]]) < 0 )
      ;                           /* mighty fast inner loop */
    if (i < (LARGE - str.size()))
      return String::npos;
    i -= LARGE;
    j = patlen1;
    s = strend + (i - j);
    while (--j >= 0 && s[j] == what[j])
      ;
    if ( j < 0 )
      return (s - str.begin());
    if ( (i += skip2) >= 0 )
      return String::npos;
  }
} // end CFindDlg::bmh_search

StrIdx CFindDlg::bmhi_search(const String& str)
{
  const int patlen1 = what.GetLength() - 1;
  int i, j;
  StrConstItr s;

  i = patlen1 - str.size();
  if (i >= 0)
    return String::npos;
  const StrConstItr  strend = str.end();
  for ( ;; ) {
    while ( (i += skip[((uchar *)strend)[i]]) < 0 )
      ;                           /* mighty fast inner loop */
    if (i < (LARGE - str.size()))
      return String::npos;
    i -= LARGE;
    j = patlen1;
    s = strend + (i - j);
    while (--j >= 0 && toupper(s[j]) == what[j])
      ;
    if ( j < 0 )
      return (s - str.begin());
    if ( (i += skip2) >= 0 )
      return String::npos;
  }
} // end CFindDlg::bmhi_search

//====================================================================
// Read the search string and initialize the search engine:
//
// Also adds the string to the history list.
//
// Returns:
//   true:   Pattern is ready for searching
//   false:  Can't search

bool CFindDlg::updatePattern()
{
  CString  oldPattern(what);
  BOOL     oldCase(matchCase);

  if (!UpdateData(TRUE) || what.IsEmpty()) return false;

  if (mapView != commentDlg->fromView) {
     mapView = commentDlg->fromView;
     whichRoom = -2;
  }
  if (!mapView) return false;

  // Add search string to the history:
  int  i = -1;
  int  j;
  CString  s;
  while ((j = history.FindStringExact(i, what)) > i) {
    i = j;
    history.GetLBText(i, s);    // FindStringExact ignores case
    if (s == what) break;  // It matches
    else           j = -1; // Not a match
  } // end while searching for existing string

  if (j) {
    if (j > 0) history.DeleteString(j); // Moving existing string
    history.InsertString(0, what);
    if (history.GetCount() > 24)
      history.DeleteString(24);
    history.SetCurSel(0);
  } // end if string is not already the first item in the history list

  // Now initialize the pattern:
  if (!matchCase) what.MakeUpper();

  if ((oldCase != matchCase) || (oldPattern != what)) {
    bmh_init(!matchCase);
    whichRoom = -2;
  }
  return true;
} // end CFindDlg::updatePattern

/////////////////////////////////////////////////////////////////////////////
// CFindDlg message handlers
//--------------------------------------------------------------------
void CFindDlg::OnCancel()
{
  GetWindowRect(&pos);
  if (!matchCase) what.MakeLower();
  DestroyWindow();
} // end CFindDlg::OnCancel

//--------------------------------------------------------------------
void CFindDlg::OnEditComments()
{
  if (match == String::npos)
      gueApp()->editComment();
  else {
    if (commentDlg->GetSafeHwnd() == NULL)
      gueApp()->OnViewComments();
    else
      commentDlg->SetFocus();

    commentDlg->SendDlgItemMessage(IDC_CD_COMMENT, EM_SETSEL,
                                   match, match + what.GetLength());
    commentDlg->SendDlgItemMessage(IDC_CD_COMMENT, EM_SCROLLCARET, 0,0);
  }
} // end CFindDlg::OnEditComments

//--------------------------------------------------------------------
void CFindDlg::OnFindNext()
{
  GotoDlgCtrl(GetDlgItem(IDC_FIND_WHAT));

  if (!updatePattern())
    return;

  const CMapDoc*  doc = mapView->GetDocument();
  ASSERT_VALID(doc);

  const int
    startRoom = whichRoom,
    roomCount = doc->getRoomCount();
  int tries = roomCount;

  match = String::npos;
  StrIdx  (CFindDlg::*searchFunc)(const String& s);

  if (matchCase)
    searchFunc = &CFindDlg::bmh_search;
  else
    searchFunc = &CFindDlg::bmhi_search;

  if (searchIn == searchInComments) {
    ++tries;                    // Search map comments too
    while (--tries >= 0) {
      if (++whichRoom >= roomCount) whichRoom = -1;

      if (whichRoom < 0) {
        if ((match = (this->*searchFunc)(doc->getNote())) != String::npos) {
          gueApp()->setComment(mapView, NULL);
          break;
        }
      } else if ((match = (this->*searchFunc)(doc->getRoom(whichRoom).note))
                 != String::npos) {
        mapView->clearSelection(); // Make sure selectOnlyRoom does something
        mapView->selectOnlyRoom(whichRoom);
        break;
      }
    } // end while tries

    if (match == String::npos || startRoom == whichRoom)
      MessageBeep(MB_ICONASTERISK);
    else {
      if (commentDlg->GetSafeHwnd() == NULL) {
        gueApp()->OnViewComments();
        SetFocus();
      }
      if (whichRoom >= 0) mapView->makeRoomVisible(whichRoom);
      commentDlg->SendDlgItemMessage(IDC_CD_COMMENT, EM_SETSEL,
                                     match, match + what.GetLength());
      commentDlg->SendDlgItemMessage(IDC_CD_COMMENT, EM_SCROLLCARET, 0,0);
    }
  } // end if search in comments
  else {
    if (whichRoom < 0) whichRoom = -1;
    String  s;
    s.reserve(80);
    while (--tries >= 0) {
      if (++whichRoom >= roomCount) whichRoom = 0;

      copyToOneLine(s, doc->getRoom(whichRoom).name);
      if ((this->*searchFunc)(s) != String::npos) {
        mapView->clearSelection(); // Make sure selectOnlyRoom does something
        mapView->selectOnlyRoom(whichRoom);
        mapView->makeRoomVisible(whichRoom);
        return;
      } // end if found match
    } // end while tries

    MessageBeep(MB_ICONASTERISK); // Not found
  } // end else search in room names
} // end CFindDlg::OnFindNext

//--------------------------------------------------------------------
BOOL CFindDlg::OnInitDialog()
{
  CDialog::OnInitDialog();

  VERIFY(history.SubclassDlgItem(IDC_FIND_WHAT, this));

  return TRUE;  // return TRUE unless you set the focus to a control
} // end CFindDlg::OnInitDialog
