//--------------------------------------------------------------------
// $Id: FindPath.cpp 4718 2007-09-15 02:34:13Z cjm $
//--------------------------------------------------------------------
//
//   GUEmap
//   Copyright 1997-2007 by Christopher J. Madsen
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License along
//   with this program; if not, write to the Free Software Foundation, Inc.,
//   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//   Find the shortest path from one room to another
//
//--------------------------------------------------------------------

#include "StdAfx.hpp"
#include "GUEmap.hpp"

#include "MapDoc.hpp"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

enum StatusType {intree, fringe, unseen};
#define fringeList fringeLink[0]

static const char *const directions[rcNumDirections] =
{
  "n",   "s",   "e",   "w",
  "ne",  "nw",  "se",  "sw",
  "nne", "nnw", "sse", "ssw", NULL,
  "u",   "d",   "in",  "out",
}; // end directions

//====================================================================
// Class CMapDoc:
//--------------------------------------------------------------------
// Find the shortest path:

void CMapDoc::shortestPath(String& path, RoomNum start, RoomNum end) const
{
  const int numRooms = room.size();
  const int numEdges = edge.size();

  ByteVec     status;
  RoomNumVec  parent;
  ByteVec     dir;
  RoomNumVec  fringeLink;
  WordVec     dist;
  int         ptr;
  int         x,y;
  RoomCorner  yDir;
  EdgeType    yType;

  path.erase();

  dir.resize(numRooms, 0);
  dist.resize(numRooms, 0);
  parent.resize(numRooms, 0);
  fringeLink.resize(numRooms+1, 0);

  status.resize(numRooms, unseen);

  // We start at the ending room and work backwards,
  // so we can read the path forwards when we're done:
  status[end] = intree;
  dist[end]   = 0;
  parent[end] = 0;
  fringeList  = 0;

  x = end;

  while (x != start) {
    for (ptr = 0; ptr < numEdges; ++ptr) {
      const MapEdge& e = edge[ptr];
      if (!(e.type1 & etNoRoom2)) {
        if (x == e.room2) {
          y = e.room1;
          yDir = e.end1;
          yType = e.type1;
        } else if (x == e.room1 && !(e.type1 & etOneWay)) {
          y = e.room2;
          yDir = e.end2;
          yType = e.type2;
        } else continue;
      } else continue;
      switch (yType & etDirection) {
       case etUp:   yDir = rcUp;   break;
       case etDown: yDir = rcDown; break;
       case etIn:   yDir = rcIn;   break;
       case etOut:  yDir = rcOut;  break;
      }
      const short weight = (yDir == rcCorner ? 0 : 1);
      if (status[y] == unseen) {
        status[y] = fringe;
        fringeLink[y+1] = fringeList;
        fringeList = y+1;
        parent[y] = x;
        dir[y] = yDir;
        dist[y] = dist[x] + weight;
      } // end if y is unseen
      else if ((status[y] == fringe) && (dist[x] + weight < dist[y])) {
        parent[y] = x;
        dir[y] = yDir;
        dist[y] = dist[x] + weight;
      } // end else if y is on the fringe
    } // end for each edge

    if (!fringeList)
      return;  // There's nowhere left to go, return empty path
    else {
      ptr = 0;
      y = fringeLink[0];
      while (fringeLink[y]) {
        if (dist[fringeLink[y]-1] < dist[fringeLink[ptr]-1])
          ptr = y;
        y = fringeLink[y];
      }
      x = fringeLink[ptr]-1;
      fringeLink[ptr] = fringeLink[fringeLink[ptr]];
      status[x] = intree;
    } // end else
  } // end while x != start && !stuck

  while (x != end) {
    if (dir[x] != rcCorner) {
      if (!path.empty()) path += _T('.');
      path += directions[dir[x]];
    }
    x = parent[x];
  } // end while x != end
} // end CMapDoc::shortestPath

/////////////////////////////////////////////////////////////////////////////
// Local Variables:
// compile-command: "nmake -f GUEmap.mak"
// End:
