//--------------------------------------------------------------------
// $Id: MapDoc.cpp 4718 2007-09-15 02:34:13Z cjm $
//--------------------------------------------------------------------
//
//   GUEmap
//   Copyright 1996-2007 by Christopher J. Madsen
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License along
//   with this program; if not, write to the Free Software Foundation, Inc.,
//   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//   Implementation of the CMapDoc class
//
//--------------------------------------------------------------------

#include "StdAfx.hpp"
#include "GUEmap.hpp"

#include <AfxPriv.h>            // Need CMirrorFile

#include "File.hpp"
#include "MapDoc.hpp"
#include "Undo.hpp"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

const RoomCorner  oppositeCorner[rcNumCorners] =
{ rcS, rcN, rcW, rcE,
  rcSW, rcSE, rcNW, rcNE,
  rcSSW, rcSSE, rcNNW, rcNNE
}; // end oppositeCorner

const BYTE  fileHeader[8] = {0xC7,0x55,0xC5,0x6D,0xE1,0x70,0x83,0x0D};

//===========================================================================
// Class MapEdge:
//
// Member Variables:
//   end1:   The corner of room1 where this edge begins
//   end2:   The corner of room2 where this edge ends
//   room1:  The room where this edge begins
//   room2:  The room where this edge ends
//   type1:
//     The type of exit this is from room1
//     The etSpecial codes are only used with type1
//   type2:
//     The type2 of exit this is from room2

//===========================================================================
// Class MapRoom:
//
// Member Variables:
//   flags:
//     Bitmask of RoomFlags constants:
//       rfBorder:  Room has a border (default)
//       rfCorner:  Room is a corner connector (no name, one grid square)
//   pos:
//     The position of the lower left corner of the room
//     Given in logical MM_HIMETRIC coordinates (y is negative).
//   name:
//     The name of the room.
//   lines:
//     The number of lines in the room name (0 to maxNameLines)
//   nameLength[i]:
//     The number of characters on line i+1
//--------------------------------------------------------------------
MapRoom::MapRoom()
: flags(0),
  lines(0)
{
}

MapRoom& MapRoom::operator=(const MapRoom& source)
{
  pos = source.pos;
  name = source.name;
  note = source.note;
  flags = source.flags;
  lines = source.lines;
  memcpy(nameLength, source.nameLength, sizeof(nameLength));
  CJM_MEMCHECK;
  return *this;
} // end MapRoom::operator=

//---------------------------------------------------------------------------
// Compute the length of the room name:
//
// Input Variables:
//   name
//
// Output Variables:
//   name:  Has leading and trailing whitespace stripped
//   lines
//   nameLength

void MapRoom::computeNameLength()
{
  trimLeft(name);
  trimRight(name);

  int len = 0;
  lines = 0;
  if (!name.empty()) {
    StrConstItr n = name.begin();
    for (; n[len]; len++)
      if (n[len] == '\r') {
        nameLength[lines++] = len;
        if (lines == maxNameLines) return; // Name has too many lines to fit
        n += len + 2;             // Skip over \r\n
        len = -1;                 // Gets incremented to 0
      }
    nameLength[lines++] = len;
  } // end if name not empty
} // end MapRoom::computeNameLength

/////////////////////////////////////////////////////////////////////////////
// CMapDoc

IMPLEMENT_DYNCREATE(CMapDoc, CDocument)

BEGIN_MESSAGE_MAP(CMapDoc, CDocument)
	//{{AFX_MSG_MAP(CMapDoc)
	ON_COMMAND(ID_EDIT_NAVIGATE, OnNavigationMode)
	ON_COMMAND(ID_EDIT_UNDO, OnEditUndo)
	ON_COMMAND(ID_FILE_EXPORT, OnFileExport)
	ON_UPDATE_COMMAND_UI(ID_EDIT_NAVIGATE, OnUpdateNavigationMode)
	ON_UPDATE_COMMAND_UI(ID_EDIT_UNDO, OnUpdateEditUndo)
	//}}AFX_MSG_MAP
  // Send Mail command:
  ON_COMMAND(ID_FILE_SEND_MAIL, OnFileSendMail)
  ON_UPDATE_COMMAND_UI(ID_FILE_SEND_MAIL, OnUpdateFileSendMail)
  // My commands
  ON_COMMAND_RANGE(ID_EDIT_MOVE_BOTTOM, ID_EDIT_MOVE_TOP, OnEditMoveMap)
  ON_UPDATE_COMMAND_UI_RANGE(ID_EDIT_MOVE_BOTTOM, ID_EDIT_MOVE_TOP,
                             OnUpdateEditMoveMap)
END_MESSAGE_MAP();

/////////////////////////////////////////////////////////////////////////////
// CMapDoc construction/destruction

CMapDoc::CMapDoc()
: locked(false),
  pageSize(42 * gridX, 53 * gridY),
  undoData(NULL)
{
  // TODO: add one-time construction code here
} // end CMapDoc::CMapDoc

CMapDoc::~CMapDoc()
{
  delete undoData;
}

BOOL CMapDoc::OnNewDocument()
{
  if (!CDocument::OnNewDocument())
    return FALSE;

  // TODO: add reinitialization code here
  // (SDI documents will reuse this document)
  DeleteContents();

  return TRUE;
} // end CMapDoc::OnNewDocument

/////////////////////////////////////////////////////////////////////////////
// CMapDoc serialization
//--------------------------------------------------------------------
// Make a backup file when we Save As:

BOOL CMapDoc::DoSave(LPCTSTR pathName, BOOL replace)
{
  const bool  neededBackup = needBackup;
  if (!pathName) needBackup = true; // Saving to new name

  const BOOL  result = CDocument::DoSave(pathName, replace);
  if (!pathName && (!result || !replace))
    needBackup = neededBackup;  // Did not save (or saved a copy)

  return result;
} // end CMapDoc::DoSave

//--------------------------------------------------------------------
// Allocate a CFile object for loading/saving a file:
//
// Rename the existing file if this is the first save after a load.

CFile* CMapDoc::GetFile(LPCTSTR fileName, UINT openFlags, CFileException* error)
{
  if (needBackup && (openFlags & (CFile::modeReadWrite | CFile::modeWrite))) {
    String  backupFileName(fileName);

    // Change the extension to .GM~:
    //   Look for a period after the last backslash (a period before
    //   the last backslash would be part of a directory name).
    StrIdx  dot = backupFileName.find_last_of(_T('.'));
    if ((dot==String::npos) || (dot < backupFileName.find_last_of(_T('\\'))))
      dot = backupFileName.size(); // No extension
    backupFileName.replace(dot, String::npos, _T(".gm~"));

    if (_tcsicmp(fileName, backupFileName.c_str()) == 0)
      needBackup = false;       // This file already has a .GM~ extension
    else {
      ::DeleteFile(backupFileName.c_str());
      if (::MoveFile(fileName, backupFileName.c_str()))
        needBackup = false;      // We made a backup file
    } // end else try to make a backup file
  } // end if we need to make a backup file

  // The rest is copied from CDocument::GetFile:
  CMirrorFile*  file = new CMirrorFile;
  ASSERT(file != NULL);
  if (!file->Open(fileName, openFlags, error)) {
    delete file;
    file = NULL;
  }
  return file;
} // end CMapDoc::GetFile

//--------------------------------------------------------------------
// Read or write the document:
//
// Exceptions:
//   badClass:   Not a GUEmap document
//   badIndex:   File corrupt
//   badSchema:  Invalid version

void CMapDoc::Serialize(CArchive& ar)
{
  if (ar.IsStoring()) {
    const WORD edgeCount = edge.size();
    const WORD roomCount = room.size();

    ar.Write(fileHeader,sizeof(fileHeader));

    CSize  gridSize;
    getGridSize(gridSize);

    // Write minimum version number:
    if ((roomCount > 255) || (gridSize.cx > 42) || (gridSize.cy > 53) ||
        (page.size() > 1))
      ar << WORD(4);
    else
      ar << WORD(3);

    //................................................................
    // Write the map information record:
    ar << WORD(4)               // Version 4
       << fcStartRecord << recMap;
    if (!name.empty())
      ar << fcMapName << name;
    if (!note.empty())
      ar << fcMapNote << note;
    ar << fcMapNumRooms << roomCount
       << fcMapNumEdges << edgeCount
       << fcMapNumPages << BYTE(page.size())
       << fcMapSize     << WORD(gridSize.cx) << WORD(gridSize.cy)
       << fcEndRecord   << recMap;

    //................................................................
    // Write the rooms:
    for (RoomConstItr r = room.getVector().begin();
         r != room.getVector().end(); ++r) {
      ar << fcStartRecord << recRoom
         << fcRoomPosition
         << WORD((*r)->pos.x / gridX)
         << WORD((*r)->pos.y /-gridY);
      if (!(*r)->name.empty())
        ar << fcRoomName << (*r)->name;
      if (!(*r)->note.empty())
        ar << fcRoomNote << (*r)->note;
      if ((*r)->flags != rfBorder)
        ar << fcRoomFlags << (*r)->flags;
      ar << fcEndRecord   << recRoom;
    } // end for rooms

    //................................................................
    // Write the edges:
    for (EdgeConstItr e = edge.begin(); e != edge.end(); ++e) {
      ar << fcStartRecord << recEdge;
      if (e->room1 > 0xFF)
        ar << fcEdgeEnd1L << BYTE(e->end1) << e->room1;
      else
        ar << fcEdgeEnd1 << BYTE(e->end1) << BYTE(e->room1);
      if (e->type1 || e->type2)
        ar << fcEdgeType << e->type1 << e->type2;
      if (!(e->type1 & etNoRoom2)) {
        if (e->room2 > 0xFF)
          ar << fcEdgeEnd2L << BYTE(e->end2) << e->room2;
        else
          ar << fcEdgeEnd2 << BYTE(e->end2) << BYTE(e->room2);
      } // end if there is a room 2
      ar << fcEndRecord << recEdge;
    } // end for edges

    //................................................................
    // Write the pages:
    for (PageConstItr p = page.begin(); p != page.end(); ++p) {
      ar << fcStartRecord << recPage
         << fcPagePosition
         << WORD(p->pos.x / gridX)
         << WORD(p->pos.y /-gridY)
         << fcEndRecord << recPage;
    } // end for pages
  } // end if saving file
  //::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
  else { // loading file
    BYTE  buf[sizeof(fileHeader)];
    if ((ar.Read(buf,sizeof(buf)) != sizeof(buf)) ||
        memcmp(fileHeader, buf, sizeof(buf))) {
      AfxThrowArchiveException(CArchiveException::badClass,
                               ar.GetFile()->GetFilePath());
    } // end if not a GUEmap file

    WORD  version;
    ar >> version;
    if ((version < 2) || (version > 4))
      AfxThrowArchiveException(CArchiveException::badSchema,
                               ar.GetFile()->GetFilePath());
    FileReader  read(ar);
    WORD  edgeCount, roomCount;
    BYTE  pageCount = 0;
    BYTE  b;
    CRect rect;

    ar >> version;            // Get version number

    //................................................................
    // Read map information:
    read.findRecord(recMap);
    while ((b = read.byte()) != fcEndRecord) {
      switch (b) {
       case fcMapName:     ar >> name;      break;
       case fcMapNote:     ar >> note;      break;
       case fcMapNumEdges: ar >> edgeCount; break;
       case fcMapNumPages: ar >> pageCount; break;
       case fcMapNumRooms: ar >> roomCount; break;
       default:  read.skipField(b);         break;
      } // end switch field type
    } // end while map information record
    if (read.byte() != recMap) {
     corruptFile:
      AfxThrowArchiveException(CArchiveException::badIndex,
                               ar.GetFile()->GetFilePath());
    }

    //................................................................
    // Read rooms:
    room.resize(roomCount);
    for (RoomItr r = room.getVector().begin(); r != room.getVector().end();
         ++r) {
      read.findRecord(recRoom);
      *r = new MapRoom();
      (**r).flags = rfBorder;
      while ((b = read.byte()) != fcEndRecord) {
        switch (b) {
         case fcRoomFlags: ar >> (**r).flags; break;
         case fcRoomName:  ar >> (**r).name;  break;
         case fcRoomNote:  ar >> (**r).note;  break;
         case fcRoomPosition:
          (**r).pos.x = read.word() *  gridX;
          (**r).pos.y = read.word() * -gridY;
          break;
         default: read.skipField(b); break;
        } // end switch field type
      } // end while room record
      (**r).getRect(rect);
      if ((read.byte() != recRoom) || (rect.right > docSize.cx) ||
          (rect.bottom > 0) || (-rect.top > docSize.cy))
        goto corruptFile;
      (**r).computeNameLength();
    } // end for rooms

    //................................................................
    // Read edges:
    edge.resize(edgeCount);
    for (EdgeItr e = edge.begin(); e != edge.end(); ++e) {
      read.findRecord(recEdge);
      memset(e, 0, sizeof(*e));
      while ((b = read.byte()) != fcEndRecord) {
        switch (b) {
         case fcEdgeEnd1:
          e->end1  = RoomCorner(read.byte());
          e->room1 = read.byte();
          break;
         case fcEdgeEnd1L:
          e->end1  = RoomCorner(read.byte());
          e->room1 = read.word();
          break;
         case fcEdgeEnd2:
          e->end2  = RoomCorner(read.byte());
          e->room2 = read.byte();
          break;
         case fcEdgeEnd2L:
          e->end2  = RoomCorner(read.byte());
          e->room2 = read.word();
          break;
         case fcEdgeType:
          ar >> e->type1 >> e->type2;
          break;
         default: read.skipField(b); break;
        } // end switch field type
      } // end while edge record
      if ((read.byte() != recEdge) ||
          (e->room1 >= roomCount) ||
          (e->end1 < 0) || (e->end1 >= rcNumCorners) ||
          ((e->type1 & etDirection) > etOut) ||
          (!(e->type1 & etNoRoom2)
           && ((e->room2 >= roomCount) || (e->room2 == e->room1) ||
               (e->end2 < 0) || (e->end2 >= rcNumCorners) ||
               ((e->type2 & etDirection) > etOut))))
        goto corruptFile;
    } // end for edges

    //................................................................
    // Read pages:
    if (pageCount) {
      page.resize(pageCount);
      for (PageItr p = page.begin(); p != page.end(); ++p) {
        read.findRecord(recPage);
        while ((b = read.byte()) != fcEndRecord) {
          switch (b) {
           case fcPagePosition:
            p->pos.x = read.word() *  gridX;
            p->pos.y = read.word() * -gridY;
            break;
           default: read.skipField(b); break;
          } // end switch field type
        } // end while page record
        getPageRect(*p, rect);
        if ((read.byte() != recPage) || (rect.right > docSize.cx) ||
            (-rect.top > docSize.cy))
          goto corruptFile;
      } // end for pages
    } // end if pageCount
    needBackup = true;
  } // end else loading file
} // end CMapDoc::Serialize

//--------------------------------------------------------------------
// Report errors:

void CMapDoc::ReportSaveLoadException(LPCTSTR lpszPathName, CException* e,
                                      BOOL bSaving, UINT nIDPDefault)
{
  if (e && e->IsKindOf(RUNTIME_CLASS(CArchiveException))) {
    UINT msg = 0;

    switch (static_cast<CArchiveException*>(e)->m_cause) {
     case CArchiveException::badClass:  msg = IDS_AR_NOT_GUEMAP;  break;
     case CArchiveException::badSchema: msg = IDS_AR_BAD_VERSION; break;

     case CArchiveException::badIndex:
     case CArchiveException::endOfFile:
      msg = IDS_AR_CORRUPT;
      break;
    } // end switch

    if (msg) {
      AfxMessageBox(msg, MB_ICONEXCLAMATION);
      return;
    }
  } // end if CArchiveException

  CDocument::ReportSaveLoadException(lpszPathName, e, bSaving, nIDPDefault);
} // end CMapDoc::ReportSaveLoadException

/////////////////////////////////////////////////////////////////////////////
// CMapDoc diagnostics

#ifdef _DEBUG
void CMapDoc::AssertValid() const
{
  CDocument::AssertValid();
}

void CMapDoc::Dump(CDumpContext& dc) const
{
  CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// My Functions:
//--------------------------------------------------------------------
// Add a corner to an edge:
//
// Input:
//   rNum:  The room to add a corner by
//   eNum:  The edge to add a corner in
//
// Returns:
//   The number of the newly created room
//   -1 if we couldn't add a corner

static const short cornerXoffset[] =
{
  gridX, gridX, roomWidth, -gridX,
  roomWidth, -gridX, roomWidth, -gridX,
  2*gridX, 0, 2*gridX, 0
}; // end cornerXoffset

static const short cornerYoffset[] =
{
  roomHeight, -gridY, gridY, gridY,
  roomHeight, roomHeight, -gridY, -gridY,
  roomHeight, roomHeight, -gridY, -gridY
}; // end cornerYoffset

int CMapDoc::addCorner(RoomNum rNum, int eNum)
{
  if (room.size() == maxRooms) return -1;

  MapEdge  e1(edge[eNum]);
  MapEdge  e2(e1);

  ASSERT(!(e1.type1&etNoRoom2));

  RoomCorner  corner = ((e1.room1 == rNum) ? e1.end1 : e1.end2);

  e1.type1 &= ~etOneWay;             // Clear one-way
  e1.type2 = etNormal;
  e2.type1 &= etObstructed|etOneWay; // Clear all but obstructed & one-way
  e1.end2  = e2.end1  = rcCorner;
  e1.room2 = e2.room1 = room.size();

  CPoint pos(room[rNum].pos);
  pos.x += cornerXoffset[corner];
  pos.y += cornerYoffset[corner];

  if (pos.x < 0 || pos.x+gridX > docSize.cx ||
      pos.y+gridY > 0 || -pos.y > docSize.cy)
    return -1;

  MapRoom*  newRoom = new MapRoom();
  newRoom->flags = rfCorner;
  newRoom->pos = pos;
  newRoom->computeNameLength();

  setUndoData(new UndoChanges(IsModified(), e1.room2, 2, edge[eNum]));
  deleteEdge(eNum);
  addRoom(e1.room2, newRoom);
  addEdge(e1);
  addEdge(e2);

  return e1.room2;
} // end CMapDoc::addCorner

//--------------------------------------------------------------------
// Remove a corner from a connection:
//
// Input:
//   r:  The room number of the corner to be removed

void CMapDoc::deleteCorner(RoomNum r)
{
  ASSERT(r >= 0 && r < room.size() && room[r].isCorner());

  MapEdge  edges[2];
  EdgeVec  delEdges;
  int      count = 0;

  memset(edges, 0, sizeof(edges));

  delEdges.reserve(2);

  for (EdgeConstItr e = edge.begin(); (count < 2) && (e != edge.end()); ++e) {
    if (e->room1 == r) {
      edges[count].room1 = e->room2;
      edges[count].end1  = e->end2;
      edges[count].type1  |= e->type2;
      edges[!count].type1 |= e->type1 & etSpecial;
      delEdges.push_back(*e);
      ++count;
    } else if (e->room2 == r) {
      edges[count].room1 = e->room1;
      edges[count].end1  = e->end1;
      edges[count].type1|= e->type1;
      delEdges.push_back(*e);
      ++count;
    }
  }

  if (count < 2) {
    MessageBeep(MB_ICONASTERISK); // Corner without 2 edges
    deleteRoom(r);
    setUndoData(NULL);
    return;
  }

  int dest   = 0;
  int source = 1;

  if (edges[1].type1 & etSpecial) {
    dest = 1;  source = 0;
  }

  edges[dest].end2  = edges[source].end1;
  edges[dest].room2 = edges[source].room1;
  edges[dest].type2 = edges[source].type1 & ~etSpecial;

  if (edges[dest].room1 > r) --edges[dest].room1;
  if (edges[dest].room2 > r) --edges[dest].room2;

  const bool isMod = IsModified();
  setUndoData(new UndoChanges(isMod, r, extractRoom(r), delEdges));
  addEdge(edges[dest]);
} // end CMapDoc::deleteCorner

//--------------------------------------------------------------------
void CMapDoc::addEdge(const MapEdge& newEdge)
{
  edge.push_back(newEdge);
  SetModifiedFlag();
  CRect  rect;
  getEdgeRect(newEdge, rect);
  UpdateAllViews(NULL,0,reinterpret_cast<CObject*>(&rect));
} // end CMapDoc::addEdge

void CMapDoc::deleteEdge(int n)
{
  ASSERT(n >= 0 && n < edge.size());
  SetModifiedFlag();
  CRect  rect;
  getEdgeRect(edge[n], rect);

  edge.erase(edge.begin() + n);

  UpdateAllViews(NULL,0,reinterpret_cast<CObject*>(&rect));
} // end CMapDoc::deleteEdge

//--------------------------------------------------------------------
// Find the edge (if any) connected to a given corner:
//
// Input:
//   n:       The room number to look for
//   corner:  The corner of room N we want
//
// Output:
//   newEdge.room1:  The room at the other end of this edge
//   newEdge.end1:   The corner of the other room
//   newEdge.type1:  The type of connection at the other end
//   newEdge.type2:  The type of connection at this end
//     etObstructed is moved from type2 to type1 if necessary
//
// Returns:
//   The edge number
//   -1 if no edge is attached to the specified corner

static const EdgeType  corner2et[] = { etUp, etDown, etIn, etOut };

int CMapDoc::findEdge(RoomNum n, RoomCorner corner, MapEdge& newEdge) const
{
  EdgeConstItr  e = edge.begin();
  int i;

  if (corner < rcNumCorners) {
    for (i = edge.size()-1; i >= 0; i--) {
      if (e[i].room1 == n && e[i].end1 == corner)
        goto found1;
      if (e[i].room2==n && e[i].end2==corner && !(e[i].type1 & etNoRoom2))
        goto found2;
    } // end for edges
  } // end if corner
  else { // direction (up/down/in/out)
    const EdgeType  dir = corner2et[corner - rcNumCorners];
    for (i = edge.size()-1; i >= 0; i--) {
      if (e[i].room1 == n && (e[i].type1 & etDirection) == dir)
        goto found1;
      if (e[i].room2 == n && (e[i].type2 & etDirection) == dir &&
          !(e[i].type1 & etNoRoom2))
        goto found2;
    } // end for edges
  } // end else direction (up/down/in/out)

  return -1;

 found1: // The room is at the start of the edge
  newEdge.room1 = e[i].room2;
  newEdge.type1 = e[i].type2 | (e[i].type1 & etObstructed);
  newEdge.type2 = e[i].type1 & ~etObstructed;
  newEdge.end1  = e[i].end2;
  return i;

 found2: // The room is at the end of the edge
  newEdge.room1 = e[i].room1;
  newEdge.type1 = e[i].type1;
  newEdge.type2 = e[i].type2;
  newEdge.end1  = e[i].end1;
  return i;
} // end CMapDoc::findEdge

//--------------------------------------------------------------------
// Find the other end of a connection with corners:
//
// Input:
//   start:
//     The edge we want to follow
//
// Output:
//   cornerRooms:  (if not NULL)
//     Contains the room numbers of all corners in this connection.
//     If the vector was not empty, the new entries are appended.
//
// Returns:
//   The number of the room at the other end
//   -1 if the connection doesn't lead anywhere

int CMapDoc::findOtherEnd(EdgeConstItr start, RoomNumVec* cornerRooms) const
{
  EdgeConstItr  oldEdge = start;
  RoomNum room = ((start->end1 == rcCorner) ? start->room1 : start->room2);

  if (cornerRooms) cornerRooms->push_back(room);

  for (;;) {
    EdgeConstItr  lastTime = oldEdge;
    for (EdgeConstItr e = edge.begin(); e < edge.end(); e++) {
      if (e != oldEdge) {
        if (e->room1 == room) {
          room = e->room2;
          if (e->end2 != rcCorner) return room;
          if (cornerRooms) cornerRooms->push_back(room);
          oldEdge = e;
        } else if (e->room2 == room) {
          room = e->room1;
          if (e->end1 != rcCorner) return room;
          if (cornerRooms) cornerRooms->push_back(room);
          oldEdge = e;
        }
      } // end if not the edge we just came in on
    } // end for each edge
    if (lastTime == oldEdge)
      return -1;                // We didn't go anywhere
  } // end forever
} // end CMapDoc::findOtherEnd

//--------------------------------------------------------------------
static const int  cornerXOffset[] = {
  roomWidth/2, roomWidth/2, roomWidth, 0,
  roomWidth, 0, roomWidth, 0,
  3*roomWidth/4, roomWidth/4, 3*roomWidth/4, roomWidth/4,
  gridX/2
}; // end cornerXOffset

static const int  cornerYOffset[] = {
  roomHeight, 0, roomHeight/2, roomHeight/2,
  roomHeight, roomHeight, 0, 0,
  roomHeight, roomHeight, 0, 0,
  gridY/2
}; // end cornerYOffset

static const int  nowhereXOffset[] = {
  0, 0, gridX/2, -gridX/2,              // N,S,E,W
  gridX/2, -gridX/2, gridX/2, -gridX/2, // NE,NW,SE,SW
  0, 0, 0, 0                            // NNE, NNW, SSE, SSW
}; // end nowhereXOffset

static const int  nowhereYOffset[] = {
  gridY/2, -gridY/2, 0, 0,              // N,S,E,W
  gridY/2, gridY/2, -gridY/2, -gridY/2, // NE,NW,SE,SW
  gridY/2, gridY/2, -gridY/2, -gridY/2, // NNE, NNW, SSE, SSW
}; // end nowhereYOffset

void CMapDoc::getEdgePoints(const MapEdge& e, LONG& x1, LONG& y1,
                            LONG& x2, LONG& y2) const
{
  x1 = room[e.room1].pos.x + cornerXOffset[e.end1];
  y1 = room[e.room1].pos.y + cornerYOffset[e.end1];
  if (e.type1 & etNoRoom2) {
    x2 = x1 + nowhereXOffset[e.end1];
    y2 = y1 + nowhereYOffset[e.end1];
  } else {
    x2 = room[e.room2].pos.x + cornerXOffset[e.end2];
    y2 = room[e.room2].pos.y + cornerYOffset[e.end2];
  } // end else normal edge
} // end CMapDoc::getEdgePoints

void CMapDoc::getEdgeRect(const MapEdge& e, CRect& r) const
{
  getEdgePoints(e, r.left, r.top, r.right, r.bottom);
  r.NormalizeRect();
  r.InflateRect(gridX,gridY);       // PORT
} // end CMapDoc::getEdgeRect

//--------------------------------------------------------------------
// Compute the actual size of the map:
//
// Output:
//   r:  The rectangle enclosing the map in logical coordinates

void CMapDoc::getGridRect(CRect& rect) const
{
  CRect  roomRect;

  rect.SetRectEmpty();
  for (RoomConstItr  r = room.getVector().begin();
       r != room.getVector().end();
       ++r)
    rect |= (*r)->getRect(roomRect);

  for (PageConstItr  p = page.begin(); p != page.end(); ++p) {
    getPageRect(*p, roomRect);
    rect |= roomRect;
  }
} // end CMapDoc::getGridRect

//--------------------------------------------------------------------
// Compute the actual size of the map:
//
// Output:
//   size: The size of the map in grid units
//     Does not consider stubs or labels that may extend off the edge

void CMapDoc::getGridSize(CSize& size) const
{
  CRect  rect;
  getGridRect(rect);

  size.cx = rect.right / gridX;
  size.cy = rect.top / -gridY;
} // end CMapDoc::getGridSize

//--------------------------------------------------------------------
// Add a page to the map:
//
// Input:
//   n:        The new page number (0 based)
//   newPage:  The page information

void CMapDoc::addPage(int n, const MapPage& newPage)
{
  ASSERT(n >= 0 && n <= page.size());
  if (page.size() == maxRooms)
    return;

  page.insert(page.begin() + n, newPage);
  SetModifiedFlag();
  UpdateAllViews(NULL, dupPageCount, NULL);
} // end CMapDoc::addPage

//--------------------------------------------------------------------
// Append pages to the map:
//
// Input:
//   newPages:  The page information

void CMapDoc::addPages(const PageVec& newPages)
{
  if (page.size() + newPages.size() > maxRooms)
    return;

  page.insert(page.end(), newPages.begin(), newPages.end());
  SetModifiedFlag();
  UpdateAllViews(NULL, dupPageCount, NULL);
} // end CMapDoc::addPages

//--------------------------------------------------------------------
// Delete a page from the map:
//
// Input:
//   n:  The (0 based) page number to delete

void CMapDoc::deletePage(int n)
{
  ASSERT(n >= 0 && n < page.size());
  ASSERT(page.size() > 1);
  SetModifiedFlag();
  CRect  rect;
  getPageRect(n, rect);

  page.erase(page.begin() + n);

  UpdateAllViews(NULL, dupPageCount, NULL);
} // end CMapDoc::deletePage

//--------------------------------------------------------------------
// Place pages to cover the entire map:

void CMapDoc::layoutPages()
{
  UndoPaginate*  undoRec = new UndoPaginate(*this);

  CRect r;
  getGridRect(r); // Just counts rooms, since UndoPaginate cleared the pages

  const int
    width  = r.Width(),
    height = r.Height(),
    numX   = (width  + pageSize.cx - 1) / pageSize.cx,
    numY   = (height + pageSize.cy - 1) / pageSize.cy;

  if (!numX)
    page.insert(page.begin());  // One page at 0,0
  else {
    int i;
    float delta,step;
    PageItr  p, p2;
    page.resize(numX * numY);
    if (numX == 1) {
      i = r.left - (pageSize.cx - width)/2;
      i -= i % gridX;
      if (i < 0) i = 0;
      for (p = page.begin(); p != page.end(); ++p)
        p->pos.x = i;
    } else {
      step = float(width - pageSize.cx) / float(numX);
      delta = 0;
      for (p = page.begin(), p2 = p + numX; p != p2; ++p, delta += step) {
        p->pos.x = r.left + delta;
        p->pos.x -= p->pos.x % gridX;
      }
      p[-1].pos.x = r.right - pageSize.cx;
      for (p2 = page.begin(); p != page.end(); ++p, ++p2)
        p->pos.x = p2->pos.x;
    } // end else more than 1 page across
    if (numY == 1) {
      i = r.bottom + (pageSize.cy - height)/2;
      i -= i % gridY;
      if (i > 0) i = 0;
      for (p = page.begin(); p != page.end(); ++p)
        p->pos.y = i;
    } else {
      step = float(pageSize.cy - height) / float(numY);
      delta = 0;
      p = p2 = page.begin();
      while (p2 != page.end()) {
        p2 += numX;
        if (p2 == page.end())
          i = r.top + pageSize.cy;
        else {
          i = r.bottom + delta;
          i -= i % gridY;
          delta += step;
        }
        while (p != p2)
          (p++)->pos.y = i;
      } // end while more rows
    } // end else more than 1 page down
  } // end else at least 1 page

  if (undoRec->getPages() == page)
    delete undoRec;             // No changes
  else {
    setUndoData(undoRec);
    SetModifiedFlag();
    UpdateAllViews(NULL, dupPageCount, NULL);
  }
} // end CMapDoc::layoutPages

//--------------------------------------------------------------------
// Move a page:
//
// Input:
//   n:       The (0 based) page number to move
//   offset:  How much to move it
//
// Note:
//   Must not change undo data (may be used during undo)

void CMapDoc::movePage(int n, const CSize& offset)
{
  SetModifiedFlag();

  page[n].pos += offset;
  UpdateAllViews(NULL);
} // end CMapDoc::movePage

//--------------------------------------------------------------------
// Decide if we need to repaginate:
//
// Returns:
//   true:   Some rooms are not (completely) on any page
//   false:  All rooms are on a page

bool CMapDoc::needRepaginate() const
{
  const PageConstItr
    pBegin = page.begin(),
    pEnd   = page.end();
  CRect  rr, pr, i;

  for (RoomConstItr r = room.getVector().begin();
       r != room.getVector().end(); ++r) {
    (**r).getRect(rr);
    for (PageConstItr p = pBegin; p != pEnd; ++p) {
      getPageRect(*p, pr);
      if (i.IntersectRect(&pr, &rr) && (i == rr))
        goto nextRoom;
    } // end for pages
    return true;                // This room not on any page
   nextRoom:
    ;
  } // end for rooms

  return false;
} // end CMapDoc::needRepaginate

//--------------------------------------------------------------------
// Determine which page (if any) contains a given point:
//
// The point must be in the page border (one of the grid cells which
// displays the page number).
//
// Input:
//   pos:  The position to test
//
// Returns:
//   The (0 based) page number of the page containing the point
//   -1 if no page contains the point

int CMapDoc::pageBorderHit(const CPoint& pos) const
{
  CRect r;
  for (int i = page.size()-1; i >= 0; --i) {
    getPageRect(i, r);
    if (r.PtInRect(pos) &&
        (((pos.x - r.left < gridX) || (r.right  - pos.x < gridX)) !=
         ((pos.y - r.top  < gridY) || (r.bottom - pos.y < gridY))))
      // Point is in the X border or the Y border, but not both:
      return i;
  } // end for

  return -1;                    // Not in any page border
} // end CMapDoc::pageBorderHit

//--------------------------------------------------------------------
// Add a new room:
//
// Returns:
//   The room number of the new room (-1 if error)

int CMapDoc::addRoom(const CPoint& pos)
{
  if (room.size() == maxRooms)
    return -1;

  MapRoom*  newRoom = new MapRoom();
  newRoom->flags = rfBorder;
  newRoom->pos = pos;
  newRoom->computeNameLength();
  addRoom(room.size(), newRoom);

  return room.size() - 1;
} // end CMapDoc::addRoom

//--------------------------------------------------------------------
// Add a new room:
//
// Input:
//   n:        The position for the new room [0..size]
//   newRoom:  A MapRoom object allocated by new

void CMapDoc::addRoom(int n, MapRoom* newRoom)
{
  ASSERT(n >= 0 && n <= room.size());
  ASSERT(newRoom);

  room.insert(n, newRoom);
  CJM_MEMCHECK;

  if (n != room.size()-1)
    for (int i=edge.size()-1; i >= 0; i--) {
      if (edge[i].room1 >= n) ++edge[i].room1;
      if (edge[i].room2 >= n) ++edge[i].room2;
    }

  SetModifiedFlag();
  CRect  rect;
  newRoom->getRect(rect);
  UpdateAllViews(NULL,0,reinterpret_cast<CObject*>(&rect));
} // end CMapDoc::addRoom

//--------------------------------------------------------------------
// Extract a room from the map:
//
// Input:
//   n:  The room number to remove
//
// Returns:
//   A pointer to the room extracted from the map
//   You can use this in the undo record

MapRoom* CMapDoc::extractRoom(int n)
{
  ASSERT(n >= 0 && n < room.size());

  SetModifiedFlag();

  for (int i=edge.size()-1; i >= 0; i--) {
    if ((edge[i].room1 == n) ||
        (!(edge[i].type1 & etNoRoom2) && (edge[i].room2 == n)))
      deleteEdge(i);
    else {
      if (edge[i].room1 > n) --edge[i].room1;
      if (edge[i].room2 > n) --edge[i].room2;
    }
  }

  MapRoom*  oldRoom = room.extract(n);
  CRect     rect;
  oldRoom->getRect(rect);

  UpdateAllViews(NULL,dupDeletedRoom,reinterpret_cast<CObject*>(&rect));

  return oldRoom;
} // end CMapDoc::extractRoom

//--------------------------------------------------------------------
// Find out if a new room would intersect an old one:
//
// There must be at least one grid of space around the room.
// Ignores corners.
//
// Input:
//   pos:  The proposed location for a new room
//
// Returns:
//   The room number of an intersecting room
//   -1 if the space is unoccupied
//   -2 if the space is off the grid

int CMapDoc::findRoom(const CPoint& pos) const
{
  if (pos.x < 0 || pos.x+roomWidth > docSize.cx ||
      pos.y+roomHeight > 0 || -pos.y > docSize.cy)
    return -2;

  RoomConstItr  rm = room.getVector().begin();
  CRect  newRoom(pos, defaultRoomSize);
  newRoom.InflateRect(gridX-1, gridY-1);
  CRect  intersect,temp;

  for (int i = room.size()-1; i >= 0; i--)
    if (!rm[i]->isCorner() &&
        intersect.IntersectRect(newRoom, rm[i]->getRect(temp)))
      return i;

  return -1;
} // end CMapDoc::findRoom

//--------------------------------------------------------------------
// Move a room:
//
// Input:
//   n:       The room number to move
//   offset:  How much to move it
//
// Note:
//   Must not change undo data (may be used during undo)

void CMapDoc::moveRoom(int n, const CSize& offset)
{
  SetModifiedFlag();

  room[n].pos += offset;
  UpdateAllViews(NULL);
} // end CMapDoc::moveRoom

//--------------------------------------------------------------------
// Determine which room (if any) contains a given point:
//
// Input:
//   pos:  The position to test
//
// Output:
//   corner:
//     Which corner of the room was hit (rcNone if in center)
//
// Returns:
//   The room number of the room containing the point
//   -1 if no room contains the point

int CMapDoc::roomHit(const CPoint& pos, RoomCorner* corner) const
{
  CRect r;
  for (int i = room.size()-1; i >= 0; i--) {
    if (room[i].getRect(r).PtInRect(pos)) {
      if (corner) {
        if (room[i].isCorner())
          *corner = rcNone;
        else {
          // FIXME this assumes all rooms are the same size
          LONG x = pos.x - r.left;
          LONG y = pos.y - r.top; // Actually, this is the bottom
          if (y < roomHeight/4) {
            if (x < roomWidth*3/16)
              *corner = rcSW;
            else if (x < roomWidth*5/16)
              *corner = rcSSW;
            else if (x > roomWidth*13/16)
              *corner = rcSE;
            else if (x > roomWidth*11/16)
              *corner = rcSSE;
            else
              *corner = rcS;
          } else if (y > roomHeight*3/4) {
            if (x < roomWidth*3/16)
              *corner = rcNW;
            else if (x < roomWidth*5/16)
              *corner = rcNNW;
            else if (x > roomWidth*13/16)
              *corner = rcNE;
            else if (x > roomWidth*11/16)
              *corner = rcNNE;
            else
              *corner = rcN;
          } else {
            if (x < roomWidth/4)
              *corner = rcW;
            else if (x > roomWidth*3/4)
              *corner = rcE;
            else
              *corner = rcNone;
          }
        } // end else not a corner room
      } // end if corner
      return i;
    } // end if point in current room
  } // end for

  if (corner) *corner = rcNone;
  return -1;                    // Not in any room
} // end CMapDoc::roomHit

//--------------------------------------------------------------------
// Change the map's title:
//
// Input:
//   newName:  The new map title

void CMapDoc::setName(LPCTSTR newName)
{
  if (name == newName) return;  // Nothing to do
  name = newName;
  SetModifiedFlag();
} // end CMapDoc::setName

//--------------------------------------------------------------------
// Change the map's comments:
//
// Input:
//   newNote:  The new map comments

void CMapDoc::setNote(LPCTSTR newNote)
{
  if (note == newNote) return;  // Nothing to do
  note = newNote;
  trimRight(note);
  SetModifiedFlag();
} // end CMapDoc::setNote

//--------------------------------------------------------------------
// Change a room's flags:
//
// Input:
//   n:        The room number to change
//   set:      The flags to set
//   clear:    The flags to clear (default 0)
//
// Note:
//   flags in both SET and CLEAR will be set.

void CMapDoc::setRoomFlags(int n, RoomFlags set, RoomFlags clear)
{
  MapRoom& rm = room[n];
  RoomFlags  old = rm.flags;

  rm.flags &= ~clear;
  rm.flags |= set;

  if (rm.flags != old) {
    SetModifiedFlag();

    CRect  rect;
    rm.getRect(rect).InflateRect(5,5); // PORT
    UpdateAllViews(NULL,0,reinterpret_cast<CObject*>(&rect));
  }
} // end CMapDoc::setRoomFlags

//--------------------------------------------------------------------
// Change a room's name:
//
// Input:
//   n:        The room number to change
//   newName:  The new room name

void CMapDoc::setRoomName(int n, LPCTSTR newName)
{
  MapRoom& rm = room[n];
  if (rm.name == newName) return; // Nothing to do
  rm.name = newName;
  rm.computeNameLength();
  SetModifiedFlag();

  CRect  rect;
  rm.getRect(rect).DeflateRect(5,5); // PORT
  UpdateAllViews(NULL,0,reinterpret_cast<CObject*>(&rect));
} // end CMapDoc::setRoomName

//--------------------------------------------------------------------
// Change a room's comments:
//
// Input:
//   n:        The room number to change
//   newNote:  The new room note

void CMapDoc::setRoomNote(int n, LPCTSTR newNote)
{
  MapRoom& rm = room[n];
  if (rm.note == newNote) return; // Nothing to do
  const bool  redraw = (rm.note.empty() || !*newNote);
  rm.note = newNote;
  trimRight(rm.note);
  SetModifiedFlag();

  if (redraw) {
    CRect  rect;
    rm.getRect(rect).DeflateRect(5,5); // PORT
    UpdateAllViews(NULL,dupRoomComment,reinterpret_cast<CObject*>(&rect));
  } else
    UpdateAllViews(NULL,dupRoomComment,NULL);
} // end CMapDoc::setRoomNote

//////////////////////////////////////////////////////////////////////
// Undo handling:
//--------------------------------------------------------------------
// Set undo data:

void CMapDoc::setUndoData(UndoRec* newUndo)
{
  if (undoData != newUndo) {
    delete undoData;
    undoData = newUndo;
  }
  CJM_MEMCHECK;
} // end TClassViewer::setUndoData

//--------------------------------------------------------------------
// Undo the last operation:

void CMapDoc::OnEditUndo()
{
  CJM_MEMCHECK;
  if (undoData)
    setUndoData(undoData->undo(*this));
} // end CMapDoc::OnEditUndo

//--------------------------------------------------------------------
// Update the Undo menu item:

void CMapDoc::OnUpdateEditUndo(CCmdUI* pCmdUI)
{
  pCmdUI->Enable(undoData && !locked);

  CString  text;
  if (undoData) {
    text  = _T("&Undo ");
    text += undoData->getName();
  } else
    text = _T("Can't &Undo");

  text += _T("\tCtrl+Z");
  pCmdUI->SetText(text);
} // end CMapDoc::OnUpdateEditUndo

/////////////////////////////////////////////////////////////////////////////
// CMapDoc commands

// Erase contents of document:
void CMapDoc::DeleteContents()
{
  locked = false;
  edge.erase(edge.begin(), edge.end());
  page.resize(1);
  page.front().pos.x = 0;
  page.front().pos.y = 0;
  room.removeAll();
  name.erase();
  note.erase();
  docSize = CSize(271 * gridX, 451 * gridY); // PORT
  setUndoData(NULL);
  needBackup = true;
} // end CMapDoc::DeleteContents

//--------------------------------------------------------------------
// Move all the rooms around:
//
// This moves only rooms that have connections.  Rooms that stand
// alone (used as a title box, for example) are not moved.

void CMapDoc::OnEditMoveMap(UINT cmd)
{
  const int  numEdges = edge.size();
  const int  numRooms = room.size();

  ByteVec  selected;
  selected.resize(numRooms, false);

  int i;
  // Select all rooms that have a connection to somewhere:
  for (i=0; i < numEdges; i++) {
    selected[edge[i].room1] = true;
    if ((edge[i].type1 & etNoRoom2) == 0)
      selected[edge[i].room2] = true;
  }

  // Compute the bounding rectangle of the selected rooms:
  int    numSelected = 0;
  CRect  r,temp;

  r.SetRectEmpty();
  for (i=0; i < numRooms; i++)
    if (selected[i]) {
      ++numSelected;
      r |= room[i].getRect(temp);
    }

  // Calculate the distance to move them:
  CSize offset(0,0);
  switch (cmd) {
   case ID_EDIT_MOVE_CENTER:
    offset.cx = (docSize.cx - r.right - r.left) / 2;
    offset.cy = (docSize.cy + r.top + r.bottom) /-2;
    break;
   case ID_EDIT_MOVE_TOP:    offset.cy = -r.bottom;             break;
   case ID_EDIT_MOVE_BOTTOM: offset.cy = -(docSize.cy + r.top); break;
   case ID_EDIT_MOVE_LEFT:   offset.cx = -r.left;               break;
   case ID_EDIT_MOVE_RIGHT:  offset.cx = docSize.cx - r.right;  break;
#ifdef _DEBUG
   default: ASSERT(0);          // Can't get here
#endif
  } // end switch

  offset.cx -= offset.cx % gridX; // Make sure the rooms stay on the grid
  offset.cy -= offset.cy % gridY;

  // Do the move:

  if (!offset.cx && !offset.cy)
    return;                     // Do nothing

  setUndoData(new UndoMove(IsModified(), offset, numSelected, selected,
                           0, selected)); // No selected pages

  for (i=0; i < numRooms; i++)
    if (selected[i]) moveRoom(i,offset);
} // end CMapDoc::OnEditMoveMap

void CMapDoc::OnNavigationMode()
{
  locked = !locked;
  if (locked)
    if (getRoomCount() > 0)
      UpdateAllViews(NULL, dupNavigationMode);
    else
      locked = false;           // Can't enter navigation mode
} // end CMapDoc::OnNavigationMode

void CMapDoc::OnUpdateEditMoveMap(CCmdUI* pCmdUI)
{
  // You can't use Move Map unless you have edges:
  pCmdUI->Enable(!locked && edge.size());
} // end CMapDoc::OnUpdateEditMoveMap

void CMapDoc::OnUpdateNavigationMode(CCmdUI* pCmdUI)
{
  pCmdUI->SetCheck(locked);
  if (!locked)
    pCmdUI->Enable(getRoomCount() > 0);
} // end CMapDoc::OnUpdateNavigationMode

//////////////////////////////////////////////////////////////////////
// Local Variables:
// compile-command: "nmake -f GUEmap.mak"
// End:
