//--------------------------------------------------------------------
// $Id: MapDoc.hpp 4718 2007-09-15 02:34:13Z cjm $
//--------------------------------------------------------------------
//
//   GUEmap
//   Copyright 1996-2007 by Christopher J. Madsen
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License along
//   with this program; if not, write to the Free Software Foundation, Inc.,
//   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//   MapDoc.h : interface of the CMapDoc class
//
//--------------------------------------------------------------------

const int
  gridX      = 120,             // PORT 0.25 in
  gridY      =  72,             // PORT 0.15 in
  maxNameLines = 3,
  maxPages   = 250,
  maxRooms   = 65535,
  roomWidth  = 3 * gridX,       // PORT
  roomHeight = 3 * gridY;       // PORT

const LPARAM
  dupDeletedRoom    = 1,
  dupNavigationMode = 2,
  dupRoomComment    = 3,
  dupPageCount      = 4;

typedef WORD          RoomNum;
typedef unsigned char EdgeType;
typedef unsigned char RoomFlags;

const EdgeType
  etNormal     = 0x00,
  etUp         = 0x01,
  etDown       = 0x02,
  etIn         = 0x03,
  etOut        = 0x04,
  etDirection  = 0x07,
  etObstructed = 0x08,
  etNoExit     = 0x10,
  etUnexplored = 0x20,
  etOneWay     = 0x40,
  etNoRoom2    = etNoExit|etUnexplored,
  etSpecial    = etObstructed|etNoExit|etOneWay|etUnexplored;

const RoomFlags
  rfBorder = 0x01,
  rfCorner = 0x02;

enum RoomCorner
{
  rcNone = -1,
  rcN, rcS, rcE, rcW,
  rcNE, rcNW, rcSE, rcSW,
  rcNNE, rcNNW, rcSSE, rcSSW,
  rcCorner,
  rcNumCorners,
  rcUp = rcNumCorners, rcDown, rcIn, rcOut,
  rcNumDirections
}; // end RoomCorner

extern const RoomCorner  oppositeCorner[rcNumCorners];

const CSize  defaultRoomSize(roomWidth,roomHeight);

class MapEdge
{
 public:
  RoomNum    room1;
  RoomNum    room2;
  EdgeType   type1;
  EdgeType   type2;
  RoomCorner end1;
  RoomCorner end2;
}; // end MapEdge

class MapPage
{
 public:
  CPoint  pos;

 public:
  MapPage() : pos(0,0) {};
  bool  operator==(const MapPage& o) const { return pos == o.pos; };
}; // end MapPage

class MapRoom
{
 public:
  CPoint  pos;
  String  name;
  String  note;
  RoomFlags flags;
  BYTE    lines;
  BYTE    nameLength[maxNameLines];
 public:
  MapRoom();
  MapRoom(const MapRoom& o) { *this = o; };
  MapRoom&  operator=(const MapRoom& source);
  void  computeNameLength();
  CRect&  getRect(CRect& r) const {
    r.right  = (r.left = pos.x) + ((flags & rfCorner) ? gridX : roomWidth);
    r.bottom = (r.top = pos.y)  + ((flags & rfCorner) ? gridY : roomHeight);
    return r;
  }; // end MapRoom::getRect
  RoomFlags  isCorner() const { return (flags & rfCorner); };
}; // end MapRoom

class UndoRec;
class UndoPaginate;

typedef vector<MapEdge> EdgeVec;
typedef EdgeVec::iterator       EdgeItr;
typedef EdgeVec::const_iterator EdgeConstItr;

typedef vector<MapPage> PageVec;
typedef PageVec::iterator       PageItr;
typedef PageVec::const_iterator PageConstItr;

typedef vector<MapRoom*> RoomVec;
typedef RoomVec::iterator       RoomItr;
typedef RoomVec::const_iterator RoomConstItr;
typedef Array<MapRoom, RoomVec> RoomArray;

typedef ByteVec PageNumVec;
typedef ByteVec::iterator       PNItr;
typedef ByteVec::const_iterator PNConstItr;

typedef WordVec RoomNumVec;
typedef WordVec::iterator       RNItr;
typedef WordVec::const_iterator RNConstItr;

class CMapDoc : public CDocument
{
  friend class UndoPaginate;

 protected: // create from serialization only
  CMapDoc();
  DECLARE_DYNCREATE(CMapDoc);

  // Attributes
 protected:
  RoomArray room;
  EdgeVec   edge;
  PageVec   page;
  String    name;
  String    note;
  CSize     docSize;
  CSize     pageSize;
  UndoRec*  undoData;
  bool      needBackup;

 public:
  bool  locked;

 public:
  const CSize&  getDocSize()    const { return docSize;  };
  int    getEdgeCount() const { return edge.size(); };
  void   getGridRect(CRect& rect) const;
  void   getGridSize(CSize& size) const;
  int    getRoomCount() const { return room.size(); };
  const MapEdge&  getEdge(int i) const { return edge[i]; };
  const EdgeVec&  getEdges()     const { return edge;    };
  const String&   getName()      const { return name;    };
  const String&   getNote()      const { return note;    };
  const MapPage&  getPage(int i) const { return page[i]; };
  const PageVec&  getPages()     const { return page;    };
  const MapRoom&  getRoom(int i) const { return room[i]; };
  const RoomVec&  getRooms()     const { return room.getVector(); };
  void            setUndoData(UndoRec* newUndo);

  // Overrides
 public:
  virtual CFile* GetFile(LPCTSTR fileName, UINT openFlags,
                         CFileException* error);

 protected:
  virtual BOOL   DoSave(LPCTSTR pathName, BOOL replace = TRUE);

  // ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CMapDoc)
	public:
	virtual BOOL OnNewDocument();
	virtual void DeleteContents();
	virtual void ReportSaveLoadException(LPCTSTR lpszPathName, CException* e, BOOL bSaving, UINT nIDPDefault);
	virtual void Serialize(CArchive& ar);
	//}}AFX_VIRTUAL

  // Operations:
 public:

  int   addCorner(RoomNum r, int e);
  void  addEdge(const MapEdge& newEdge);
  void  addEdges(int n) { edge.reserve(edge.size() + n); };
  void  deleteCorner(RoomNum r);
  void  deleteEdge(int n);
  int   findEdge(RoomNum n, RoomCorner corner, MapEdge& newEdge) const;
  int   findOtherEnd(EdgeConstItr start, RoomNumVec* cornerRooms=NULL) const;
  int   findOtherEnd(int n) const { return findOtherEnd(edge.begin() + n); }
  void  getEdgePoints(const MapEdge& e, LONG& x1, LONG& y1,
                      LONG& x2, LONG& y2) const;
  void  getEdgePoints(int n, LONG& x1, LONG& y1, LONG& x2, LONG& y2) const
  {  getEdgePoints(edge[n],x1,y1,x2,y2); };
  void  getEdgeRect(const MapEdge& e, CRect& r) const;

  void  addPage(int n, const MapPage& newPage);
  void  addPages(int n) { page.reserve(page.size() + n); };
  void  addPages(const PageVec& newPages);
  void  deletePage(int n);
  void  getPageRect(const MapPage& p, CRect& r) const
    {
      r.right = (r.left   = p.pos.x) + pageSize.cx;
      r.top   = (r.bottom = p.pos.y) - pageSize.cy;
    };
  void  getPageRect(int p, CRect& r) const
    { ASSERT(p>=0 && p<page.size());  getPageRect(page[p], r); };
  void  layoutPages();
  void  movePage(int n, const CSize& offset);
  bool  needRepaginate() const;
  int   pageBorderHit(const CPoint& pos) const;

  int   addRoom(const CPoint& pos);
  void  addRoom(int n, MapRoom* newRoom);
  void  addRooms(int n) { room.reserve(room.size() + n); };
  MapRoom*  extractRoom(int n);
  void  deleteRoom(int n) { delete extractRoom(n); };
  int   findRoom(const CPoint& pos) const;
  void  moveRoom(int n, const CSize& offset);
  int   roomHit(const CPoint& pos, RoomCorner* corner=NULL) const;
  void  setName(LPCTSTR newName);
  void  setNote(LPCTSTR newNote);
  void  setRoomFlags(int n, RoomFlags set, RoomFlags clear=0);
  void  setRoomName(int n, LPCTSTR newName);
  void  setRoomNote(int n, LPCTSTR newNote);
  void  shortestPath(String& path, RoomNum start, RoomNum end) const;

  // Implementation
 public:
  virtual ~CMapDoc();
#ifdef _DEBUG
  virtual void AssertValid() const;
  virtual void Dump(CDumpContext& dc) const;
#endif

  afx_msg void  OnNavigationMode();

  // Generated message map functions
 protected:
  //{{AFX_MSG(CMapDoc)
	afx_msg void OnEditUndo();
	afx_msg void OnFileExport();
	afx_msg void OnUpdateEditUndo(CCmdUI* pCmdUI);
	afx_msg void OnUpdateNavigationMode(CCmdUI* pCmdUI);
	//}}AFX_MSG
  afx_msg void OnEditMoveMap(UINT cmd);
  afx_msg void OnUpdateEditMoveMap(CCmdUI* pCmdUI);
  DECLARE_MESSAGE_MAP();
}; // end CMapDoc

class RoomScrap
{
 protected:
  EdgeVec    edges;
  PageVec    pages;
  RoomArray  rooms;

 public:
  RoomScrap(const CMapDoc& doc, int numSelected, const ByteVec& selectedRooms,
            int numSelectedPages, const ByteVec& selectedPages);
  ~RoomScrap() {};
  UndoRec*  paste(CMapDoc& doc) const;
}; // end RoomScrap

/////////////////////////////////////////////////////////////////////////////
// Local Variables:
// mode: c++
// compile-command: "nmake -f GUEmap.mak"
// End:
