//--------------------------------------------------------------------
// $Id: MapView.cpp 4718 2007-09-15 02:34:13Z cjm $
//--------------------------------------------------------------------
//
//   GUEmap
//   Copyright 1996-2007 by Christopher J. Madsen
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License along
//   with this program; if not, write to the Free Software Foundation, Inc.,
//   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//   MapView.cpp : implementation of the CMapView class
//
//--------------------------------------------------------------------

#include "StdAfx.hpp"
#include "GUEmap.hpp"

#include "MapDoc.hpp"
#include "MapView.hpp"
#include "MainFrm.hpp"
#include "Properties.hpp"
#include "Undo.hpp"

#include <math.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

const int
  penEdgeWidth = 9,
  penPageWidth = 20,
  penRoomWidth = 5,
  maxZoom      = 200,
  minZoom      = 20,
  minZoomGrid  = 50;

const UINT
  clickTicks   = 500,
  menuTimer    = 1;

//===========================================================================
// Miscellaneous functions:
//---------------------------------------------------------------------------
// Draw dashed line:
//
// Input:
//   dc:  The device context to use
//        The current point must be START, and the pen must be suitable
//   start:  The starting point of the line
//   end:    The end point of the line

void drawDashedLine(CDC* dc, const CPoint& start, const CPoint& end)
{
  const int  dashSize = 24;    // PORT

  CSize step(0,0);
  int   count;

  if (start.x == end.x) {
    step.cy = dashSize * ((end.y > start.y) ? 1 : -1);
    count = abs(end.y - start.y) / (2*dashSize);
  } else if (start.y == end.y) {
    step.cx = dashSize * ((end.x > start.x) ? 1 : -1);;
    count = abs(end.x - start.x) / (2*dashSize);
  } else {
    const int x = end.x - start.x;
    const int y = end.y - start.y;
    const int d = sqrt(x*x + y*y);
    step.cx = x * dashSize / d;
    step.cy = y * dashSize / d;
    count = d / (2*dashSize);
  }

  CPoint pos(start);
  while (count--) {
    pos += step;
    dc->LineTo(pos);
    pos += step;
    dc->MoveTo(pos);
  }

  dc->LineTo(end);
} // end drawDashedLine

//---------------------------------------------------------------------------
struct DirectionName
{
  RoomCorner   dir;
  const char*  name;
}; // end DirectionName

static const DirectionName  directions[] = {
  { rcN,   "n" },   { rcS,   "s" },   { rcE,   "e" },   { rcW,   "w" },
  { rcNE,  "ne" },  { rcNW,  "nw" },  { rcSE,  "se" },  { rcSW,  "sw" },
  { rcNNE, "nne" }, { rcNNW, "nnw" }, { rcSSE, "sse" }, { rcSSW, "ssw" },
  { rcUp,  "u" },   { rcDown, "d" },  { rcIn,  "i" },   { rcOut, "o" },
  { rcN, "north" }, { rcS, "south" }, { rcE, "east" },  { rcW, "west" },
  { rcNE, "northeast" }, { rcNW, "northwest" },
  { rcSE, "southeast" }, { rcSW, "southwest" },
  { rcUp, "up" }, { rcDown, "down" }, { rcIn, "in" }, { rcOut, "out" },
  { rcNone, NULL }
}; // end directions

static const char *const a2z = "abcdefghijklmnopqrstuvwxyz";

//---------------------------------------------------------------------------
// Parse a direction word:
//
// Input:
//   text:  The string to be tested (must be in lowercase, may be empty)
//
// Output:
//   text:  Has the direction word and whitespace stripped from the beginning
//
// Returns:
//   The direction indicated by the text
//   rcNone if not a direction

RoomCorner  parseDirection(CString& text)
{
  const CString  word(text.SpanIncluding(a2z));

  if (!word.IsEmpty())
    for (int i=0; directions[i].dir != rcNone; i++)
      if (word == directions[i].name) {
        text = text.Mid(word.GetLength());
        text.TrimLeft();
        return directions[i].dir;
      }

  return rcNone;
} // end parseDirection

/////////////////////////////////////////////////////////////////////////////
// CRepaginateDlg dialog

class CRepaginateDlg : public CDialog
{
  // Construction
 public:
  CRepaginateDlg();   // standard constructor

  // Dialog Data
	//{{AFX_DATA(CRepaginateDlg)
	enum { IDD = IDD_REPAGINATE };
	//}}AFX_DATA

  // Overrides
  // ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CRepaginateDlg)
	//}}AFX_VIRTUAL

  // Implementation
 protected:

  // Generated message map functions
	//{{AFX_MSG(CRepaginateDlg)
	afx_msg void OnYes();
	virtual BOOL OnInitDialog();
	//}}AFX_MSG
  DECLARE_MESSAGE_MAP();
}; // end CRepaginateDlg



/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// CMapView

IMPLEMENT_DYNCREATE(CMapView, CScrollZoomView)

BEGIN_MESSAGE_MAP(CMapView, CScrollZoomView)
	//{{AFX_MSG_MAP(CMapView)
	ON_COMMAND(ID_EDIT_ADD_CORNER, OnEditAddCorner)
	ON_COMMAND(ID_EDIT_ADD_PAGE, OnEditAddPage)
	ON_COMMAND(ID_EDIT_ADD_ROOM, OnEditAddRoom)
	ON_COMMAND(ID_EDIT_CLEAR, OnEditClear)
	ON_COMMAND(ID_EDIT_COPY, OnEditCopy)
	ON_COMMAND(ID_EDIT_CUT, OnEditCut)
	ON_COMMAND(ID_EDIT_PAGINATE, OnEditPaginate)
	ON_COMMAND(ID_EDIT_PASTE, OnEditPaste)
	ON_COMMAND(ID_EDIT_SELECT_ALL, OnEditSelectAll)
	ON_COMMAND(ID_EDIT_SELECT_CONNECTED, OnEditSelectConnected)
	ON_COMMAND(ID_VIEW_GRID, OnViewGrid)
	ON_COMMAND(ID_WINDOW_REFRESH, OnWindowRefresh)
	ON_UPDATE_COMMAND_UI(ID_EDIT_ADD_CORNER, OnUpdateEditAddCorner)
	ON_UPDATE_COMMAND_UI(ID_EDIT_CLEAR, OnUpdateSelectedUnlocked)
	ON_UPDATE_COMMAND_UI(ID_EDIT_COPY, OnUpdateSelectedUnlocked)
	ON_UPDATE_COMMAND_UI(ID_EDIT_CUT, OnUpdateSelectedUnlocked)
	ON_UPDATE_COMMAND_UI(ID_EDIT_PASTE, OnUpdateEditPaste)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SELECT_ALL, OnUpdateUnlocked)
	ON_UPDATE_COMMAND_UI(ID_EDIT_SELECT_CONNECTED, OnUpdateSelectedUnlocked)
	ON_UPDATE_COMMAND_UI(ID_VIEW_GRID, OnUpdateViewGrid)
	ON_WM_CHAR()
	ON_WM_KEYDOWN()
	ON_WM_LBUTTONDBLCLK()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_WM_MOUSEWHEEL()
	ON_WM_RBUTTONDOWN()
	ON_WM_RBUTTONUP()
	ON_WM_SIZE()
	ON_WM_TIMER()
	//}}AFX_MSG_MAP
  ON_WM_MOUSEACTIVATE()
  ON_BN_CLICKED(IDC_NAV_GO, OnNavGo)
  ON_COMMAND_RANGE(ID_EDIT_PROPERTIES, ID_EDIT_MAP_PROPERTIES, OnEditProperties)
  ON_COMMAND_RANGE(ID_NAV_GO_NORTH, ID_NAV_STUB2_OUT, OnNavGoDir)
  ON_COMMAND_RANGE(ID_VIEW_ZOOM_IN, ID_VIEW_ZOOM_OUT, OnViewZoom)
  ON_UPDATE_COMMAND_UI_RANGE(ID_VIEW_ZOOM_IN, ID_VIEW_ZOOM_OUT,
                             OnUpdateViewZoom)
  // Standard printing commands
  ON_COMMAND(ID_FILE_PRINT, CView::OnFilePrint)
  ON_COMMAND(ID_FILE_PRINT_DIRECT, CView::OnFilePrint)
  ON_COMMAND(ID_FILE_PRINT_PREVIEW, CView::OnFilePrintPreview)
END_MESSAGE_MAP();

//===========================================================================
// Class CMapView:
//
// Member Variables:
//   opInProgress:  Indicates what's happening while the mouse button is down
//     gmoAddCorner:
//       iTmp:     The edge that needs a corner
//     gmoAddEdge:
//       bTmp:     The modification state of the document
//       edgeTmp:  The edge being created
//       e2Tmp:    The original edge data (if iTmp >= 0)
//       iTmp:     -1 if this is a new edge, >=0 if changing an old edge
//       p1Tmp:    The start point (fixed) of the rubber band cursor
//       p2Tmp:    The end point (moving) of the rubber band cursor
//     gmoChangeEdge:
//       bTmp:          The modification state of the document
//       edgeTmp:       Info about the other end of this edge
//       iTmp:          The edge number about to be changed
//       p1Tmp, p2Tmp:  The start point (fixed) of the rubber band cursor
//     gmoDeleting:
//       Used in OnEditClear to tell OnUpdate not to clear the selection
//     gmoSelectBox:
//       p1Tmp:  The fixed corner of the selection rectangle
//       p2Tmp:  The moving corner of the selection rectangle
//       rTmp:   The current selection rectangle
//     gmoControlDown:
//       bTmp:   TRUE if the room or page was already selected
//       b2Tmp:  True if this is a page instead of a room
//       iTmp:   The room number where the button was pressed (if bTmp==TRUE)
//       p1Tmp:  The grid cell where the mouse button was pressed
//       p2Tmp:  The offset to the current grid cell (in logical coordinates)
//     gmoShiftDown:
//       iTmp:   The number of the selected page (-1 if a room)
//       p1Tmp:  The grid cell where the mouse button was pressed
//       p2Tmp:  The position of the selected room
//     gmoDrag: (all dragging operations)
//       p2Tmp:
//         The offset to the current grid cell (in logical coordinates)
//       rTmp:
//         A rectangle enclosing the original position of all rooms being
//         dragged.  Used to ensure we don't push a room off the edge.
//   font:
//     The font used for room names
//   numSelected:
//     The number of rooms currently selected
//   numSelectedPages:
//     The number of pages currently selected
//   printingPage:
//     The number of the page being printed or previewed
//     0 means normal screen display
//   scrollDrag:
//     True if we are in the middle of a scroll-by-drag operation
//   scrollDragStart:
//     The point under the cursor when a scrollDrag started
//   selectedOne:
//     The selected room's number, if exactly 1 room is selected
//     -1 otherwise
//   selected:
//     selected[N] is non-zero iff room N is selected
//       2 means the room should stay selected during gmoSelectBox
//   selectedPage:
//     selectedPage[N] is non-zero iff page N is selected
//       2 means the page should stay selected during gmoSelectBox
//   zoom:
//     Zoom factor in percent
//---------------------------------------------------------------------------
// CMapView construction/destruction

CMapView::CMapView()
: opInProgress(gmoNone),
  printingPage(0),
  scrollDrag(false),
  scrollDragTimer(0),
  showCorners(false),
  showGrid(true),
  showPages(false),
  zoom(100)
{
} // end CMapView::CMapView

CMapView::~CMapView()
{
  gueApp()->closingView(this);  // Remove any comment for this view
} // end CMapView::~CMapView

BOOL CMapView::PreCreateWindow(CREATESTRUCT& cs)
{
  cs.lpszClass = mapWinClass.c_str();
  return CScrollZoomView::PreCreateWindow(cs);
} // end CMapView::PreCreateWindow

/////////////////////////////////////////////////////////////////////////////
// CMapView drawing

static const int  textOffset[maxNameLines][maxNameLines] = // PORT
{
  { roomHeight/2 - 19 },
  { roomHeight/2 + 9, roomHeight/2 - 48 },
  { roomHeight/2 + 38, roomHeight/2 - 19, roomHeight/2 - 77 }
}; // end textOffset

static const char  edgeLabel[] = { 'u', 'd', 'i', 'o' };

static const int  edgLblXOffset[] = { // PORT
  roomWidth/2+24, roomWidth/2-24, roomWidth+24, -24, // N,S,E,W
  roomWidth-24, 24, roomWidth-24, 24,                // NE,NW,SE,SW
  3*roomWidth/4+24, roomWidth/4+24, 3*roomWidth/4-24, roomWidth/4-24
}; // end edgLblXOffset

static const int  edgLblYOffset[] = { // PORT
  roomHeight+5, -43, roomHeight/2-57, roomHeight/2+19, // N,S,E,W
  roomHeight+5, roomHeight+5, -43, -43,                // NE,NW,SE,SW
  roomHeight+5, roomHeight+5, -43, -43
}; // end edgLblYOffset

static const int // PORT
  noteXOffset = roomWidth - 24,
  noteYOffset = 57;

static const int // PORT
  cantGoSide  = 19,
  cantGoAngleX= gridY/6,
  cantGoAngleY= gridX/6,
  oneWaySide  = 19,
  oneWayBack1 = 24,
  oneWayBack2 =  5,
  oneWayCrnrO1 = 28,
  oneWayCrnrO2 = 7,
  oneWayLen      = 28;

void CMapView::OnDraw(CDC* dc)
{
  const CMapDoc* doc = GetDocument();
  ASSERT_VALID(doc);

  CBrush  focusBrush;
  CPen penEdge;
  CPen penGrid;
  CPen penRoom;
  LOGBRUSH  lb;
  lb.lbStyle = BS_SOLID;
  lb.lbColor = RGB(0,0,0);

  if (!focusBrush.CreateHatchBrush(HS_BDIAGONAL,PALETTERGB(128,128,128)) ||
      !penEdge.CreatePen(PS_GEOMETRIC|PS_ENDCAP_FLAT, penEdgeWidth, &lb) ||
      !penGrid.CreatePen(PS_DOT, 0, RGB(0,0,0)) ||
      !penRoom.CreatePen(PS_SOLID, penRoomWidth, RGB(0,0,0)))
    return;
  CBrush *const  whiteBrush = dc->GetCurrentBrush();
  CPen *const    oldPen     = dc->GetCurrentPen();
  CFont *const   oldFont    = dc->SelectObject(&font);
  const UINT  oldAlign = dc->SetTextAlign(TA_CENTER|TA_BASELINE|TA_NOUPDATECP);
  int i;
  CRect  r;

  if (printingPage) {
    // Clipping in OnPrepareDC screws up print preview
    doc->getPageRect(printingPage-1, r);
    dc->SetWindowOrg(r.left, r.bottom); // FIXME adjust for margins
    dc->IntersectClipRect(&r);
  } // end if printing or previewing

  //..................................................................
  // Draw the page outlines:
  if ((showPages || numSelectedPages) && !dc->IsPrinting()) {
    dc->SetBkMode(TRANSPARENT);
    CPen  penPage;
    lb.lbColor = RGB(192,192,192);
    VERIFY(penPage.CreatePen(PS_GEOMETRIC|PS_ENDCAP_FLAT, penPageWidth, &lb));
    dc->SelectObject(&penPage);
    const PageVec&  pageVec = doc->getPages();
    TCHAR  pageNum[4];
    int  pNum = 1;
    dc->SelectStockObject(NULL_BRUSH);
    CBrush *const  nullBrush = dc->GetCurrentBrush();
    bool  focused = false;
    for (PageConstItr  p = pageVec.begin(); p != pageVec.end(); ++p, ++pNum) {
      doc->getPageRect(*p, r);
      if (dc->RectVisible(r)) {
        const bool selectedI = selectedPage[pNum-1] != 0;
        if (!selectedI && !showPages) continue; // Only show selected pages
        if (selectedI != focused) {
          focused = selectedI;
          dc->SelectObject(focused ? &focusBrush : nullBrush);
        }
///    CRgn inside,outside;
///    VERIFY(outside.CreateRectRgnIndirect(&r));
///    VERIFY(inside.CreateRectRgn(r.left + gridX, r.bottom - gridY,
///                                r.right - gridX, r.top + gridY));
///    outside.CombineRgn(&outside, &inside, RGN_DIFF);
///    dc->FillRgn(&outside, &focusBrush);
        dc->Rectangle(r.left+penPageWidth/4, r.bottom-penPageWidth/4,
                      r.right, r.top);
        ASSERT(pNum < 1000);
        _itot(pNum, pageNum, 10);
        const int pageNumL = strlen(pageNum);
        for (i = r.left + 3*gridX/2; i < r.right - gridX; i += gridX) {
          dc->TextOut(i, r.bottom - gridY*7/8, pageNum, pageNumL);
          dc->TextOut(i, r.top + gridY/4, pageNum, pageNumL);
        }
        for (i = r.bottom - gridY*15/8; i > r.top + gridY; i -= gridY) {
          dc->TextOut(r.left + gridX/2 + penPageWidth/4, i, pageNum, pageNumL);
          dc->TextOut(r.right- gridX/2 - penPageWidth/4, i, pageNum, pageNumL);
        }
      } // end if page visible
    } // end for pages
    dc->SetBkMode(OPAQUE);
    dc->SelectObject(whiteBrush);
    dc->SelectObject(&penEdge); // Before destroying penPage
  } else // not showing pages
    dc->SelectObject(&penEdge);

  //..................................................................
  // Draw the room connections:
  i = doc->getEdgeCount();
  EdgeConstItr  edge = doc->getEdges().end() - 1;
  RoomConstItr  room = doc->getRooms().begin();
  while (i-- > 0) {
    CPoint start, end;
    doc->getEdgePoints(*edge, start.x, start.y, end.x, end.y);
    dc->MoveTo(start);
    if (edge->type1 & etObstructed)
      drawDashedLine(dc, start, end);
    else
      dc->LineTo(end);
    if (edge->type1 & etOneWay) {
      if (start.x == end.x) {
        const int  sign = (start.y > end.y ? 1 : -1);
        dc->MoveTo(end.x - oneWaySide, end.y + sign*oneWayBack1);
        dc->LineTo(end.x,              end.y + sign*oneWayBack2);
        dc->LineTo(end.x + oneWaySide, end.y + sign*oneWayBack1);
      } else if (start.y == end.y) {
        const int  sign = (start.x > end.x ? 1 : -1);
        dc->MoveTo(end.x + sign*oneWayBack1, end.y + oneWaySide);
        dc->LineTo(end.x + sign*oneWayBack2, end.y);
        dc->LineTo(end.x + sign*oneWayBack1, end.y - oneWaySide);
      } else if (abs((start.x-end.x)/gridX) == abs((start.y-end.y)/gridY)) {
        const int  signX = (start.x > end.x ? 1 : -1);
        const int  signY = (start.y > end.y ? 1 : -1);
        dc->MoveTo(end.x +signX*oneWayCrnrO1, end.y -signY*oneWayCrnrO2);
        dc->LineTo(end);
        dc->LineTo(end.x +signX*oneWayCrnrO2, end.y +signY*oneWayCrnrO1);
      } else {
        const double  a = atan2(double(start.y - end.y), // 45 degree angle
                                double(start.x - end.x)) + 0.785398163;
        const int  o1 = int(oneWayLen * cos(a));
        const int  o2 = int(oneWayLen * sin(a));
        dc->MoveTo(end.x + o1, end.y + o2);
        dc->LineTo(end);
        dc->LineTo(end.x + o2, end.y - o1);
      } // end else different angle
    } // end if one-way connection
    else if (edge->type1 & etNoExit) {
      if (start.x == end.x) {
        dc->MoveTo(end.x - cantGoSide, end.y);
        dc->LineTo(end.x + cantGoSide, end.y);
      } else if (start.y == end.y) {
        dc->MoveTo(end.x, end.y - cantGoSide);
        dc->LineTo(end.x, end.y + cantGoSide);
      } else if ((start.x > end.x) == (start.y > end.y)) {
        dc->MoveTo(end.x + cantGoAngleX, end.y - cantGoAngleY);
        dc->LineTo(end.x - cantGoAngleX, end.y + cantGoAngleY);
      } else {
        dc->MoveTo(end.x - cantGoAngleX, end.y - cantGoAngleY);
        dc->LineTo(end.x + cantGoAngleX, end.y + cantGoAngleY);
      }
    } // end else if can't go that way
    if (edge->type1 & etDirection) {
      const CPoint&  p = room[edge->room1]->pos;
      dc->TextOut(p.x + edgLblXOffset[edge->end1],
                  p.y + edgLblYOffset[edge->end1],
                  &edgeLabel[(edge->type1 & etDirection)-1], 1);
    }
    if (edge->type2 & etDirection) {
      const CPoint&  p = room[edge->room2]->pos;
      dc->TextOut(p.x + edgLblXOffset[edge->end2],
                  p.y + edgLblYOffset[edge->end2],
                  &edgeLabel[(edge->type2 & etDirection)-1], 1);
    }
    --edge;
  } // end while more edges

  //..................................................................
  // Draw the background grid:
  if (showGrid && (zoom >= minZoomGrid) && !dc->IsPrinting()) {
    dc->SetBkMode(TRANSPARENT);
    dc->SelectObject(&penGrid);
    CRect clip;
    dc->GetClipBox(&clip);
    clip.left   -= clip.left % gridX;
    clip.bottom -= clip.bottom % gridY + gridY;
    if (clip.top < -gridY) clip.top += gridY;
    CSize  docSize(doc->getDocSize());
    docSize.cy *= -1;
    if (clip.right > docSize.cx) clip.right = docSize.cx;
    if (clip.bottom < docSize.cy) clip.bottom = docSize.cy;
    for (i = clip.left; i <= clip.right; i += gridX) {
      dc->MoveTo(i,0);
      dc->LineTo(i,docSize.cy);
    }
    for (i = clip.bottom; i <= clip.top; i += gridY) {
      dc->MoveTo(0,i);
      dc->LineTo(docSize.cx,i);
    }
    dc->SetBkMode(OPAQUE);
  } // end if drawing grid

  //..................................................................
  // Draw the rooms:
  dc->SelectObject(&penRoom);
  bool  focused = false;
  i = 0;
  for (const int count = doc->getRoomCount(); i < count; ++i, ++room) {
    if (dc->RectVisible((*room)->getRect(r))) {
      const bool selectedI = selected[i] != 0;
      if (selectedI != focused) {
        focused = selectedI;
        dc->SelectObject(focused ? &focusBrush : whiteBrush);
      }
      bool  needOpaqueText = false;
      if ((*room)->flags & rfBorder)
        dc->Rectangle(r);
      else if (focused)
        dc->FillRect(r, &focusBrush);
      else // Room has no border and is not focused
        needOpaqueText = !(*room)->name.empty();

      if ((*room)->isCorner() && (focused || showCorners)) {
        r.DeflateRect(gridX*343/1000,gridY*238/1000); // PORT
        dc->FillSolidRect(r, RGB(0,0,0));
        dc->SetBkColor(RGB(255,255,255)); // FIXME change background?
      }

      dc->SetBkMode(TRANSPARENT);
      if (!(*room)->note.empty()) {
        COLORREF oldColor = dc->SetTextColor(RGB(0xFF,0,0));
        dc->TextOut(r.left + noteXOffset, r.bottom - noteYOffset, _T("*"),1);
        dc->SetTextColor(oldColor);
      }
      if (needOpaqueText)
        dc->SetBkMode(OPAQUE);
      StrConstItr  name((*room)->name.begin());
      const int*  offset = textOffset[(*room)->lines-1];
      for (int j = 0; j < (*room)->lines; j++) {
        dc->TextOut(r.left + roomWidth/2, r.top + *(offset++),
                    name, (*room)->nameLength[j]);
        name += (*room)->nameLength[j] + 2;
      } // end for
      dc->SetBkMode(OPAQUE);
    } // end if room is visible
  } // end for rooms
  dc->SelectObject(whiteBrush);
  dc->SelectObject(oldFont);
  dc->SelectObject(oldPen);
  dc->SetTextAlign(oldAlign);
} // end CMapView::OnDraw

//--------------------------------------------------------------------
// Draw the grid and non-selected rooms & edges:
//
// Prepares for a drag operation by erasing the selected rooms and edges.

void CMapView::drawBackground(CDC* dc)
{
  const CMapDoc* doc = GetDocument();
  ASSERT_VALID(doc);

  ValidateRect(NULL);           // Mark the window as updated

  CPen penEdge;
  CPen penGrid;
  CPen penRoom;
  LOGBRUSH  lb;
  lb.lbStyle = BS_SOLID;
  lb.lbColor = RGB(0,0,0);

  if (!penEdge.CreatePen(PS_GEOMETRIC|PS_ENDCAP_FLAT, penEdgeWidth, &lb) ||
      !penGrid.CreatePen(PS_DOT, 0, RGB(0,0,0)) ||
      !penRoom.CreatePen(PS_SOLID, penRoomWidth, RGB(0,0,0)))
    return;
  CPen*  oldPen = dc->GetCurrentPen();
  UINT oldAlign = dc->SetTextAlign(TA_CENTER|TA_BASELINE|TA_NOUPDATECP);
  CFont* oldFont = dc->SelectObject(&font);

  int i;
  CRect  clip, r;
  dc->GetClipBox(&clip);
  dc->FillSolidRect(clip,dc->GetBkColor());

  //..................................................................
  // Draw the page outlines:
  if (showPages && !dc->IsPrinting()) {
    dc->SetBkMode(TRANSPARENT);
    CPen  penPage;
    lb.lbColor = RGB(192,192,192);
    VERIFY(penPage.CreatePen(PS_GEOMETRIC|PS_ENDCAP_FLAT, penPageWidth, &lb));
    dc->SelectObject(&penPage);
    const PageVec&  pageVec = doc->getPages();
    TCHAR  pageNum[4];
    int  pNum = 1;
    CBrush *const  whiteBrush =
      static_cast<CBrush*>(dc->SelectStockObject(NULL_BRUSH));
    for (PageConstItr  p = pageVec.begin(); p != pageVec.end(); ++p, ++pNum) {
      doc->getPageRect(*p, r);
      if (dc->RectVisible(r) && !selectedPage[pNum-1]) {
        dc->Rectangle(r.left+penPageWidth/4, r.bottom-penPageWidth/4,
                      r.right, r.top);
        ASSERT(pNum < 1000);
        _itot(pNum, pageNum, 10);
        const int pageNumL = strlen(pageNum);
        for (i = r.left + 3*gridX/2; i < r.right - gridX; i += gridX) {
          dc->TextOut(i, r.bottom - gridY*7/8, pageNum, pageNumL);
          dc->TextOut(i, r.top + gridY/4, pageNum, pageNumL);
        }
        for (i = r.bottom - gridY*15/8; i > r.top + gridY; i -= gridY) {
          dc->TextOut(r.left + gridX/2 + penPageWidth/4, i, pageNum, pageNumL);
          dc->TextOut(r.right- gridX/2 - penPageWidth/4, i, pageNum, pageNumL);
        }
      } // end if page visible
    } // end for pages
    dc->SetBkMode(OPAQUE);
    dc->SelectObject(whiteBrush);
    dc->SelectObject(&penEdge); // Before destroying penPage
  } else // not showing pages
    dc->SelectObject(&penEdge);

  //..................................................................
  // Draw the room connections:
  i = doc->getEdgeCount();
  EdgeConstItr  edge = doc->getEdges().end() - 1;
  RoomConstItr  room = doc->getRooms().begin();
  while (i-- > 0) {
    if (!selected[edge->room1] &&
        ((edge->type1 & etNoRoom2) || !selected[edge->room2])) {
      CPoint start, end;
      doc->getEdgePoints(i, start.x, start.y, end.x, end.y);
      dc->MoveTo(start);
      if (edge->type1 & etObstructed)
        drawDashedLine(dc, start, end);
      else
        dc->LineTo(end);
      if (edge->type1 & etOneWay) {
        if (start.x == end.x) {
          const int  sign = (start.y > end.y ? 1 : -1);
          dc->MoveTo(end.x - oneWaySide, end.y + sign*oneWayBack1);
          dc->LineTo(end.x,              end.y + sign*oneWayBack2);
          dc->LineTo(end.x + oneWaySide, end.y + sign*oneWayBack1);
        } else if (start.y == end.y) {
          const int  sign = (start.x > end.x ? 1 : -1);
          dc->MoveTo(end.x + sign*oneWayBack1, end.y + oneWaySide);
          dc->LineTo(end.x + sign*oneWayBack2, end.y);
          dc->LineTo(end.x + sign*oneWayBack1, end.y - oneWaySide);
        } else if (abs((start.x-end.x)/gridX) == abs((start.y-end.y)/gridY)) {
          const int  signX = (start.x > end.x ? 1 : -1);
          const int  signY = (start.y > end.y ? 1 : -1);
          dc->MoveTo(end.x +signX*oneWayCrnrO1, end.y -signY*oneWayCrnrO2);
          dc->LineTo(end);
          dc->LineTo(end.x +signX*oneWayCrnrO2, end.y +signY*oneWayCrnrO1);
        } else {
          const double  a = atan2(double(start.y - end.y), // 45 degree angle
                                  double(start.x - end.x)) + 0.785398163;
          const int  o1 = int(oneWayLen * cos(a));
          const int  o2 = int(oneWayLen * sin(a));
          dc->MoveTo(end.x + o1, end.y + o2);
          dc->LineTo(end);
          dc->LineTo(end.x + o2, end.y - o1);
        } // end else different angle
      } // end if one-way connection
      else if (edge->type1 & etNoExit) {
        if (start.x == end.x) {
          dc->MoveTo(end.x - cantGoSide, end.y);
          dc->LineTo(end.x + cantGoSide, end.y);
        } else if (start.y == end.y) {
          dc->MoveTo(end.x, end.y - cantGoSide);
          dc->LineTo(end.x, end.y + cantGoSide);
        } else if ((start.x > end.x) == (start.y > end.y)) {
          dc->MoveTo(end.x + cantGoAngleX, end.y - cantGoAngleY);
          dc->LineTo(end.x - cantGoAngleX, end.y + cantGoAngleY);
        } else {
          dc->MoveTo(end.x - cantGoAngleX, end.y - cantGoAngleY);
          dc->LineTo(end.x + cantGoAngleX, end.y + cantGoAngleY);
        }
      } // end else if can't go that way
      if (edge->type1 & etDirection) {
        const CPoint&  p = room[edge->room1]->pos;
        dc->TextOut(p.x + edgLblXOffset[edge->end1],
                    p.y + edgLblYOffset[edge->end1],
                    &edgeLabel[(edge->type1 & etDirection)-1], 1);
      }
      if (edge->type2 & etDirection) {
        const CPoint&  p = room[edge->room2]->pos;
        dc->TextOut(p.x + edgLblXOffset[edge->end2],
                    p.y + edgLblYOffset[edge->end2],
                    &edgeLabel[(edge->type2 & etDirection)-1], 1);
      }
    } // end if rooms are not selected
    --edge;
  } // end while edges

  //..................................................................
  // Draw the grid:
  if (showGrid && (zoom >= minZoomGrid)) {
    dc->SetBkMode(TRANSPARENT);
    dc->SelectObject(&penGrid);
    clip.left   -= clip.left % gridX;
    clip.bottom -= clip.bottom % gridY + gridY;
    if (clip.top < -gridY) clip.top += gridY;
    CSize  docSize(doc->getDocSize());
    docSize.cy *= -1;
    if (clip.right > docSize.cx) clip.right = docSize.cx;
    if (clip.bottom < docSize.cy) clip.bottom = docSize.cy;
    for (i = clip.left; i <= clip.right; i += gridX) {
      dc->MoveTo(i,0);
      dc->LineTo(i,docSize.cy);
    }
    for (i = clip.bottom; i <= clip.top; i += gridY) {
      dc->MoveTo(0,i);
      dc->LineTo(docSize.cx,i);
    }
    dc->SetBkMode(OPAQUE);
  } // end if showing grid

  //..................................................................
  // Draw the rooms:
  dc->SelectObject(&penRoom);
  i = 0;
  for (const int count = doc->getRoomCount(); i < count; ++i, ++room) {
    if (!selected[i] && dc->RectVisible((*room)->getRect(r))) {
      if ((*room)->flags & rfBorder)
        dc->Rectangle(r);
      if (!(*room)->note.empty()) {
        dc->SetBkMode(TRANSPARENT);
        COLORREF oldColor = dc->SetTextColor(RGB(0xFF,0,0));
        dc->TextOut(r.left + noteXOffset, r.bottom - noteYOffset, _T("*"),1);
        dc->SetTextColor(oldColor);
        dc->SetBkMode(OPAQUE);
      }
      StrConstItr  name((*room)->name.begin());
      const int*  offset = textOffset[(*room)->lines-1];
      for (int j = 0; j < (*room)->lines; j++) {
        dc->TextOut(r.left + roomWidth/2, r.top + *(offset++),
                    name, (*room)->nameLength[j]);
        name += (*room)->nameLength[j] + 2;
      }
    }
  } // end for rooms
  dc->SelectObject(oldFont);
  dc->SelectObject(oldPen);
  dc->SetTextAlign(oldAlign);
} // end CMapView::drawBackground

//--------------------------------------------------------------------
// Draw the selected rooms & edges:
//
// The selected rooms and edges are offset by p2Tmp.
// This should be called with the DC in R2_NOT mode.

void CMapView::drawSelected(CDC* dc)
{
  const CMapDoc* doc = GetDocument();
  ASSERT_VALID(doc);

  CPen penEdge;
  if (!penEdge.CreatePen(PS_SOLID, penEdgeWidth, RGB(0,0,0)))
    return;
  CBrush* oldBrush = static_cast<CBrush*>(dc->SelectStockObject(NULL_BRUSH));
  CPen*  oldPen = dc->SelectObject(&penEdge);
  CRect  r;
  int    i;

  if (numSelectedPages)
    for (i = doc->getPages().size()-1; i >= 0; --i) {
      if (selectedPage[i]) {
        doc->getPageRect(i, r);
        r += p2Tmp;
        dc->Rectangle(r);
      }
    } // end for pages

  if (numSelected) {
    i = doc->getEdgeCount();
    EdgeConstItr  edge = doc->getEdges().end() - 1;
    while (i-- > 0) {
      if (selected[edge->room1] ||
          (!(edge->type1 & etNoRoom2) && selected[edge->room2])) {
        CPoint start, end;
        doc->getEdgePoints(i, start.x, start.y, end.x, end.y);
        if (selected[edge->room1]) start += p2Tmp;
        if ((edge->type1 & etNoRoom2) || selected[edge->room2]) end += p2Tmp;
        dc->MoveTo(start);
        dc->LineTo(end);
      }
      --edge;
    } // end while more edges

    RoomConstItr  room = doc->getRooms().begin();
    i = 0;
    for (const int count = doc->getRoomCount(); i < count; i++) {
      if (selected[i] && !room[i]->isCorner()) {
        room[i]->getRect(r) += p2Tmp;
        dc->Rectangle(r);
      }
    }
  } // end if selected rooms

  dc->SelectObject(oldBrush);
  dc->SelectObject(oldPen);
} // end CMapView::drawSelected

/////////////////////////////////////////////////////////////////////////////
// CMapView printing
//--------------------------------------------------------------------
// Record which page we're printing, if any:
//
// Output Variables:
//   printingPage

void CMapView::OnPrepareDC(CDC* dc, CPrintInfo* pInfo)
{
  CScrollZoomView::OnPrepareDC(dc, pInfo);
  if (pInfo)
    printingPage = pInfo->m_nCurPage;
  else
    printingPage = 0;
} // end CMapView::OnPrepareDC

//--------------------------------------------------------------------
BOOL CMapView::OnPreparePrinting(CPrintInfo* pInfo)
{
  // Require registration before printing map with more than 10 rooms:
  CMapDoc *const  doc = GetDocument();
  ASSERT_VALID(doc);

  if (doc->needRepaginate()) {
    CRepaginateDlg  d;
    switch (d.DoModal()) {
     case IDCANCEL: return false;
     case IDYES:    doc->layoutPages(); break;
    }
  } // end if need to repaginate

  pInfo->SetMaxPage(doc->getPages().size());
  return DoPreparePrinting(pInfo);
} // end CMapView::OnPreparePrinting

void CMapView::OnBeginPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
  // Don't show selection or corners while printing:
  //   We do this here instead of OnPreparePrinting, because OnEndPrinting
  //   is only guaranteed to be called if OnBeginPrinting is.
  clearSelection();
  selectDone();
  showCorners = false;
} // end CMapView::OnBeginPrinting

void CMapView::OnEndPrinting(CDC* /*pDC*/, CPrintInfo* /*pInfo*/)
{
  if (GetDocument()->locked)
    OnUpdate(NULL, dupNavigationMode, NULL); // Select a room
} // end CMapView::OnEndPrinting

/////////////////////////////////////////////////////////////////////////////
// CMapView diagnostics

#ifdef _DEBUG
void CMapView::AssertValid() const
{
  CScrollZoomView::AssertValid();
}

void CMapView::Dump(CDumpContext& dc) const
{
  CScrollZoomView::Dump(dc);
}

const CMapDoc* CMapView::GetDocument() const // non-debug version is inline
{
  ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CMapDoc)));
  return (CMapDoc*)m_pDocument;
}

CMapDoc* CMapView::GetDocument() // non-debug version is inline
{
  ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CMapDoc)));
  return (CMapDoc*)m_pDocument;
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CMapView miscellaneous
//--------------------------------------------------------------------
// Add a room:
//
// Input:
//   point:  The point where the mouse was clicked (in logical coordinates)

void CMapView::addRoom(CPoint& point)
{
  CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);

  CSize  docSize(doc->getDocSize());
  docSize.cy *= -1;

  if (doc->locked || (point.x > docSize.cx) || (point.y < docSize.cy)) {
    MessageBeep(MB_ICONASTERISK);
    return;
  }
  point.x -= roomWidth/2 - gridX/2;
  point.x -= point.x % gridX;
  point.y -= roomHeight/2 + gridY/2;
  point.y -= point.y % gridY;

  if (point.x < 0)
    point.x = 0;
  else if (point.x + roomWidth > docSize.cx)
    point.x = docSize.cx - roomWidth;

  if (point.y > -roomHeight)
    point.y = -roomHeight;
  else if (point.y < docSize.cy)
    point.y = docSize.cy;

  clearSelection();
  const bool  wasModified = doc->IsModified();
  const int   rNum = doc->addRoom(point);
  if (rNum >= 0) selectRoom(rNum);
  selectDone();               // FIXME disable floating comments
  editProperties(epuAddRoom, wasModified);
} // end CMapView::addRoom

//---------------------------------------------------------------------------
// Make sure we aren't dragging a room off the edge:
//
// Input:
//   p:  The proposed offset for the selected rooms
//
// Input Variables:
//   rTmp:  Rectangle enclosing the original positions of the selected rooms
//
// Output:
//   p:  The corrected (if necessary) offset for selected rooms

void CMapView::adjustOffset(CPoint& p) const
{
  const CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);
  const CSize size(doc->getDocSize());

  if (p.x + rTmp.left   < 0)       p.x = -rTmp.left;
  if (p.x + rTmp.right  > size.cx) p.x = size.cx - rTmp.right;
  if (-p.y - rTmp.top   > size.cy) p.y = -size.cy - rTmp.top;
  if (p.y + rTmp.bottom > 0)       p.y = -rTmp.bottom;
} // end CMapView::adjustOffset

//---------------------------------------------------------------------------
// Deselect all rooms:
//
// Input:
//   update:  TRUE means the selected rooms should be redrawn

void CMapView::clearSelection(bool update)
{
  const CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);

  if (doc->getRoomCount() != selected.size())
    selected.resize(doc->getRoomCount(), 0);

  if (update && numSelected) {
    CRect r;
    CClientDC dc(this);
    OnPrepareDC(&dc);
    RoomConstItr  room = doc->getRooms().begin();
    for (int i = doc->getRoomCount()-1; numSelected && i>=0; --i)
      if (selected[i]) {
        --numSelected;
        room[i]->getRect(r);
        dc.LPtoDP(r);
        r.NormalizeRect();
        InvalidateRect(r);
      }
  } // end if update && numSelected

  numSelected = 0;
  for (ByteItr b = selected.begin(); b != selected.end(); ++b)
    *b = 0;
  deselectPages(update);
} // end CMapView::clearSelection

//---------------------------------------------------------------------------
// Compute a rectangle enclosing all selected rooms:
//
// Input Variables:
//   numSelected
//   numSelectedPages
//   selected
//   selectedPage
//
// Output:
//   r:  A rectangle enclosing all selected rooms

void CMapView::computeSelectedRect(CRect& r) const
{
  const CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);

  r.SetRectEmpty();

  int i;
  CRect temp;
  if (numSelected) {
    RoomConstItr  room = doc->getRooms().begin();
    for (i = doc->getRoomCount()-1; i >= 0; --i)
      if (selected[i])
        r |= room[i]->getRect(temp);
  } // end if selected rooms

  if (numSelectedPages)
    for (i = doc->getPages().size()-1; i >= 0; --i)
      if (selectedPage[i]) {
        doc->getPageRect(i, temp);
        r |= temp;
      }
} // end CMapView::computeSelectedRect

//---------------------------------------------------------------------------
// Deselect a page:
//
// Input:
//   n:       The page number to be deselected
//   update:  TRUE means the page should be redrawn if necessary
//
// Output Variables:
//   numSelectedPages
//   selectedOne
//   selectedPage

void CMapView::deselectPage(int n, bool update)
{
  const CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);
  ASSERT((n >= 0) && (n < doc->getPages().size()));

  if (selectedPage[n]) {
    selectedPage[n] = false;
    if ((--numSelectedPages == 0) && (numSelected == 1))
      for (selectedOne=0; !selected[selectedOne]; selectedOne++)
        ;                       // Find the selected room
    else
      selectedOne = -1;

    if (update) {
      CClientDC dc(this);
      OnPrepareDC(&dc);
      CRect r;
      doc->getPageRect(n, r);
      dc.LPtoDP(r);
      r.NormalizeRect();
      InvalidateRect(r);
    } // end if update
  } // end if page was selected
} // end CMapView::deselectPage

//---------------------------------------------------------------------------
// Deselect all pages:
//
// Input:
//   update:  TRUE means the selected pages should be redrawn
//
// Output Variables:
//   numSelectedPages
//   selectedOne
//   selectedPage

void CMapView::deselectPages(bool update)
{
  const CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);

  ASSERT(selectedPage.size() == doc->getPages().size());

  if (update && numSelectedPages) {
    CRect r;
    CClientDC dc(this);
    OnPrepareDC(&dc);
    for (int i = doc->getPages().size()-1; numSelectedPages && i>=0; --i)
      if (selectedPage[i]) {
        --numSelectedPages;
        doc->getPageRect(i, r);
        if (!showPages) {
          r.bottom += penPageWidth;
          r.left   -= penPageWidth;
        } // make sure it disappears
        dc.LPtoDP(r);
        r.NormalizeRect();
        InvalidateRect(r);
      }
  } // end if update

  numSelectedPages = 0;
  for (ByteItr b = selectedPage.begin(); b != selectedPage.end(); ++b)
    *b = 0;

  if (numSelected == 1)
    for (selectedOne=0; !selected[selectedOne]; selectedOne++)
      ;                         // Find the selected room
  else
    selectedOne = -1;
} // end CMapView::deselectPages

//---------------------------------------------------------------------------
// Deselect a room:
//
// Input:
//   n:       The room number to be deselected
//   update:  TRUE means the room should be redrawn if necessary
//
// Output Variables:
//   numSelected
//   selectedOne
//   selected

void CMapView::deselectRoom(RoomNum n, bool update)
{
  const CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);
  ASSERT((n >= 0) && (n < doc->getRoomCount()));

  if (doc->getRoomCount() != selected.size())
    selected.resize(doc->getRoomCount(), 0);

  if (selected[n]) {
    selected[n] = false;
    if ((--numSelected == 1) && !numSelectedPages)
      for (selectedOne=0; !selected[selectedOne]; selectedOne++)
        ;                       // Find the selected room
    else
      selectedOne = -1;

    if (update) {
      CClientDC dc(this);
      OnPrepareDC(&dc);
      CRect r;
      doc->getRoom(n).getRect(r);
      dc.LPtoDP(r);
      r.NormalizeRect();
      InvalidateRect(r);
    } // end if update
  } // end if room was selected
} // end CMapView::deselectRoom

//--------------------------------------------------------------------
// Update the room comments dialog after the selection changes:

void CMapView::selectDone()
{
  static_cast<CMapApp*>(AfxGetApp())->setComment(this,
    (selectedOne >= 0)
    ? &(GetDocument()->getRoom(selectedOne))
    : NULL
  );
} // end CMapView::selectDone

//---------------------------------------------------------------------------
// Select a page:
//
// Input:
//   n:       The page number to select
//   update:  TRUE means the selected page should be redrawn if necessary
//
// Output Variables:
//   numSelectedPages
//   selectedOne
//   selectedPage

void CMapView::selectPage(int n, bool update)
{
  const CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);

  if (!selectedPage[n]) {
    selectedPage[n] = true;
    ++numSelectedPages;
    selectedOne = -1;

    if (update) {
      CClientDC dc(this);
      OnPrepareDC(&dc);
      CRect r;
      doc->getPageRect(n, r);
      dc.LPtoDP(r);
      r.NormalizeRect();
      InvalidateRect(r);
    } // end if update
  } // end if page not selected
} // end CMapView::selectPage

//---------------------------------------------------------------------------
// Select a room:
//
// Input:
//   n:       The room number to select
//   update:  TRUE means the selected room should be redrawn if necessary
//
// Output Variables:
//   numSelected
//   selectedOne
//   selected

void CMapView::selectRoom(RoomNum n, bool update)
{
  const CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);

  if (doc->getRoomCount() != selected.size())
    selected.resize(doc->getRoomCount(), 0);

  if (!selected[n]) {
    selected[n] = true;
    if ((++numSelected == 1) && !numSelectedPages)
      selectedOne = n;
    else
      selectedOne = -1;

    if (update) {
      CClientDC dc(this);
      OnPrepareDC(&dc);
      CRect r;
      doc->getRoom(n).getRect(r);
      dc.LPtoDP(r);
      r.NormalizeRect();
      InvalidateRect(r);
    }
  }
} // end CMapView::selectRoom

//---------------------------------------------------------------------------
// Make sure that a room is visible:
//
// Input:
//   n:  The number of the room to be made visible

void CMapView::makeRoomVisible(RoomNum n)
{
  const CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);

  CRect  clientRect;
  GetClientRect(&clientRect);
  CSize  viewSize(clientRect.right, clientRect.bottom);

  CClientDC  cdc(this);
  OnPrepareDC(&cdc);
  cdc.DPtoLP(&viewSize);

  const CSize   docSize(doc->getDocSize());
  const CPoint  curPos(getScrollPosition());
  CRect roomRect;
  doc->getRoom(n).getRect(roomRect);

  CPoint  newPos(curPos);

  if (roomRect.left - curPos.x < gridX) {
    newPos.x -= curPos.x - roomRect.left + gridX;
    if (newPos.x < 0) newPos.x = 0;
  } else if (roomRect.right - curPos.x + gridX > viewSize.cx) {
    newPos.x += roomRect.right + gridX - viewSize.cx - curPos.x;
    if (newPos.x + viewSize.cx > docSize.cx)
      newPos.x = docSize.cx - viewSize.cx;
  }

  if (curPos.y - roomRect.bottom < gridY) {
    newPos.y += roomRect.bottom + gridY - curPos.y;
    if (newPos.y > 0) newPos.y = 0;
  } else if (curPos.y - roomRect.top + gridY > viewSize.cy) {
    newPos.y += roomRect.top + viewSize.cy - curPos.y - gridY;
    if (viewSize.cy - newPos.y > docSize.cy)
      newPos.y = viewSize.cy - docSize.cy;
  }

  if (newPos != curPos) {
    scrollToPosition(newPos);
    // Must adjust the room position because the DC won't be updated:
    roomRect.OffsetRect(curPos.x - newPos.x, curPos.y - newPos.y);
  }

  // Make sure the floating comments dialog isn't in the way:
  cdc.LPtoDP(&roomRect);
  ClientToScreen(&roomRect);
  roomRect.NormalizeRect();
  static_cast<CMapApp*>(AfxGetApp())->adjustCommentPos(roomRect);
} // end CMapView::makeRoomVisible

//--------------------------------------------------------------------
// Paste a string to the clipboard:
//
// Input:
//   text:  The string to paste
//
// Returns:
//   True:   String was successfully pasted
//   False:  Could not paste (clipboard unavailable or out of memory)

bool CMapView::pasteClipboard(const String& text)
{
  if (!OpenClipboard())
    return false;

  EmptyClipboard();

  // Allocate a global memory object for the text

  HGLOBAL mem = GlobalAlloc(GMEM_DDESHARE, text.length() + sizeof(TCHAR));
  if (mem == NULL) {
    CloseClipboard();
    return false;
  }

  // Lock the handle and copy the text to the buffer

  strcpy(LPTSTR(GlobalLock(mem)), text.c_str());
  GlobalUnlock(mem);

  SetClipboardData(CF_TEXT, mem);  // Place the handle on the clipboard
  CloseClipboard();

  return true;
} // end CMapView::pasteClipboard

//--------------------------------------------------------------------
// Update the selected room's comment from the floating dialog:
//
// Input:
//   comment:  The new room comment (must not be NULL)

void CMapView::setRoomNote(LPCTSTR comment)
{
  ASSERT(selectedOne>=0);
  CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);
  if (doc->getRoom(selectedOne).note != comment) {
    doc->setUndoData(new UndoRoomInfo(*doc, selectedOne));
    doc->setRoomNote(selectedOne, comment);
  }
} // end CMapView::setRoomNote

//--------------------------------------------------------------------
// Set the step size of the scroll bars:

void CMapView::setScrollBars()
{
  CRect  clientRect;
  GetClientRect(&clientRect);
  CSize  viewSize(clientRect.right, clientRect.bottom);

  CClientDC  cdc(this);
  OnPrepareDC(&cdc);
  cdc.DPtoLP(&viewSize);

  viewSize.cx -= gridX/2;
  viewSize.cx -= viewSize.cx % gridX + roomWidth;
  viewSize.cx = max(viewSize.cx, gridX);

  viewSize.cy -= gridY/2;
  viewSize.cy -= viewSize.cy % gridY + roomHeight;
  viewSize.cy = max(viewSize.cy, gridY);

  setPageStep(viewSize);
} // end CMapView::setScrollBars

//--------------------------------------------------------------------
// Wrap room name:
//
// Input:
//   name:
//     The name to wrap
//     Must not contain any CRLF characters
//
// Output:
//   name:  Wrapped to fit the room

void CMapView::wrapRoomName(CString& name)
{
  ASSERT(name.Find(_T('\n')) < 0);

  CDC  pdc;
  const int maxWidth = roomWidth; // FIXME PORT

  if (gueApp()->CreatePrinterDC(pdc)) {
    name.TrimLeft();
    name.TrimRight();

    CFont* oldFont = pdc.SelectObject(&font);
    CSize box(pdc.GetTextExtent(name));

    if (box.cx > maxWidth) {    // then name needs to be wrapped
      LPCTSTR  s = name;  // S is the start of the line we're looking at
      LPCTSTR  e = s;     // E is the end of the line we're looking at
      LPCTSTR  s1 = NULL; // S1 is a valid break site or NULL
      String  wrapped;

      while (*e) {
        while (*e && (*e != _T(' '))) ++e; // Find a space or EOS

        box = pdc.GetTextExtent(s, e - s);
        if (box.cx <= maxWidth) {
          // It fits, make this a possible break site:
          s1 = e;
          while (*e == _T(' ')) ++e; // Find beginning of next word
        } else {
          // It doesn't fit, break as much as we can:
          if (!s1) {
            s1 = e;  // Break after the last character that fits
            while (box.cx > maxWidth)
              box = pdc.GetTextExtent(s, --s1 - s);
          } // end if no space to break at
          wrapped.append(s, s1-s);
          wrapped.append(_T("\r\n"));
          s = s1;
          s1 = NULL;
          while (*s == _T(' ')) ++s;
          e = s;
        }
      } // end while e

      if (*s) wrapped.append(s);

      Strcpy(name, wrapped);
    } // end if name needs wrapping

    pdc.SelectObject(oldFont);
  }
} // end CMapView::wrapRoomName

//--------------------------------------------------------------------
// Zoom the display:
//
// Input:
//   newZoom:  The new zoom factor (in percent)

void CMapView::zoomTo(short newZoom)
{
  if (opInProgress == gmoAddCorner) opInProgress = gmoNone;
  if (newZoom == zoom) return;

  if (opInProgress || scrollDrag || (newZoom < minZoom) || (newZoom > maxZoom))
    MessageBeep(MB_OK);
  else {
    setDisplayZoom(zoom = newZoom);
    if (selectedOne >= 0)
      makeRoomVisible(selectedOne);
    setScrollBars();
    GetParentFrame()->OnUpdateFrameTitle(TRUE);
  }
} // end CMapView::zoomTo

//====================================================================
// Navigation:
//---------------------------------------------------------------------------
// Keyboard navigation:

static const short roomXoffset[] =
{
  0, 0, gridX+roomWidth, -(gridX+roomWidth),
  gridX+roomWidth, -(gridX+roomWidth), gridX+roomWidth, -(gridX+roomWidth)
}; // end roomXoffset  FIXME add NNE, etc.

static const short roomYoffset[] =
{
  gridY+roomHeight, -(gridY+roomHeight), 0, 0,
  gridY+roomHeight, gridY+roomHeight, -(gridY+roomHeight), -(gridY+roomHeight)
}; // end roomYoffset  FIXME add NNE, etc.

// Input:
//   corner:       The direction to move
//   toggleStubs:  TRUE means adding an existing stub should delete it
//
// Input Variables:
//   edgeTmp:      If type1 or type2 is nonzero, use that for a new edge
//   selectedOne:  Must be a valid room number

void CMapView::navigate(RoomCorner corner, bool toggleStubs)
{
  MapEdge e;

  CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);
  ASSERT(selectedOne>=0);

  if (doc->getRoom(selectedOne).isCorner()) {
    MessageBeep(MB_ICONASTERISK);
    return;
  } // can't navigate from corner

  // Look for an existing connection:
  int eNum = doc->findEdge(selectedOne, corner, e);

  if ((eNum >= 0) && !(e.type2 & etUnexplored)) {
    // Found existing connection:
    if (toggleStubs && (e.type2 & etNoExit)) {
      if (edgeTmp.type1 & etNoExit) {
        doc->setUndoData(new UndoDelete(doc->IsModified(), doc->getEdge(eNum)));
      } else {
        doc->setUndoData(new UndoChangeEdge(doc->IsModified(),
                                            doc->getEdge(eNum)));
        edgeTmp.room1 = selectedOne;
        edgeTmp.end1  = corner;
        doc->addEdge(edgeTmp);
      }
      doc->deleteEdge(eNum);
      return;
    } // end if toggling a can't go stub

    if (edgeTmp.type1 || edgeTmp.type2 || // We were trying to add connection
        (e.type1 & etOneWay) || (e.type2 & etNoRoom2)) {
      MessageBeep(MB_ICONASTERISK);
      return;
    }
    if (e.end1 == rcCorner) {
      int r = doc->findOtherEnd(eNum);
      if (r >= 0)
        e.room1 = r;
      else {
        MessageBeep(MB_ICONASTERISK);
        return;
      }
    } // end if this connection leads to a corner

    selectOnlyRoom(e.room1);
    makeRoomVisible(e.room1);
    eNum = -1;                  // Don't delete this edge
  } else { // Try to add a new connection
    if (doc->locked || (corner >= rcNNE)) {
      MessageBeep(MB_ICONASTERISK);
      return;
    }
    const bool  wasModified = doc->IsModified(); // Was it already modified?
    bool  added = false;                         // Was a room added?

    if ((eNum >= 0) && (e.type2 & etUnexplored) &&
        (edgeTmp.type1 & etUnexplored)) {
      if (toggleStubs) { // Remove stub connection:
        doc->setUndoData(new UndoDelete(wasModified, doc->getEdge(eNum)));
        doc->deleteEdge(eNum);
      } else MessageBeep(MB_ICONASTERISK);
      return;
    } // end if adding stub where there's already a stub

    e.room1 = selectedOne;
    e.end1  = corner;
    if (edgeTmp.type1 || edgeTmp.type2) {
      e.type1 = edgeTmp.type1;
      e.type2 = edgeTmp.type2;
    } else
      setEdgeType(e);

    EdgeVec  deletedEdges;

    if (!(e.type1 & etNoRoom2)) {
      CPoint pos(doc->getRoom(selectedOne).pos);
      pos.x += roomXoffset[corner];
      pos.y += roomYoffset[corner];

      int room = doc->findRoom(pos);

      if (room < -1) {          // We're off the grid
        MessageBeep(MB_ICONASTERISK); return;
      } else if (room < 0) {    // We need to add a new room
        room = doc->addRoom(pos);
        added = true;
      } else {                  // Existing room, check for existing connection
        MapEdge oEdge;
        int oeNum = doc->findEdge(room, oppositeCorner[corner], oEdge);
        if (oeNum >= 0) {       // There is an existing connection
          if (oEdge.type2 & etUnexplored) {
            deletedEdges.push_back(doc->getEdge(oeNum));
            doc->deleteEdge(oeNum);
            if (eNum > oeNum) --eNum; // The edge number might have changed
          } else room = -1; // Room has a non-stub connection there
        } // end if room has existing connection
      } // end else room exists, check for existing connection

      if (room < 0) { MessageBeep(MB_ICONASTERISK); return; }

      selectOnlyRoom(room);
      makeRoomVisible(room);

      e.room2 = room;
      e.end2  = oppositeCorner[corner];
    } // end if edge connects two rooms

    if (eNum >= 0) {
      deletedEdges.push_back(doc->getEdge(eNum));
      doc->deleteEdge(eNum);
    }
    doc->addEdge(e);
    if (added) { // Enter room name
      if (gueApp()->editAfterAdd())
        editProperties(epuAddRoomEdge, wasModified,
                       (deletedEdges.size() ? &deletedEdges : NULL));
      else if (deletedEdges.size())
        doc->setUndoData(new UndoChanges(
          wasModified, selectedOne, 1, deletedEdges)
        );
      else
        doc->setUndoData(new UndoAdd(wasModified, selectedOne, 1));
    } else if (deletedEdges.size()) {
      if (deletedEdges.size() == 1)
        doc->setUndoData(new UndoChangeEdge(wasModified, deletedEdges[0]));
      else
        doc->setUndoData(new UndoChanges(wasModified, -1, 1, deletedEdges));
    } else
      doc->setUndoData(new UndoAdd(wasModified));  // Undo new edge
  } // end else add new edge
} // end CMapView::navigate

//--------------------------------------------------------------------
// Fill in the edge type from the navigation box:
//
// If the text in the navigation bar is up, down, in, or out, then set
// the edge type appropriately.  Otherwise, set the edge type to normal.
// Always clears the navigation bar text.
//
// Can only create up-down and in-out passages.
//
// Output:
//   e.type1 and e.type2

void CMapView::setEdgeType(MapEdge& e)
{
  e.type1 = e.type2 = etNormal;

  RoomCorner  corner1, corner2;
  char  initial, separator;
  getNavText(initial, corner1, separator, corner2);

  if (initial && initial != '%' && initial != '>') {
    MessageBeep(MB_ICONASTERISK); return;
  }

  if (separator) {
    switch (corner1) {
     case rcUp:   e.type1 = etUp;   break;
     case rcDown: e.type1 = etDown; break;
     case rcIn:   e.type1 = etIn;   break;
     case rcOut:  e.type1 = etOut;  break;
    } // end switch corner1
    if (separator == '-')
      switch (corner2) {
       case rcUp:   e.type2 = etUp;   break;
       case rcDown: e.type2 = etDown; break;
       case rcIn:   e.type2 = etIn;   break;
       case rcOut:  e.type2 = etOut;  break;
      } // end switch corner2
  } else // standard edge
    switch (corner1) {
     case rcUp:   e.type1 = etUp;   e.type2 = etDown; break;
     case rcDown: e.type1 = etDown; e.type2 = etUp;   break;
     case rcIn:   e.type1 = etIn;   e.type2 = etOut;  break;
     case rcOut:  e.type1 = etOut;  e.type2 = etIn;   break;
    } // end switch

  if (initial == '%')
    e.type1 |= etObstructed;
  if (initial == '>' || separator == '>')
    e.type1 |= etOneWay;
  if (initial == '?' || separator == '?') {
    e.type1 &= ~etSpecial;
    e.type1 |= etUnexplored;
    e.room2 = 0;
    e.end2 = rcSW;
  } else if (initial == '!' || separator == '!') {
    e.type1 &= ~etSpecial;
    e.type1 |= etNoExit;
    e.room2 = 0;
    e.end2 = rcSW;
  }
} // end CMapView::setEdgeType

//--------------------------------------------------------------------
// Start typing in the navigation bar:
//
// Stuffs the given char in the navigation bar and sets focus to there.
//
// Input:
//   c:  The character to put in the navigation bar.

void CMapView::setNavText(char c)
{
  CDialogBar& nb = static_cast<CMainFrame*>(AfxGetMainWnd())->wndNavBar;

  CString  text(c);
  nb.SetDlgItemText(IDC_NAV_DIR, text);
  nb.SetFocus();
  CEdit* dir = static_cast<CEdit*>(nb.GetDlgItem(IDC_NAV_DIR));
  dir->SetSel(1,1);
  dir->SetSel(-1,2);
} // end CMapView::setNavText

void CMapView::OnChar(UINT c, UINT nRepCnt, UINT nFlags)
{
  if (((c >= 'A') && (c <= 'Z')) ||
      ((c >= 'a') && (c <= 'z')) ||
      (c == '-') || (c == '!') || (c == '?') || (c == '/') ||
      (c == '>') || (c == '.') || (c == '~') || (c == '`') ||
      (c == '%'))
    setNavText(c);
} // end CMapView::OnChar

//--------------------------------------------------------------------
// Parse the text in the navigation bar:
//
// Input:
//   mustBeMove: (default FALSE)
//     If TRUE, beep and don't clear text if not a direction.
//     If FALSE, clear text and don't beep even if it's meaningless.
//
// Output:
//   initial:
//     The non-alphabetic character beginning the text (0 if none)
//   dir1:
//     The direction indicated by the first direction word
//     rcNone if no direction words
//   separator:
//     The non-alphabetic character following the first direction word
//     0 if none
//     Must be either '-', '>', or '?'
//   dir2:
//     The direction indicated by the second direction word
//     rcNone if no second direction

void CMapView::getNavText(char& initial, RoomCorner& dir1,
                          char& separator, RoomCorner& dir2,
                          bool mustBeMove)
{
  CDialogBar& nb = static_cast<CMainFrame*>(AfxGetMainWnd())->wndNavBar;
  CString  text;

  nb.GetDlgItemText(IDC_NAV_DIR, text);
  text.MakeLower();

  dir1    = dir2      = rcNone;
  initial = separator = '\0';

  if (!text.IsEmpty() && (text[0] < 'a' || text[0] > 'z')) {
    initial = text[0];
    text = text.Mid(1);
  }

  if (initial == '-') {
    initial = '\0';
    separator = '-';
  } else if (!text.IsEmpty() && (text[0] == '-' || text[0] == '>')) {
    separator = text[0];
    text = text.Mid(1);
  } else {
    dir1 = parseDirection(text);
    if (!text.IsEmpty() && (text[0]=='-' || text[0]=='>' || text[0]=='?')) {
      separator = text[0];
      text = text.Mid(1);
    }
  }

  if (separator)
    dir2 = parseDirection(text);

  if (mustBeMove && (dir1 == rcNone || separator == '-' ||
                     ((initial || separator) && (dir1 >= rcNumCorners))))
    MessageBeep(MB_ICONASTERISK);
  else {
    nb.SetDlgItemText(IDC_NAV_DIR, _T(""));
    SetFocus();
  }
} // end CMapView::getNavText

//--------------------------------------------------------------------
// Handle the Go button in the navigation bar:

void CMapView::OnNavGo()
{
  if (selectedOne < 0) { MessageBeep(MB_ICONASTERISK); return; }

  RoomCorner  corner, junk2;
  char  initial, separator;
  getNavText(initial, corner, separator, junk2, true);

  if (separator == '-' || junk2 != rcNone) {
    MessageBeep(MB_ICONASTERISK);
    return;
  }

  if (corner != rcNone) {
    edgeTmp.type1 = edgeTmp.type2 = etNormal;
    if (initial == '%')
      edgeTmp.type1 = etObstructed;
    if (initial == '>' || separator == '>')
      edgeTmp.type1 |= etOneWay;
    if (initial == '?' || separator == '?')
      edgeTmp.type1 = etUnexplored;
    else if (initial == '!' || separator == '!')
      edgeTmp.type1 = etNoExit;
    navigate(corner);
  }
} // end CMapView::OnNavGo

//--------------------------------------------------------------------
// Handle the direction keys on the numeric keypad:
//
// Input:
//   cmd:  Which command was executed

void CMapView::OnNavGoDir(UINT cmd)
{
  ASSERT(cmd >= ID_NAV_GO_NORTH && cmd <= ID_NAV_STUB2_OUT);

  if (selectedOne >= 0) {
    SetFocus();
    edgeTmp.type1 = edgeTmp.type2 = etNormal;
    bool toggleStubs = false;
    if (cmd >= ID_NAV_STUB1_NORTH) {
      if (cmd >= ID_NAV_STUB2_NORTH) { // Stub 2
        edgeTmp.type1 = (gueApp()->stub1Unexplored() ? etNoExit : etUnexplored);
        cmd -= ID_NAV_STUB2_NORTH - ID_NAV_GO_NORTH;
      } else {                         // Stub 1
        edgeTmp.type1 = (gueApp()->stub1Unexplored() ? etUnexplored : etNoExit);
        cmd -= ID_NAV_STUB1_NORTH - ID_NAV_GO_NORTH;
      }
      toggleStubs = true;
    } // end if create/remove stub connection
    navigate(RoomCorner(cmd - ID_NAV_GO_NORTH), toggleStubs);
  } else
    MessageBeep(MB_ICONASTERISK);
} // end CMapView::OnNavGoDir

/////////////////////////////////////////////////////////////////////////////
// CMapView message handlers
//---------------------------------------------------------------------------
// Update the comments dialog when the view is activated:

void CMapView::OnActivateView(BOOL bActivate, CView* pActivateView,
                              CView* pDeactiveView)
{
  if (bActivate) selectDone();  // Update comments dialog

  CScrollZoomView::OnActivateView(bActivate, pActivateView, pDeactiveView);
} // end CMapView::OnActivateView

//--------------------------------------------------------------------
void CMapView::OnEditAddCorner()
{
  ASSERT(opInProgress==gmoAddCorner);
  ASSERT(selectedOne>=0);

  CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);
  int room = doc->addCorner(selectedOne, iTmp);
  if (room >= 0)
    selectOnlyRoom(room);
  else
    MessageBeep(MB_ICONASTERISK);

  opInProgress = gmoNone;
} // end CMapView::OnEditAddCorner

//--------------------------------------------------------------------
void CMapView::OnEditAddPage()
{
  CMapDoc *const  doc = GetDocument();
  ASSERT_VALID(doc);

  const int  p = doc->getPages().size();
  if (p < maxPages) {
    MapPage  page;
    page.pos = scrollDragStart;
    page.pos.x -= page.pos.x % gridX;
    page.pos.y -= page.pos.y % gridY;
    doc->setUndoData(new UndoAdd(doc->IsModified(), p));
    showPages = true;
    doc->addPage(p, page);
    clearSelection();
    selectPage(p);
  }
} // end CMapView::OnEditAddPage

//--------------------------------------------------------------------
void CMapView::OnEditAddRoom()
{
  addRoom(scrollDragStart);
} // end CMapView::OnEditAddRoom

//--------------------------------------------------------------------
// Delete the selected rooms:

void CMapView::OnEditClear()
{
  deleteSelection();
} // end CMapView::OnEditClear

// Input:
//   removeCorner: (default true)
//     When TRUE, if a single corner is selected, remove it but leave
//     the connection.  When FALSE, the connection is always removed.

void CMapView::deleteSelection(bool removeCorner)
{
  if (numSelected || numSelectedPages) {
    CMapDoc*  doc = GetDocument();
    ASSERT_VALID(doc);

    if (doc->locked || (numSelectedPages == doc->getPages().size())) {
      MessageBeep(MB_ICONASTERISK);
      return;
    }

    if (removeCorner &&
        (selectedOne >= 0) && doc->getRoom(selectedOne).isCorner()) {
      // Replace this corner with a single edge:
      doc->deleteCorner(selectedOne);
      clearSelection(false);
      selectDone();
      return;
    } else {
      // Select all corners if one corner or one end is selected:
      const EdgeVec&  edge = doc->getEdges();
      int wereSelected = 0;
      while (wereSelected != numSelected) {
        wereSelected = numSelected;
        for (EdgeConstItr e = edge.begin(); e != edge.end(); ++e) {
          if ((e->end1 == rcCorner) && selected[e->room2] &&
              !selected[e->room1])
            selectRoom(e->room1, false);
          else if ((e->end2==rcCorner) && selected[e->room1] &&
                   !selected[e->room2])
            selectRoom(e->room2, false);
        }
      } // end while something changed on this pass
    } // end else not just a corner selected

    UndoDelete*  undoRec = new UndoDelete(*doc, numSelected, selected,
                                          numSelectedPages, selectedPage);
    opInProgress = gmoDeleting; // Don't clear selection after first room
    int pos = numSelected;
    for (int i = doc->getRoomCount()-1; i >= 0; --i)
      if (selected[i])
        undoRec->addRoom(--pos, i, doc->extractRoom(i));
    for (i = doc->getPages().size()-1; i >= 0; --i)
      if (selectedPage[i])
        doc->deletePage(i);
    selectedPage.resize(doc->getPages().size(), 0);
    opInProgress = gmoNone;
    doc->setUndoData(undoRec);

    clearSelection(false);
    selectDone();
  }
} // end CMapView::deleteSelection

//--------------------------------------------------------------------
// Cut, Copy & Paste:

void CMapView::OnEditCopy()
{
  if (numSelected || numSelectedPages) {
    CMapDoc*  doc = GetDocument();
    ASSERT_VALID(doc);

    // Deselect all corner rooms:
    for (int i = doc->getRoomCount()-1; i >= 0; --i)
      if (selected[i] && doc->getRoom(i).isCorner())
        deselectRoom(i);

    if (!numSelected && !numSelectedPages) {
      selectDone();
      MessageBeep(MB_ICONASTERISK);
      return;
    } // end if only corners were selected

    // Select corners between selected rooms:
    const EdgeVec& edges = doc->getEdges();
    for (EdgeConstItr e = edges.begin(); e != edges.end(); ++e)
      if ((selected[e->room1] && (e->end2 == rcCorner) &&
           (e->end1 != rcCorner) && !selected[e->room2])   ||
          (selected[e->room2] && (e->end1 == rcCorner) &&
           (e->end2 != rcCorner) && !selected[e->room1])) {
        RoomNumVec  cornerRooms;
        cornerRooms.reserve(16); // Should be plenty
        int  otherEnd = doc->findOtherEnd(e, &cornerRooms);
        if ((otherEnd >= 0) && selected[otherEnd])
          for (RNConstItr rn = cornerRooms.begin();
               rn != cornerRooms.end(); ++rn)
            selectRoom(*rn);
      } // end if edge connects selected room to unselected corner

    selectDone();

    // Copy selected rooms and pages:
    gueApp()->setClipboard(
      new RoomScrap(*doc, numSelected, selected, numSelectedPages, selectedPage)
    );
  }
} // end CMapView::OnEditCopy

void CMapView::OnEditCut()
{
  OnEditCopy();
  deleteSelection();
} // end CMapView::OnEditCut

void CMapView::OnEditPaste()
{
  const RoomScrap*  scrap = static_cast<CMapApp*>(AfxGetApp())->getClipboard();

  if (scrap) {
    CMapDoc*  doc = GetDocument();
    ASSERT_VALID(doc);

    int i = doc->getRoomCount();

    doc->setUndoData(scrap->paste(*doc));
    const int  numRooms = doc->getRoomCount();

    if (numRooms > i) makeRoomVisible(i);

    clearSelection(true);
    while (i < numRooms)
      selectRoom(i++,true);
    selectDone();
  } // end if scrap
} // end CMapView::OnEditPaste

//--------------------------------------------------------------------
void CMapView::OnEditPaginate()
{
  showPages = true;
  GetDocument()->layoutPages();
} // end CMapView::OnEditPaginate

//--------------------------------------------------------------------
// Handle Edit Properties:

void CMapView::OnEditProperties(UINT cmd)
{
  editProperties((cmd==ID_EDIT_MAP_PROPERTIES) ? epuMapInfo : epuRoomInfo);
} // end CMapView::OnEditProperties

//--------------------------------------------------------------------
// Bring up the Properties dialog:
//
// Input:
//   undoType:
//     The type of action we should create an undo record for
//       epuRoomInfo:     Just changing room info
//       epuAddRoom:      Adding a new room
//       epuAddRoomEdge:  Adding a new room and a new edge
//   wasModified:
//     The modification state of the document
//       (Necessary only if undoType is not epuRoomInfo)
//   edges: (default NULL)
//     A pointer to the edges that were deleted before adding this room
//       NULL if no edges were deleted
//       Must be NULL unless undoType is epuAddRoomEdge
//       The vector is emptied but not deleted

void CMapView::editProperties(EditPropUndo undoType, bool wasModified,
                              EdgeVec* edges)
{
  bool forceMap = false;

  if (undoType == epuMapInfo) {
    undoType = epuRoomInfo;
    forceMap = true;
  }

  CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);

  CPropertySheet dlg(_T("Properties"));
  CRoomPropPage  room;
  UndoRoomInfo*  undoRoom = NULL;
  const bool  roomTab =
    (selectedOne >= 0) && !doc->getRoom(selectedOne).isCorner();

  CMapPropPage   map;
  Strcpy(map.name, doc->getName());
  Strcpy(map.note, doc->getNote());
  map.corners = showCorners;
  map.locked = doc->locked;
  map.grid   = showGrid;
  map.pages  = showPages;
  dlg.AddPage(&map);

  if (roomTab) {
    VERIFY(undoRoom = new UndoRoomInfo(*doc, selectedOne));
    room.border = BOOL((undoRoom->getRoom().flags & rfBorder) != 0);
    Strcpy(room.name, undoRoom->getRoom().name);
    Strcpy(room.note, undoRoom->getRoom().note);
    dlg.AddPage(&room);
    static_cast<CMapApp*>(AfxGetApp())->setComment(this,NULL,false);
    if (!forceMap)
      dlg.SetActivePage(&room);
  }

  if (dlg.DoModal() == IDOK) {
    if (roomTab) {
      bool  changed = false;
      if (room.border != (0 != (undoRoom->getRoom().flags & rfBorder))) {
        doc->setRoomFlags(selectedOne,(room.border ? rfBorder : 0),rfBorder);
        changed = true;
      }
      if (room.name != undoRoom->getRoom().name.c_str()) {
        if (room.name.Find('\n') < 0)
          wrapRoomName(room.name);
        doc->setRoomName(selectedOne, room.name);
        changed = true;
      } // end if name changed
      room.note.TrimRight();
      if (room.note != undoRoom->getRoom().note.c_str()) {
        doc->setRoomNote(selectedOne, room.note);
        changed = true;
      } // end if note changed
      if (changed && (undoType == epuRoomInfo)) {
        doc->setUndoData(undoRoom);
        undoRoom = NULL; // undoRoom now belongs to the document
      } // end if not a new room
    } // end if room properties page
    if (undoType != epuRoomInfo) {
      ASSERT(selectedOne>=0);
      if (edges)
        doc->setUndoData(new UndoChanges(wasModified, selectedOne, 1, *edges));
      else
        doc->setUndoData(new UndoAdd(wasModified, selectedOne,
                                     (undoType == epuAddRoomEdge ? 1 : 0)));
    } // end if adding room
    doc->setName(map.name);
    doc->setNote(map.note);
    if (doc->locked != bool(map.locked))
      doc->OnNavigationMode();
    if ((showCorners != bool(map.corners)) ||
        (showGrid != bool(map.grid)) ||
        (showPages != bool(map.pages))) {
      showCorners = map.corners != 0;
      showGrid    = map.grid    != 0;
      showPages   = map.pages   != 0;
      if (!showPages) deselectPages(false);
      InvalidateRect(NULL);
    }
  } // end if OK
  else if (undoType != epuRoomInfo) {
    // Cancel adding the room:
    ASSERT(selectedOne>=0);
    const RoomNum  roomNum = selectedOne;
    clearSelection();
    if (undoType == epuAddRoomEdge) {
      const MapEdge& edge = doc->getEdge(doc->getEdgeCount()-1);
      selectOnlyRoom((edge.room1 == roomNum) ? edge.room2 : edge.room1);
    }
    opInProgress = gmoDeleting; // Don't clear selection
    doc->deleteRoom(roomNum);
    opInProgress = gmoNone;
    if (edges) {
      doc->addEdges(edges->size());
      for (EdgeConstItr e = edges->begin(); e != edges->end(); ++e)
        doc->addEdge(*e);
    }
    doc->SetModifiedFlag(wasModified);
  } // end if canceled adding room

  selectDone();
  delete undoRoom;
} // end CMapView::editProperties

//---------------------------------------------------------------------------
// Select all rooms:

void CMapView::OnEditSelectAll()
{
  const CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);

  for (int i = doc->getRoomCount()-1; i >= 0; --i)
    selectRoom(i);

  selectDone();
} // end CMapView::OnEditSelectAll

//---------------------------------------------------------------------------
// Select connected rooms:

void CMapView::OnEditSelectConnected()
{
  const CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);

  const int   numEdges = doc->getEdgeCount();
  const EdgeVec&  edge = doc->getEdges();

  int wasSelected = 0;
  while (wasSelected != numSelected) {
    wasSelected = numSelected;
    for (int i = 0; i < numEdges; i++)
      if ((edge[i].type1 & etNoRoom2) == 0) {
        if      (selected[edge[i].room1])  selectRoom(edge[i].room2);
        else if (selected[edge[i].room2])  selectRoom(edge[i].room1);
      } // end if edge connects two rooms
  } // end while more rooms were selected on this pass

  selectDone();
} // end CMapView::OnEditSelectConnected

//---------------------------------------------------------------------------
void CMapView::OnInitialUpdate()
{
  const CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);

  selectedPage.resize(doc->getPages().size(), 0);
  clearSelection(false);

  // Set the document size:
  init(4*gridX, doc->getDocSize(), CSize(gridX, gridY));
  //  CClientDC dc(this);
  //  OnPrepareDC(&dc);
  LOGFONT lf;
  memset(&lf, 0, sizeof(lf));
  lf.lfHeight = 62; // PORT
  strcpy(lf.lfFaceName, "Arial");
  font.CreateFontIndirect(&lf);
  if (doc->locked)
    OnUpdate(NULL, dupNavigationMode, NULL);
  CScrollZoomView::OnInitialUpdate();
  setScrollBars();              // Now fix the scroll bars
} // end OnInitialUpdate

//--------------------------------------------------------------------
void CMapView::OnKeyDown(UINT c, UINT nRepCnt, UINT nFlags)
{
  bool ctrl = (GetKeyState(VK_CONTROL) < 0);

  switch (c) {
   case VK_LEFT:  OnHScroll(ctrl ? SB_PAGELEFT  : SB_LINELEFT, 0,NULL); break;
   case VK_RIGHT: OnHScroll(ctrl ? SB_PAGERIGHT : SB_LINERIGHT,0,NULL); break;
   case VK_UP:    OnVScroll(ctrl ? SB_PAGEUP    : SB_LINEUP,   0,NULL); break;
   case VK_DOWN:  OnVScroll(ctrl ? SB_PAGEDOWN  : SB_LINEDOWN, 0,NULL); break;
   case VK_PRIOR: OnVScroll(ctrl ? SB_TOP       : SB_PAGEUP,   0,NULL); break;
   case VK_NEXT:  OnVScroll(ctrl ? SB_BOTTOM    : SB_PAGEDOWN, 0,NULL); break;
   case VK_TAB:   gueApp()->editComment();                              break;
   default:
    if ((GetKeyState(VK_SHIFT) >= 0) && (c >= '0') && (c <= '9')) {
      if (!ctrl && (c < '2'))
        zoomTo(100);
      else if (ctrl && (c == '0'))
        zoomTo(200);
      else
        zoomTo(10 * (c - '0') + (ctrl ? 100 : 0));
    } // end if number key (not shifted)
    break;
  } // end switch c
} // end CMapView::OnKeyDown

//--------------------------------------------------------------------
void CMapView::OnLButtonDblClk(UINT nFlags, CPoint point)
{
  if (scrollDrag || scrollDragTimer) return;

  opInProgress = gmoNone;
  if (GetCapture() == this)
    ReleaseCapture();

  CClientDC dc(this);
  OnPrepareDC(&dc);
  dc.DPtoLP(&point);

  CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);

  const int i = doc->roomHit(point);
  if (i >= 0) {
    if (!doc->getRoom(i).isCorner()) {
      selectOnlyRoom(i);
      editProperties(epuRoomInfo);
    }
  } else
    addRoom(point);
} // end CMapView::OnLButtonDblClk

//--------------------------------------------------------------------
void CMapView::OnLButtonDown(UINT nFlags, CPoint point)
{
  if (scrollDrag || scrollDragTimer) return;

  CClientDC dc(this);
  OnPrepareDC(&dc);
  dc.DPtoLP(&point);
  CMapDoc*  doc = GetDocument();
  ASSERT_VALID(doc);

  RoomCorner  corner;
  int room = doc->roomHit(point,&corner);

  if (room < 0) {
    if (doc->locked) return; // Must have 1 room selected during navigation
    if (showPages && ((room = doc->pageBorderHit(point)) >= 0)) {
      iTmp  = room;
      if (nFlags & MK_CONTROL) {
        opInProgress = gmoControlDown;
        bTmp  = selectedPage[room];
        b2Tmp = true;           // In page
      } else {
        if (nFlags & MK_SHIFT) opInProgress = gmoShiftDown;
        else opInProgress = gmoNone;
        clearSelection();
      }
      selectPage(room);
      selectDone();
      if (opInProgress) {
        SetCapture();
        p1Tmp.x = point.x / gridX;
        p1Tmp.y = point.y / gridY;
        p2Tmp = doc->getPage(room).pos;
      }
    } // end if in a page border
    else {
      SetCapture();
      opInProgress = gmoSelectBox;
      if (nFlags & MK_CONTROL) {
        // If Ctrl pressed, don't clear selection:
        for (ByteItr b = selected.begin(); b != selected.end(); ++b)
          if (*b) *b = 2;         // Mark the rooms that should stay selected
        for (b = selectedPage.begin(); b != selectedPage.end(); ++b)
          if (*b) *b = 2;         // Mark the pages that should stay selected
      } else {
        clearSelection();
        selectDone();
      } // end else Ctrl is not pressed
      p1Tmp = point;
      p2Tmp = point;
      rTmp.left = rTmp.right = point.x;
      rTmp.bottom = rTmp.top = point.y;
    } // end else not in page border
  } // end if outside room
  else if (doc->locked) {
    ASSERT(selectedOne >= 0);
    if ((room != selectedOne) && !doc->getRoom(room).isCorner()) {
      String path;
      doc->shortestPath(path, selectedOne, room);
      if (path.empty())
        path = _T("Sorry, I don't know how to get there.");
      else if (path.find('.') != String::npos) {
        const CMapApp* app = static_cast<CMapApp*>(AfxGetApp());
        if (app->navigationCopy()) {
          if (app->navigationCRLF()) path += _T("\r\n");
          pasteClipboard(path);
        }
      } // end else if path has multiple steps
      static_cast<CMainFrame*>(AfxGetMainWnd())->setStatusBar(path.c_str());
      selectOnlyRoom(room);
    } // end if moving to new room in navigation mode
  } else if (nFlags & MK_CONTROL) {
    SetCapture();
    opInProgress = gmoControlDown;
    ASSERT(room < selected.size());
    bTmp = selected[room];
    b2Tmp = false;              // In room
    iTmp = room;
    selectRoom(room);
    selectDone();
    p1Tmp.x = point.x / gridX;
    p1Tmp.y = point.y / gridY;
  } else if (nFlags & MK_SHIFT) {
    SetCapture();
    opInProgress = gmoShiftDown;
    selectOnlyRoom(room);
    iTmp = -1;                  // Not a page
    p1Tmp.x = point.x / gridX;
    p1Tmp.y = point.y / gridY;
    p2Tmp = doc->getRoom(room).pos;
  } else {
    selectOnlyRoom(room);
    if (doc->getRoom(room).isCorner()) {
      SetCapture();
      opInProgress = gmoControlDown;
      bTmp = b2Tmp = false;     // Leave the corner selected; in room
      p1Tmp.x = point.x / gridX;
      p1Tmp.y = point.y / gridY;
    } else if (corner != rcNone) {
      SetCapture();
      opInProgress = gmoAddEdge;
      bTmp = doc->IsModified();
      iTmp = doc->findEdge(room, corner, edgeTmp);
      if (iTmp >= 0) {
        opInProgress = gmoChangeEdge;
        rTmp.SetRect(point.x,point.y,point.x,point.y);
        rTmp.InflateRect(29,29); // PORT
      }
      if ((iTmp < 0) || (edgeTmp.type2 & etNoRoom2)) {
        edgeTmp.room1 = room;
        edgeTmp.end1  = corner;
        setEdgeType(edgeTmp);
      }
      edgeTmp.room2 = 0;
      edgeTmp.end2  = rcSW;
      doc->getEdgePoints(edgeTmp, p1Tmp.x, p1Tmp.y, p2Tmp.x, p2Tmp.y);
      p2Tmp = p1Tmp;
      OnPaint();                // Update the view immediately
      if (opInProgress == gmoAddEdge) {
        dc.SetROP2(R2_NOT);
        CPen penEdge;
        if (!penEdge.CreatePen(PS_SOLID, penEdgeWidth, RGB(0,0,0))) // PORT
          return;
        CPen*  oldPen = dc.SelectObject(&penEdge);
        dc.MoveTo(p1Tmp);
        dc.LineTo(p2Tmp);
        dc.SelectObject(oldPen);
      } // end if adding new edge
    } // end if mouse on corner
  } // end else not Control
} // end CMapView::OnLButtonDown

//--------------------------------------------------------------------
void CMapView::OnMouseMove(UINT nFlags, CPoint point)
{
  if (scrollDragTimer) {
    KillTimer(scrollDragTimer);
    scrollDragTimer = 0;
    SetCapture();
    scrollDrag = true;
    ::SetCursor(handCursor);
  }

  if (GetCapture() != this)
    return;

  CClientDC dc(this);
  OnPrepareDC(&dc);
  dc.DPtoLP(&point);

//....................................................................
  if (scrollDrag) {
    MSG  msg;
    CPoint  p(offset);
    p.Offset(scrollDragStart - point);
    if (!calcScrollPosition(p) ||
        PeekMessage(&msg, m_hWnd, WM_MOUSEMOVE, WM_MOUSEMOVE, PM_NOREMOVE))
      return;
    if ((opInProgress == gmoAddEdge) ||
        (opInProgress == gmoSelectBox) ||
        (opInProgress >= gmoDrag)) {
      // We must handle these specially to avoid display artifacts:
      dc.SetROP2(R2_NOT);
      CGdiObject*  old = NULL;
      CPen   pen;
      CBrush brush;
      switch (opInProgress) {
       case gmoAddEdge:
        if (!pen.CreatePen(PS_SOLID, penEdgeWidth, RGB(0,0,0))) // PORT
          return;
        old = dc.SelectObject(&pen);  dc.MoveTo(p1Tmp);  dc.LineTo(p2Tmp);
        break;

       case gmoSelectBox:
        old = dc.SelectStockObject(HOLLOW_BRUSH);  dc.Rectangle(rTmp);  break;

       default:  drawSelected(&dc);  break; // Must be gmoDrag
      } // end switch opInProgress

      scrollToPosition(p);
      OnPaint();
      OnPrepareDC(&dc);         // Update with new offset

      switch (opInProgress) {
       case gmoAddEdge:    dc.MoveTo(p1Tmp);  dc.LineTo(p2Tmp);  break;
       case gmoSelectBox:  dc.Rectangle(rTmp);                   break;
       default: /* drag*/  drawSelected(&dc);                    break;
      } // end switch opInProgress
      if (old) dc.SelectObject(old);
    } // end if opInProgress requires drawing during scroll
    else
      scrollToPosition(p);
    return;
  } // end if scrollDrag

//....................................................................
  if (opInProgress == gmoChangeEdge) {
    if (rTmp.PtInRect(point))
      return;

    CMapDoc*  doc = GetDocument();
    ASSERT_VALID(doc);

    e2Tmp = doc->getEdge(iTmp); // Save the current edge data
    doc->deleteEdge(iTmp);      // Delete the old edge
    OnPaint();                  // Update the view immediately
    opInProgress = gmoAddEdge;  // Now fall through to gmoAddEdge
  } // end if gmoChangeEdge

  if (opInProgress == gmoAddEdge) {
    CMapDoc*  doc = GetDocument();
    ASSERT_VALID(doc);

    dc.SetROP2(R2_NOT);
    CPen penEdge;
    if (!penEdge.CreatePen(PS_SOLID, penEdgeWidth, RGB(0,0,0))) // PORT
      return;
    CPen*  oldPen = dc.SelectObject(&penEdge);
    dc.MoveTo(p1Tmp);
    dc.LineTo(p2Tmp);

    RoomCorner  corner;
    int room = doc->roomHit(point, &corner);
    if (corner != rcNone && room != edgeTmp.room1) {
      edgeTmp.room2 = room;
      edgeTmp.end2  = corner;
      doc->getEdgePoints(edgeTmp, p1Tmp.x, p1Tmp.y, p2Tmp.x, p2Tmp.y);
    } else {
      p2Tmp = point;
    }
    dc.MoveTo(p1Tmp);
    dc.LineTo(p2Tmp);
    dc.SelectObject(oldPen);
  } // end if gmoAddEdge
  else if (opInProgress == gmoSelectBox) {
    if (p2Tmp != point) {
      p2Tmp = point;
      dc.SetROP2(R2_NOT);
      CBrush* oldBrush =
        static_cast<CBrush*>(dc.SelectStockObject(HOLLOW_BRUSH));
      dc.Rectangle(rTmp);

      rTmp = CRect(p1Tmp,point);
      rTmp.NormalizeRect();
      CMapDoc*  doc = GetDocument();
      ASSERT_VALID(doc);
      RoomConstItr  room = doc->getRooms().begin();
      CRect intersect,r;
      for (int i = doc->getRoomCount()-1; i >= 0; i--) {
        if (selected[i] == 2)
          continue; // Room was already selected before selection box
        room[i]->getRect(r);
        intersect.IntersectRect(rTmp,r);
        if ((selected[i] != 0) != bool(r == intersect)) {
          selected[i] = !selected[i];
          numSelected += selected[i] ? 1 : -1;
          r.DeflateRect(5,5); // PORT
          dc.LPtoDP(&r);
          r.NormalizeRect();
          InvalidateRect(&r);
        }
      }
      if (showPages) for (i = doc->getPages().size()-1; i >= 0; --i) {
        if (selectedPage[i] == 2)
          continue; // Page was already selected before selection box
        doc->getPageRect(i, r);
        intersect.IntersectRect(rTmp,r);
        if ((selectedPage[i] != 0) != bool(r == intersect)) {
          selectedPage[i]  = !selectedPage[i];
          numSelectedPages += selectedPage[i] ? 1 : -1;
          r.DeflateRect(5,5); // PORT
          dc.LPtoDP(&r);
          r.NormalizeRect();
          InvalidateRect(&r);
        }
      }
      if ((numSelected == 1) && !numSelectedPages)
        for (selectedOne=0; !selected[selectedOne]; selectedOne++)
          ;                     // Find the selected room
      else
        selectedOne = -1;
      selectDone();
      if (GetUpdateRect(NULL)) OnPaint();
      dc.Rectangle(rTmp);
      dc.SelectObject(oldBrush);
    } // end if cursor moved
  } // end else if gmoSelectBox
  else if (opInProgress == gmoControlDown) {
    p2Tmp.x = gridX * (point.x / gridX - p1Tmp.x);
    p2Tmp.y = gridY * (point.y / gridY - p1Tmp.y);
    if (p2Tmp.x || p2Tmp.y) {
      opInProgress = gmoControlDrag;
      computeSelectedRect(rTmp);
      drawBackground(&dc);
      dc.SetROP2(R2_NOT);
      drawSelected(&dc);
    }
  } // end if gmoControlDown
  else if (opInProgress == gmoShiftDown) {
    LONG ox = point.x / gridX;
    LONG oy = point.y / gridY;
    if      (ox < p1Tmp.x) opInProgress = gmoShiftDragLeft;
    else if (ox > p1Tmp.x) opInProgress = gmoShiftDragRight;
    else if (oy > p1Tmp.y) opInProgress = gmoShiftDragUp;
    else if (oy < p1Tmp.y) opInProgress = gmoShiftDragDown;
    else return;

    CMapDoc*  doc = GetDocument();
    ASSERT_VALID(doc);

    if (iTmp >= 0) {
      ASSERT(!numSelected && (numSelectedPages == 1));
      doc->getPageRect(iTmp, rTmp);
      switch (opInProgress) {
       case gmoShiftDragUp:   p2Tmp.y = rTmp.top;                       break;
       case gmoShiftDragLeft: p2Tmp.x = rTmp.right - roomWidth + gridX; break;
      }
    } // end if dragging a page

    RoomConstItr  room = doc->getRooms().begin();
    const int  roomCount = doc->getRoomCount();
    for (int i=0; i < roomCount; i++)
      if ((opInProgress==gmoShiftDragLeft  && room[i]->pos.x <= p2Tmp.x) ||
          (opInProgress==gmoShiftDragRight && room[i]->pos.x >= p2Tmp.x) ||
          (opInProgress==gmoShiftDragUp    && room[i]->pos.y >= p2Tmp.y) ||
          (opInProgress==gmoShiftDragDown  && room[i]->pos.y <= p2Tmp.y))
        selectRoom(i,false);
    if (iTmp < 0) { // shift-dragging a room
      if      (opInProgress == gmoShiftDragLeft) p2Tmp.x += roomWidth;
      else if (opInProgress == gmoShiftDragDown) p2Tmp.y += roomHeight;
    } else {        // shift-dragging a page
      if      (opInProgress == gmoShiftDragLeft)  p2Tmp.x += gridX;
      else if (opInProgress == gmoShiftDragRight) p2Tmp.x -= gridX;
      else if (opInProgress == gmoShiftDragUp)    p2Tmp.y -= gridY;
      else if (opInProgress == gmoShiftDragDown)  p2Tmp.y += gridY;
    } // end else shift-dragging a page
    for (i = doc->getPages().size()-1; i >= 0; --i) {
      doc->getPageRect(i, rTmp);
      if ((opInProgress==gmoShiftDragLeft  && rTmp.right  < p2Tmp.x) ||
          (opInProgress==gmoShiftDragRight && rTmp.left   > p2Tmp.x) ||
          (opInProgress==gmoShiftDragUp    && rTmp.top    > p2Tmp.y) ||
          (opInProgress==gmoShiftDragDown  && rTmp.bottom < p2Tmp.y))
        selectPage(i,false);
    }
    selectDone();

    p2Tmp.x = gridX * (point.x / gridX - p1Tmp.x);
    p2Tmp.y = gridY * (point.y / gridY - p1Tmp.y);
    computeSelectedRect(rTmp);
    drawBackground(&dc);
    dc.SetROP2(R2_NOT);
    drawSelected(&dc);
  } // end else if gmoShiftDown
  else if (opInProgress >= gmoDrag) {
    MSG  msg;
    if (PeekMessage(&msg, m_hWnd, WM_MOUSEMOVE, WM_MOUSEMOVE, PM_NOREMOVE))
      return; // The mouse has already been moved again
    CPoint p(gridX * (point.x / gridX - p1Tmp.x),
             gridY * (point.y / gridY - p1Tmp.y));
    if (p != p2Tmp) {
      adjustOffset(p);
      if (p == p2Tmp) return;
      dc.SetROP2(R2_NOT);
      drawSelected(&dc);
      p2Tmp = p;
      drawSelected(&dc);
    }
  }
} // end CMapView::OnMouseMove

//--------------------------------------------------------------------
void CMapView::OnLButtonUp(UINT nFlags, CPoint point)
{
  if (GetCapture() != this) {
    if (selectedOne >= 0) makeRoomVisible(selectedOne);
    return;
  }
  if (!scrollDrag)
    ReleaseCapture();

  CClientDC dc(this);
  OnPrepareDC(&dc);
  dc.DPtoLP(&point);

  CMapDoc* doc = GetDocument();
  ASSERT_VALID(doc);

//....................................................................
  if (opInProgress == gmoAddEdge) {
    // Erase the rubber band:
    dc.SetROP2(R2_NOT);
    CPen penEdge;
    if (!penEdge.CreatePen(PS_SOLID, penEdgeWidth, RGB(0,0,0))) // PORT
      return;
    CPen*  oldPen = dc.SelectObject(&penEdge);
    dc.MoveTo(p1Tmp);
    dc.LineTo(p2Tmp);
    dc.SelectObject(oldPen);

    // Find out where the edge goes:
    RoomCorner  corner;
    int room = doc->roomHit(point, &corner);
    if (corner != rcNone && room != edgeTmp.room1) {// The edge goes somewhere
      // If there's a stub on the other end, delete it:
      EdgeVec  deletedEdges;
      MapEdge  oldEdge;
      int oldEdgeNum = doc->findEdge(room,corner,oldEdge);
      if (oldEdgeNum >= 0 && (oldEdge.type2 & etUnexplored)) {
        deletedEdges.push_back(doc->getEdge(oldEdgeNum));
        doc->deleteEdge(oldEdgeNum);
      }
      // Add the new or changed edge:
      if (edgeTmp.type2 & etOneWay) {
        edgeTmp.room2 = edgeTmp.room1;
        edgeTmp.end2  = edgeTmp.end1;
        edgeTmp.type1 = edgeTmp.type2 | (edgeTmp.type1 & etObstructed);
        edgeTmp.type2 = etNormal;
        edgeTmp.room1 = room;
        edgeTmp.end1  = corner;
      } else {
        edgeTmp.room2 = room;
        edgeTmp.end2  = corner;
      }
      if (iTmp >= 0)
        deletedEdges.push_back(e2Tmp);
      if (deletedEdges.size()) {
        if (deletedEdges.size() == 1)
          doc->setUndoData(new UndoChangeEdge(bTmp, deletedEdges[0]));
        else
          doc->setUndoData(new UndoChanges(bTmp, -1, 1, deletedEdges));
      } else
        doc->setUndoData(new UndoAdd(bTmp));
      doc->addEdge(edgeTmp);
    } // end if edge goes somewhere
    else {
      if (iTmp >= 0) { // We just deleted the old edge
        if ((e2Tmp.end1 == rcCorner) ||
            (!(e2Tmp.type1 & etNoRoom2) && (e2Tmp.end2 == rcCorner))) {
          doc->addEdge(e2Tmp);  // Put it back temporarily
          selectOnlyRoom((e2Tmp.end1 == rcCorner) ? e2Tmp.room1 : e2Tmp.room2);
          deleteSelection(false); // Remove entire connection
        } else // not a connection to a corner
          doc->setUndoData(new UndoDelete(bTmp, e2Tmp));
      } else if (selectedOne >= 0)
        makeRoomVisible(selectedOne); // We just clicked on a room corner
    } // end if edge doesn't go anywhere
  } // end if adding or changing edge
//....................................................................
  else if (opInProgress == gmoChangeEdge) { // We didn't change the edge
    if (selectedOne >= 0) makeRoomVisible(selectedOne);
  }
//....................................................................
  else if (opInProgress == gmoSelectBox) {
    dc.SetROP2(R2_NOT);
    CBrush* oldBrush = static_cast<CBrush*>(dc.SelectStockObject(HOLLOW_BRUSH));
    dc.Rectangle(rTmp);
    dc.SelectObject(oldBrush);
  } // end else if selection box
//....................................................................
  else if (opInProgress == gmoControlDown) {
    if (bTmp) {
      if (b2Tmp) deselectPage(iTmp);
      else       deselectRoom(iTmp);
      selectDone();
    } // end if room or page was already selected
  } // end else if gmoControlDown
//....................................................................
  else if (opInProgress >= gmoDrag) {
    CPoint  p(gridX * (point.x / gridX - p1Tmp.x),
              gridY * (point.y / gridY - p1Tmp.y));
    adjustOffset(p);
    if (p.x || p.y) {
      const CSize  offset(p);
      doc->setUndoData(new UndoMove(doc->IsModified(), offset, numSelected,
                                    selected, numSelectedPages, selectedPage));
      for (int i = doc->getRoomCount()-1; i >= 0; i--)
        if (selected[i]) doc->moveRoom(i,offset);
      for (i = doc->getPages().size()-1; i >= 0; i--)
        if (selectedPage[i]) doc->movePage(i,offset);
    } else
      InvalidateRect(NULL);
  } // end else if dragging
//....................................................................
  opInProgress = gmoNone;
} // end CMapView::OnLButtonUp

//--------------------------------------------------------------------
int CMapView::OnMouseActivate(CWnd* wnd, UINT hitTest, UINT message)
{
  const int  result = CScrollZoomView::OnMouseActivate(wnd, hitTest, message);

  if ((result == MA_ACTIVATE) && GUEmapEatClicks)
    return MA_ACTIVATEANDEAT;
  else
    return result;
} // end CMapView::OnMouseActivate

//--------------------------------------------------------------------
BOOL CMapView::OnMouseWheel(UINT nFlags, short zDelta, CPoint pt)
{
  if (nFlags == MK_CONTROL) {
    OnViewZoom((zDelta > 0) ? ID_VIEW_ZOOM_IN : ID_VIEW_ZOOM_OUT);
    return TRUE;
  }

  return CScrollZoomView::OnMouseWheel(nFlags, zDelta, pt);
} // end CMapView::OnMouseWheel

//--------------------------------------------------------------------
void CMapView::OnRButtonDown(UINT nFlags, CPoint point)
{
  if (opInProgress == gmoAddCorner)
    opInProgress = gmoNone;

  CClientDC dc(this);
  OnPrepareDC(&dc);
  dc.DPtoLP(&point);

  scrollDragStart = point;

  if (opInProgress) {
    SetCapture();
    scrollDrag = true;
    ::SetCursor(handCursor);
  } else
    scrollDragTimer = SetTimer(menuTimer, clickTicks, NULL);
} // end CMapView::OnRButtonDown

//--------------------------------------------------------------------
void CMapView::OnRButtonUp(UINT nFlags, CPoint point)
{
  if (scrollDragTimer) {
    KillTimer(scrollDragTimer);
    scrollDragTimer = 0;
    CMapDoc *const  doc = GetDocument();
    ASSERT_VALID(doc);

    CClientDC dc(this);
    OnPrepareDC(&dc);
    dc.DPtoLP(&point);
    RoomCorner  corner;
    int room = doc->roomHit(point, &corner);
    UINT id;
    if (room < 0) {
      id = IDR_MAP_MENU;
      if (showPages && (room = doc->pageBorderHit(point)) >= 0) {
        clearSelection();
        selectPage(room);
        selectDone();
      } // end if clicked on page border
    } else {
      id = IDR_ROOM_MENU;
      selectOnlyRoom(room);
      opInProgress = gmoNone;
      if ((corner != rcNone) && !doc->locked) {
        iTmp = doc->findEdge(room, corner, edgeTmp);
        if ((iTmp >= 0) && !(edgeTmp.type2 & etNoRoom2))
          opInProgress = gmoAddCorner;
      }
    } // end else clicked in room
    CMenu menu;
    VERIFY(menu.LoadMenu(id));
    CMenu* pop = menu.GetSubMenu(0);
    ASSERT(pop != NULL);

    dc.LPtoDP(&point);
    ClientToScreen(&point);
    pop->TrackPopupMenu(TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y,
                        AfxGetMainWnd());
    //opInProgress = gmoNone;  // FIXME
  } else if (!scrollDrag || (GetCapture() != this))
    CScrollZoomView::OnRButtonUp(nFlags, point);
  else {
    if (opInProgress == gmoNone)
      ReleaseCapture();
    else
      ::SetCursor(::LoadCursor(NULL, IDC_ARROW));
    scrollDrag = false;
  }
} // end CMapView::OnRButtonUp

//--------------------------------------------------------------------
void CMapView::OnSize(UINT nType, int cx, int cy)
{
  CScrollZoomView::OnSize(nType, cx, cy);

  if (initialized())  // Make sure view has been initialized
    setScrollBars();
} // end CMapView::OnSize

//--------------------------------------------------------------------
void CMapView::OnTimer(UINT idEvent)
{
  if (idEvent == menuTimer) {
    // We've held the button down long enough, switch to dragging:
    SetCapture();
    KillTimer(scrollDragTimer);
    scrollDragTimer = 0;
    scrollDrag = true;
    ::SetCursor(handCursor);
  } else
    CScrollZoomView::OnTimer(idEvent);
} // end CMapView::OnTimer

//--------------------------------------------------------------------
// Update the view after the document changes:
//
// Input:
//   lHint:
//     dupDeletedRoom:
//       A room has just been deleted
//     dupPageCount:
//       The number of pages has changed
//     dupRoomComment:
//       A room comment was just changed
//         Do not invalidate the view unless pHint is not NULL.
//     dupNavigationMode:
//       We are entering navigation mode
//         Ignore pHint, do not invalidate the view, just make sure
//         exactly one room is selected and that it's visible.
//   pHint:
//     NULL means invalidate the entire view (but see lHint)
//     Otherwise, it's a CRect* to the area to invalidate

void CMapView::OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint)
{
  if (lHint == dupDeletedRoom && opInProgress != gmoDeleting && numSelected) {
    // Clear selection unless this is the view doing the deleting:
    clearSelection();
    selectDone();
  } else if (lHint == dupPageCount && opInProgress != gmoDeleting) {
    selectedPage.resize(GetDocument()->getPages().size(), 0);
    if (numSelectedPages) {
      deselectPages();
      selectDone();
    }
  } else if (lHint == dupRoomComment) {
    // Update comment for this view, but don't override any other view:
    static_cast<CMapApp*>(AfxGetApp())->setComment(this,
      (selectedOne >= 0)
      ? &(GetDocument()->getRoom(selectedOne))
      : NULL,  false
    );
    if (!pHint) return;         // Don't invalidate entire view
  }

  if (lHint == dupNavigationMode) {
    const CMapDoc* doc = GetDocument();
    ASSERT_VALID(doc);
    const int numRooms = doc->getRoomCount();
    if (selectedOne < 0 || doc->getRoom(selectedOne).isCorner())
      selectOnlyRoom(0);
    makeRoomVisible(selectedOne);
  } // end if switching to navigation mode
  else if (pHint) {
    CRect rect(*reinterpret_cast<const CRect*>(pHint));
    CClientDC dc(this);
    OnPrepareDC(&dc);
    dc.LPtoDP(&rect);
    rect.NormalizeRect();
    InvalidateRect(rect);
  }
  else if (lHint != dupRoomComment)
    Invalidate();
} // end CMapView::OnUpdate

//--------------------------------------------------------------------
void CMapView::OnUpdateEditAddCorner(CCmdUI* pCmdUI)
{
  pCmdUI->Enable(opInProgress == gmoAddCorner);
} // end CMapView::OnUpdateEditAddCorner

//--------------------------------------------------------------------
void CMapView::OnUpdateEditPaste(CCmdUI* pCmdUI)
{
  pCmdUI->Enable((static_cast<CMapApp*>(AfxGetApp())->getClipboard() != NULL)
                 && !GetDocument()->locked);
} // end CMapView::OnUpdateEditPaste

//--------------------------------------------------------------------
// Commands which require selected rooms and unlocked document:
//
// Edit|Clear
// Edit|Cut
// Edit|Copy
// Edit|Select connected

void CMapView::OnUpdateSelectedUnlocked(CCmdUI* pCmdUI)
{
  CMapDoc *const doc = GetDocument();
  ASSERT_VALID(doc);

  bool  selected = numSelected;
  if ((pCmdUI->m_nID != ID_EDIT_SELECT_CONNECTED) && numSelectedPages) {
    if ((pCmdUI->m_nID != ID_EDIT_COPY) &&
        (doc->getPages().size() == numSelectedPages))
      selected = false;         // Can't delete all pages
    else
      selected = true;
  } // end if not Edit|Select connected

  pCmdUI->Enable(selected && !doc->locked);
} // end CMapView::OnUpdateEditClear

//--------------------------------------------------------------------
// Commands which require the document to be unlocked:
//
// Edit|Select all

void CMapView::OnUpdateUnlocked(CCmdUI* pCmdUI)
{
  pCmdUI->Enable(!GetDocument()->locked);
} // end CMapView::OnUpdateUnlocked

//--------------------------------------------------------------------
// Toggle the grid on and off:

void CMapView::OnViewGrid()
{
  showGrid = !showGrid;
  InvalidateRect(NULL);
} // end CMapView::OnViewGrid

void CMapView::OnUpdateViewGrid(CCmdUI* pCmdUI)
{
  pCmdUI->SetCheck(showGrid);
} // end CMapView::OnUpdateViewGrid

//--------------------------------------------------------------------
void CMapView::OnViewZoom(UINT cmd)
{
  ASSERT((cmd == ID_VIEW_ZOOM_IN) || (cmd == ID_VIEW_ZOOM_OUT));

  if (cmd == ID_VIEW_ZOOM_OUT)
    zoomTo(zoom % 10 ? zoom - zoom % 10 : zoom - 10);
  else
    zoomTo(zoom + 10 - zoom % 10);
} // end CMapView::OnViewZoom

void CMapView::OnUpdateViewZoom(CCmdUI* pCmdUI)
{
  pCmdUI->Enable((pCmdUI->m_nID == ID_VIEW_ZOOM_IN)
                 ? (zoom < maxZoom)
                 : (zoom > minZoom));
} // end CMapView::OnUpdateViewZoom

/////--------------------------------------------------------------------
///// Fix drawing anomalies after scrolling:
///
///void CMapView::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
///{
///  CScrollView::OnVScroll(nSBCode, nPos, pScrollBar);
///
///  if (numSelected) {
///    const CMapDoc* doc = GetDocument();
///    ASSERT_VALID(doc);
///
///    CClientDC dc(this);
///    OnPrepareDC(&dc);
///    CRect  rect;
///    for (int i = doc->getRoomCount()-1; i >= 0; --i)
///      if (selected[i]) {
///        doc->getRoom(i).getRect(rect);
///        dc.LPtoDP(&rect);
///        rect.NormalizeRect();
///        InvalidateRect(rect);
///      } // end if room i selected
///  } // end if selected rooms
///} // end CMapView::OnVScroll

//--------------------------------------------------------------------
// Redraw the window:

void CMapView::OnWindowRefresh()
{
  InvalidateRect(NULL);
} // end CMapView::OnWindowRefresh



/////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////////////////////
// CRepaginateDlg dialog

CRepaginateDlg::CRepaginateDlg()
: CDialog(CRepaginateDlg::IDD, NULL)
{
	//{{AFX_DATA_INIT(CRepaginateDlg)
	//}}AFX_DATA_INIT
}

BEGIN_MESSAGE_MAP(CRepaginateDlg, CDialog)
	//{{AFX_MSG_MAP(CRepaginateDlg)
	ON_BN_CLICKED(IDYES, OnYes)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP();

/////////////////////////////////////////////////////////////////////////////
// CRepaginateDlg message handlers

BOOL CRepaginateDlg::OnInitDialog()
{
  CDialog::OnInitDialog();

  SendDlgItemMessage(IDC_EXCLAMATION, STM_SETICON,
                     (WPARAM) gueApp()->LoadStandardIcon(IDI_EXCLAMATION));
  MessageBeep(MB_ICONEXCLAMATION);

  return TRUE;  // return TRUE unless you set the focus to a control
} // end CRepaginateDlg::OnInitDialog

void CRepaginateDlg::OnYes()
{
  EndDialog(IDYES);
} // end CRepaginateDlg::OnYes
