//--------------------------------------------------------------------
// $Id: ScrollZoom.cpp 4718 2007-09-15 02:34:13Z cjm $
//--------------------------------------------------------------------
//
//   GUEmap
//   Copyright 1999-2007 by Christopher J. Madsen
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation; either version 2 of the License, or
//   (at your option) any later version.
//
//   This program is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License along
//   with this program; if not, write to the Free Software Foundation, Inc.,
//   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
//
//   Implementation of the CScrollZoomView class
//
//--------------------------------------------------------------------

#include "StdAfx.hpp"
#include "ScrollZoom.hpp"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// Miscellaneous functions:
//--------------------------------------------------------------------
// Compute the greatest common divisor of M and N:

static int gcd(int m, int n)
{
  if (m < n) swap(m, n);

  int r;
  while (r = m % n) {
    m = n;
    n = r;
  } // end while remainder

  return n;
} // end gcd


/////////////////////////////////////////////////////////////////////////////
// Class CScrollZoomView:
//
// Member Variables:
//   scrollLine:
//     The amount to scroll when scrolling one line (in logical pixels)
//   scrollPage:
//     The amount to scroll when scrolling one pago (in logical pixels)
//   scrollStep:
//     The minimum amount to scroll (in logical pixels).
//     This should be the smallest number of logical pixels that equals
//     an integer number of device pixels.
//--------------------------------------------------------------------

IMPLEMENT_DYNAMIC(CScrollZoomView, CView);

CScrollZoomView::CScrollZoomView()
: CView(),
  displayPerInch(0),
  realPerInch(0),
  docSize(0,0),
  offset(0,0),
  scrollLine(0,0),
  scrollPage(0,0)
{
}

BEGIN_MESSAGE_MAP(CScrollZoomView, CView)
	//{{AFX_MSG_MAP(CScrollZoomView)
	ON_WM_HSCROLL()
	ON_WM_SIZE()
	ON_WM_VSCROLL()
	//}}AFX_MSG_MAP
END_MESSAGE_MAP();

//--------------------------------------------------------------------
// Initialize the view:
//
// Input:
//   logPerInch:  The number of logical pixels per inch
//   theDocSize:  The size of the document (in logical pixels)
//   lineStep:    The amount to scroll by one line (in logical pixels)

void CScrollZoomView::init(int logPerInch, const SIZE& theDocSize,
                           const SIZE& lineStep)
{
  docSize = theDocSize;
  ASSERT(logPerInch > 0);
  ASSERT(docSize.cx >= 0 && docSize.cy >= 0);
  ASSERT(lineStep.cx > 0 && lineStep.cy > 0);

  EnableScrollBarCtrl(SB_BOTH);
  realPerInch = logPerInch;
  setDisplayScale(logPerInch);

  scrollLine = lineStep;
  scrollPage = lineStep;
} // end CScrollZoomView::init

/////////////////////////////////////////////////////////////////////////////
// CScrollZoomView diagnostics
//--------------------------------------------------------------------
#ifdef _DEBUG
void CScrollZoomView::AssertValid() const
{
	CView::AssertValid();
}

void CScrollZoomView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CScrollZoomView drawing
//--------------------------------------------------------------------
void CScrollZoomView::OnPrepareDC(CDC* pDC, CPrintInfo* pInfo)
{
  pDC->SetMapMode(MM_ISOTROPIC);
  ASSERT(realPerInch > 0);
  if (pDC->IsPrinting())
    pDC->SetWindowExt(realPerInch, -realPerInch);
  else {
    pDC->SetWindowExt(displayPerInch, -displayPerInch);
    pDC->SetWindowOrg(offset);
  }
  pDC->SetViewportExt(pDC->GetDeviceCaps(LOGPIXELSX),
                      pDC->GetDeviceCaps(LOGPIXELSY));
} // end CScrollZoomView::OnPrepareDC

/////////////////////////////////////////////////////////////////////////////
// CScrollZoomView scrolling
//--------------------------------------------------------------------
void CScrollZoomView::OnHScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
{
  if (pScrollBar != NULL && pScrollBar->SendChildNotifyLastMsg())
    return;     // eat it

	// ignore scroll bar msgs from other controls
  if (pScrollBar != GetScrollBarCtrl(SB_HORZ))
    return;

  OnScroll(MAKEWORD(nSBCode, -1), nPos);
} // end CScrollZoomView::OnHScroll

//--------------------------------------------------------------------
void CScrollZoomView::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar)
{
  if (pScrollBar != NULL && pScrollBar->SendChildNotifyLastMsg())
    return;     // eat it

  // ignore scroll bar msgs from other controls
  if (pScrollBar != GetScrollBarCtrl(SB_VERT))
    return;

  OnScroll(MAKEWORD(-1, nSBCode), nPos);
} // end CScrollZoomView::OnVScroll

//--------------------------------------------------------------------
BOOL CScrollZoomView::OnScroll(UINT nScrollCode, UINT nPos, BOOL bDoScroll)
{
  // calc new x position
  SCROLLINFO  xBar;
  VERIFY(GetScrollInfo(SB_HORZ, &xBar, SIF_PAGE|SIF_POS|SIF_RANGE));
  int x = xBar.nPos;

  switch (LOBYTE(nScrollCode)) {
   case SB_TOP:        x = 0;              break;
   case SB_BOTTOM:     x = INT_MAX;        break;
   case SB_LINEUP:     x -= scrollLine.cx; break;
   case SB_LINEDOWN:   x += scrollLine.cx; break;
   case SB_THUMBTRACK: x = nPos;           break;
   case SB_PAGEUP:     x -= scrollPage.cx; break;
   case SB_PAGEDOWN:   x += scrollPage.cx; break;
   case SB_ENDSCROLL:  return FALSE;
  } // end switch x

  // calc new y position
  SCROLLINFO  yBar;
  VERIFY(GetScrollInfo(SB_VERT, &yBar, SIF_PAGE|SIF_POS|SIF_RANGE));
  int y = -yBar.nPos;

  switch (HIBYTE(nScrollCode)) {
   case SB_TOP:        y = 0;              break;
   case SB_BOTTOM:     y = INT_MIN;        break;
   case SB_LINEUP:     y += scrollLine.cy; break;
   case SB_LINEDOWN:   y -= scrollLine.cy; break;
   case SB_THUMBTRACK: y = -nPos;          break;
   case SB_PAGEUP:     y += scrollPage.cy; break;
   case SB_PAGEDOWN:   y -= scrollPage.cy; break;
   case SB_ENDSCROLL:  return FALSE;
  } // end switch y

  if (x < 0) x = 0;
  if (y > 0) y = 0;

  if (x + xBar.nPage > xBar.nMax)
    x = xBar.nMax - xBar.nPage;
  if (-y + yBar.nPage > yBar.nMax)
    y = yBar.nPage - yBar.nMax;

  x -= x % scrollStep.cx;
  y -= y % scrollStep.cy;

  if ((offset.x == x) && (offset.y == y)) return FALSE; // Didn't scroll

  if (bDoScroll) {
    CSize  delta(x - offset.x, y - offset.y);
    offset.x = x;
    offset.y = y;
    CClientDC  dc(this);
    OnPrepareDC(&dc);
    dc.LPtoDP(&delta);

    SetScrollPos(SB_VERT, -y);
    SetScrollPos(SB_HORZ, x);
    ScrollWindow(-delta.cx, delta.cy);
    UpdateWindow();
  }

  return TRUE;
} // end CScrollZoomView::OnScroll

//--------------------------------------------------------------------
// Calculate the nearest acceptable scrolling position:
//
// Input:
//   pt:  The point you want to scroll to (in logical coordinates)
//
// Output:
//   pt:  The nearest acceptable offset
//
// Returns:
//   true:   The new offset is not the same as the current one
//   false:  The view is already scrolled to the desired location

bool CScrollZoomView::calcScrollPosition(POINT& pt) const
{
  if (pt.x < 0) pt.x = 0;
  if (pt.y > 0) pt.y = 0;

  SCROLLINFO  si;
  VERIFY(const_cast<CScrollZoomView*>(this)->
         GetScrollInfo(SB_HORZ, &si, SIF_PAGE|SIF_POS|SIF_RANGE));
  if (pt.x + si.nPage > si.nMax + scrollStep.cx)
    pt.x = si.nMax - si.nPage + scrollStep.cx;

  VERIFY(const_cast<CScrollZoomView*>(this)->
         GetScrollInfo(SB_VERT, &si, SIF_PAGE|SIF_POS|SIF_RANGE));
  if (-pt.y + si.nPage > si.nMax + scrollStep.cy)
    pt.y = si.nPage - si.nMax - scrollStep.cy;

  pt.x -= pt.x % scrollStep.cx;
  pt.y -= pt.y % scrollStep.cy;

  return offset != pt;
} // end CScrollZoomView::calcScrollPosition

//--------------------------------------------------------------------
// Scroll to a position:
//
// Input:
//   pt:  The new upper left corner of the view (in logical coordinates)

void CScrollZoomView::scrollToPosition(const POINT& pt)
{
  CPoint  newOffset(pt);

  if (!calcScrollPosition(newOffset))
    return; // Didn't scroll

  CSize  delta(newOffset - offset);
  offset = newOffset;

  CClientDC  dc(this);
  OnPrepareDC(&dc);
  dc.LPtoDP(&delta);

  SetScrollPos(SB_VERT, -offset.y);
  SetScrollPos(SB_HORZ, offset.x);
  ScrollWindow(-delta.cx, delta.cy);
} // end CScrollZoomView::scrollToPosition

//--------------------------------------------------------------------
// Update the size of the scrollbar thumbs:

void CScrollZoomView::updateScrollBars()
{
  CRect  clientRect;
  GetClientRect(&clientRect);
  CSize  viewSize(clientRect.right, clientRect.bottom);

  CClientDC  cdc(this);
  OnPrepareDC(&cdc);
  cdc.DPtoLP(&viewSize);

  SCROLLINFO info;
  info.fMask = SIF_PAGE|SIF_DISABLENOSCROLL;
  info.nMin = 0;

  info.nPage = viewSize.cy;
  SetScrollInfo(SB_VERT, &info, TRUE);

  info.nPage = viewSize.cx;
  SetScrollInfo(SB_HORZ, &info, TRUE);
} // end CScrollZoomView::OnSize

//--------------------------------------------------------------------
BOOL CScrollZoomView::OnScrollBy(CSize sizeScroll, BOOL bDoScroll)
{
  ASSERT(FALSE); // Not implemented, I don't think I need it.

  return FALSE;
} // end CScrollZoomView::OnScrollBy

/////////////////////////////////////////////////////////////////////////////
// CScrollZoomView zooming
//--------------------------------------------------------------------
// Set the display scale:
//
// This is an internal function, combining actions needed by init and
// setDisplayZoom.
//
// Input:
//   logPerInch:  The number of logical pixels per inch to display

void CScrollZoomView::setDisplayScale(int logPerInch)
{
///  TRACE("Zooming to %.3f%% (%d)\n", (100.0*realPerInch) / logPerInch,
///        logPerInch);

  CClientDC  dc(this);
  int  i, minStep, step;
  const int
    displayX = dc.GetDeviceCaps(LOGPIXELSX),
    stepBy   = 8,
    range    = (logPerInch > realPerInch) ? 16 : 8,
    stopAt   = logPerInch - logPerInch % stepBy + range;

  minStep = 32767;

  i = stopAt - 2 * range;
  if (logPerInch - i > range) i += stepBy;
  for (; i <= stopAt; i += stepBy) {
    step = (i / gcd(i, displayX));
///    TRACE("  Trying LPI: %d (%.3f%%) scroll: %d\n", i,
///          (100.0*realPerInch) / i, step);
    if (step < minStep) {
      displayPerInch = i;
      minStep = step;
    }
  }

  scrollStep.cx = minStep;
  scrollStep.cy = (displayPerInch /
                   gcd(displayPerInch, dc.GetDeviceCaps(LOGPIXELSY)));

  SetScrollRange(SB_HORZ, 0,
                 docSize.cx + scrollStep.cx - (docSize.cx % scrollStep.cx),
                 FALSE);
  SetScrollRange(SB_VERT, 0,
                 docSize.cy + scrollStep.cy - (docSize.cy % scrollStep.cy),
                 FALSE);

///  TRACE("LPI: %d (%.3f%%) scroll: %d %d  (wanted %d, diff %d)\n",
///        displayPerInch, (100.0 * realPerInch)/displayPerInch, scrollStep.cx,
///        scrollStep.cy, logPerInch, displayPerInch - logPerInch);
} // end CScrollZoomView::setDisplayScale

//--------------------------------------------------------------------
// Zoom the display:
//
// Input:
//   percent:  The new zoom factor (in percent)

void CScrollZoomView::setDisplayZoom(int percent)
{
  setDisplayScale((realPerInch * 100) / percent);
  scrollToPosition(offset);     // Make sure it aligns with new scrollStep
  updateScrollBars();
  Invalidate();
} // end CScrollZoomView::setDisplayZoom

/////////////////////////////////////////////////////////////////////////////
// CScrollZoomView message handlers
//--------------------------------------------------------------------
void CScrollZoomView::OnSize(UINT nType, int cx, int cy)
{
  CView::OnSize(nType, cx, cy);

  if (realPerInch)
    updateScrollBars();
} // end CScrollZoomView::OnSize
